/*  -*- Mode: c++ -*-
 *
 *  $Id: xotcl.c,v 1.43 2001/03/09 13:34:01 neumann Exp $
 *
 *  XOTcl - Extended OTcl
 *
 *  Copyright (C) 1999-2002 Gustaf Neumann (a), Uwe Zdun (b)
 *
 * (a) Vienna University of Economics and Business Administration
 *     Dept. of Information Systems / New Media
 *     A-1090, Augasse 2-6
 *     Vienna, Austria
 *
 * (b) University of Essen
 *     Specification of Software Systems
 *     Altendorferstrae 97-101
 *     D-45143 Essen, Germany
 *
 *  Permission to use, copy, modify, distribute, and sell this
 *  software and its documentation for any purpose is hereby granted
 *  without fee, provided that the above copyright notice appear in
 *  all copies and that both that copyright notice and this permission
 *  notice appear in supporting documentation. We make no
 *  representations about the suitability of this software for any
 *  purpose.  It is provided "as is" without express or implied
 *  warranty.
 *
 *
 *  This software is based upon MIT Object Tcl by David Wetherall and
 *  Christopher J. Lindblad, that contains the following copyright
 *  message:
 *
 *   "Copyright 1993 Massachusetts Institute of Technology
 *
 *    Permission to use, copy, modify, distribute, and sell this
 *    software and its documentation for any purpose is hereby granted
 *    without fee, provided that the above copyright notice appear in
 *    all copies and that both that copyright notice and this
 *    permission notice appear in supporting documentation, and that
 *    the name of M.I.T. not be used in advertising or publicity
 *    pertaining to distribution of the software without specific,
 *    written prior permission.  M.I.T. makes no representations about
 *    the suitability of this software for any purpose.  It is
 *    provided "as is" without express or implied warranty."
 * */

#include "xotclInt.h"

#ifdef KEEP_TCL_CMD_TYPE
# include "tclCompile.h"
#endif

/*
 * Tcl_Obj Types for XOTcl Objects
 */

#ifdef USE_TCL_STUBS
# define XOTcl_ExprObjCmd(cd,in,objc,objv) \
	callCommand(in, EXPR, objc, objv)
# define XOTcl_IncrObjCmd(cd,in,objc,objv) \
	callCommand(in, INCR, objc, objv)
#else
# define XOTcl_ExprObjCmd(cd,in,objc,objv) \
	Tcl_ExprObjCmd(cd, in, objc, objv);
# define XOTcl_IncrObjCmd(cd,in,objc,objv) \
	Tcl_IncrObjCmd(cd, in, objc, objv);
#endif


# define NEW(type) \
	(type *)ckalloc(sizeof(type)); MEM_COUNT_ALLOC(#type, NULL)
# define NEW_ARRAY(type,n) \
	(type *)ckalloc(sizeof(type)*(n)); MEM_COUNT_ALLOC(#type "*", NULL)
# define FREE(type, var) \
	ckfree((char*) var); MEM_COUNT_FREE(#type,var)

static int SetXOTclObjectFromAny(Tcl_Interp* interp, Tcl_Obj* objPtr);
static void UpdateStringOfXOTclObject(Tcl_Obj* objPtr);
static void FreeXOTclObjectInternalRep(Tcl_Obj * objPtr);
static void DupXOTclObjectInternalRep(Tcl_Obj *src, Tcl_Obj *cpy);

static Tcl_ObjType XOTclObjectType = {
  "XOTclObject",			
  FreeXOTclObjectInternalRep,
  DupXOTclObjectInternalRep,
  UpdateStringOfXOTclObject,	
  SetXOTclObjectFromAny		
};

/* these are names and contents for global (corresponding) Tcl_Objs
   and Strings - otherwise these "constants" would have to be built
   every time they are used; now they are built once in XOTcl_Init */
typedef enum {
    EMPTY, UNKNOWN, CREATE, DESTROY, INSTDESTROY, ALLOC,
    INIT, INSTVAR, INTERP, AUTONAMES,
    ZERO, ONE, MOVE, SELF, CLASS, RECREATE,
    SELF_CLASS, SELF_PROC, PARAM_CL,
    DEFAULT, SEARCH_DEFAULTS, EXIT_HANDLER,
    CLEANUP, CONFIGURE, FILTER, INSTFILTER,
    INSTPROC, PROC, MKGETTERSETTER, FORMAT, NEWOBJ,
    /** these are the redefined tcl commands; leave them
	together at the end */
    EXPR, INCR, INFO, RENAME, SUBST, UPLEVEL, UPVAR
} global_names;
static char *global_strings[] = {
  "", "unknown", "create", "destroy", "instdestroy", "alloc",
  "init", "instvar", "interp", "__autonames",
  "0", "1", "move", "self", "class", "recreate",
  "self class", "self proc",  "::xotcl::Class::Parameter",
  "default", "searchDefaults", "__exitHandler",
  "cleanup", "configure", "filter", "instfilter",
  "instproc", "proc", "mkGetterSetter", "format", "__#",
  "expr", "incr", "info", "rename", "subst", "uplevel", "upvar"
};
static TclObjCmdProcType tcl_commands[UPVAR-EXPR+1];

#define global_objects RUNTIME_STATE(in)->methodObjNames

static int ObjDispatch(ClientData cd, Tcl_Interp* in,
		       int objc, Tcl_Obj *CONST objv[],
		       int flags);
static int XOTclNextMethod(XOTclObject* obj, Tcl_Interp* in,
			   XOTclClass *givenCl,
			   char *givenMethod, int objc, Tcl_Obj *CONST objv[],
			   int useCSObjs, int *isDestroyed);

static int TclCommands(Tcl_Interp* in, int load);

static XOTclObject* GetObject(Tcl_Interp* in, char* name);
static XOTclClass*  GetClass(Tcl_Interp* in, char* name);

#if !defined(NDEBUG)
static void checkAllInstances(Tcl_Interp *in, XOTclClass* startCl, int lvl);
#endif

static XOTclCallStackContent*
CallStackGetFrame(Tcl_Interp* in);

#ifdef PRE81
/* for backward compatibility only
*/
static int
Tcl_EvalObjv(Tcl_Interp* in,
	     int objc, Tcl_Obj *CONST objv[], int flags) {
  int i, result;
  Tcl_DString ds, *dsp = &ds;

  assert(flags == 0);
  DSTRING_INIT(dsp);
  for (i = 0; i < objc; i++) {
    Tcl_DStringAppendElement(dsp, ObjStr(objv[i]));
  }
  result = Tcl_Eval(in, Tcl_DStringValue(dsp));
  DSTRING_FREE(dsp);
  return result;
}
static int
Tcl_IncrObjCmd(ClientData cd, Tcl_Interp *in, int objc, Tcl_Obj *CONST objv[]) {
  int result;
  char *argv[3];
  argv[0] = global_strings[INCR];
  argv[1] = ObjStr(objv[1]);
  if (objc == 3)
    argv[2] = ObjStr(objv[2]);
  result = Tcl_IncrCmd(cd, in, objc, argv);
  return result;
}
static int
Tcl_SubstObjCmd(ClientData cd, Tcl_Interp* in,
		int objc, Tcl_Obj *CONST objv[]) {
  char *ov[20];
  int i;
  assert(objc<19);
  for (i=0; i<objc; i++)
    ov[i] = ObjStr(objv[i]);

  return Tcl_SubstCmd(cd, in, objc, ov);
}
# define SUBST_CMD Tcl_SubstObjCmd
#else
# define SUBST_CMD 0
#endif

/*
 * call a Tcl command with given objv's ... replace objv[0]
 * with the given command name
 */
static XOTCLINLINE int
callCommand(Tcl_Interp* in,  global_names name,
	    int objc, Tcl_Obj *CONST objv[]) {
  int result;
  DEFINE_NEW_TCL_OBJECTS_ON_STACK(objc, ov);

  /*
  fprintf(stderr,"calling %s (%p)\n",
     global_strings[name],tcl_commands[name]);
  */
  ov[0] = global_objects[name];
  if (objc > 1)
    memcpy(ov+1, objv+1, sizeof(Tcl_Obj *)*(objc-1));
  result = (*tcl_commands[name-EXPR])(NULL, in, objc, ov);
  FREE_TCL_OBJECTS_ON_STACK(ov);
  return result;
}

/*
 * call an XOTcl method
 */
static int
callMethod(ClientData cd, Tcl_Interp* in, Tcl_Obj* method,
	   int objc, Tcl_Obj *CONST objv[], int flags) {
  XOTclObject* obj = (XOTclObject*) cd;
  int result;
  DEFINE_NEW_TCL_OBJECTS_ON_STACK(objc, tov);

#ifdef OBJDELETION_TRACE
  if (method == global_objects[CLEANUP] ||
      method == global_objects[DESTROY]) {
    fprintf(stderr, "%s->%s id=%p destroyCalled=%d\n",
	    ObjStr(obj->cmdName), ObjStr(method), obj,
	    (obj->flags & XOTCL_DESTROY_CALLED));
  }
#endif
  tov[0] = obj->cmdName;
  tov[1] = method;
  if (objc>2)
    memcpy(tov+2, objv, sizeof(Tcl_Obj *)*(objc-2));

  result = ObjDispatch(cd, in, objc, tov, flags);
  /*fprintf(stderr, "     callMethod returns %d\n", result);*/
  FREE_TCL_OBJECTS_ON_STACK(tov);
  return result;
}

static int
callMethodWithArg(ClientData cd, Tcl_Interp* in,
		  Tcl_Obj* method, Tcl_Obj* arg,
		  int objc, Tcl_Obj *CONST objv[], int flags) {
  XOTclObject* obj = (XOTclObject*) cd;
  int result;
  DEFINE_NEW_TCL_OBJECTS_ON_STACK(objc, tov);

  tov[0] = obj->cmdName;
  tov[1] = method;
  tov[2] = arg;
  if (objc>3)
    memcpy(tov+3, objv, sizeof(Tcl_Obj *)*(objc-3));

  result = ObjDispatch(cd, in, objc, tov, flags);

  FREE_TCL_OBJECTS_ON_STACK(tov);
  return result;
}

/*
 *  realize self, class, proc through the [self] command
 */

static CONST84 XOTCLINLINE char*
GetSelfProc(Tcl_Interp *in) {
  return Tcl_GetCommandName(in, (Tcl_Command) CallStackGetFrame(in)->cmdPtr);
}

static XOTCLINLINE XOTclClass*
GetSelfClass(Tcl_Interp *in) {
  return CallStackGetFrame(in)->cl;
}

static XOTCLINLINE XOTclObject*
GetSelfObj(Tcl_Interp* in) {
  return CallStackGetFrame(in)->self;
}

static XOTCLINLINE Command*
GetSelfProcCmdPtr(Tcl_Interp *in) {
  return CallStackGetFrame(in)->cmdPtr;
}

/*
 * prints a msg to the screen that oldCmd is deprecated
 * optinal: give a new cmd
 */
extern void
XOTclDeprecatedMsg(char* oldCmd, char* newCmd) {
  fprintf(stderr, "**\n**\n** The command/method <%s> is deprecated.\n", oldCmd);
  if (newCmd)
    fprintf(stderr, "** Use <%s> instead.\n", newCmd);
  fprintf(stderr, "**\n");
}

static int
XOTcl_DeprecatedCmd(ClientData cd, Tcl_Interp *in, int objc, Tcl_Obj *CONST objv[]) {
  char* new;
  if (objc == 2)
    new = 0;
  else if (objc == 3)
    new = ObjStr(objv[2]);
  else
    return XOTclObjErrArgCnt(in, NULL, "deprecated oldcmd ?newcmd?");
  XOTclDeprecatedMsg(ObjStr(objv[1]), new);
  return TCL_OK;
}
#ifdef DISPATCH_TRACE
static void printObjv(int objc, Tcl_Obj *CONST objv[]) {
  int i, j;
  if (objc <= 3) j=objc; else j=3;
  for (i=0;i<j;i++) fprintf(stderr, " %s", ObjStr(objv[i]));
  if (objc > 3) fprintf(stderr," ...");
  fprintf(stderr," (objc=%d)",objc);
}

static void printCall(Tcl_Interp *in, char *string, 
		      int objc, Tcl_Obj *CONST objv[]) {
  Interp *iPtr = (Interp *) in;
  fprintf(stderr, "     (%d) >%s: ", iPtr->numLevels, string);
  printObjv(objc, objv);
  fprintf(stderr, "\n");
}
static void printExit(Tcl_Interp *in, char *string, 
		      int objc, Tcl_Obj *CONST objv[], int result) {
  Interp *iPtr = (Interp *) in;
  fprintf(stderr, "     (%d) <%s: ", iPtr->numLevels, string);
  printObjv(objc, objv);
  fprintf(stderr, " result=%d\n", result);
}
#endif


/*
 *  XOTclObject Reference Accounting
 */
#if defined(XOTCLOBJ_TRACE)
# define XOTclObjectRefCountIncr(obj) \
  obj->refCount++; \
  fprintf(stderr, "RefCountIncr %p count=%d %s\n", obj, obj->refCount,obj->cmdName?ObjStr(obj->cmdName):"no name"); \
  MEM_COUNT_ALLOC("XOTclObject RefCount",obj)
# define XOTclObjectRefCountDecr(obj) \
  obj->refCount--; \
  fprintf(stderr, "RefCountDecr %p count=%d\n", obj, obj->refCount); \
  MEM_COUNT_FREE("XOTclObject RefCount", obj)
#else 
# define XOTclObjectRefCountIncr(obj) \
  obj->refCount++; \
  MEM_COUNT_ALLOC("XOTclObject RefCount",obj)
# define XOTclObjectRefCountDecr(obj) \
  obj->refCount--; \
  MEM_COUNT_FREE("XOTclObject RefCount",obj)
#endif

#if defined(XOTCLOBJ_TRACE)
void objTrace(char* string, XOTclObject *obj) {
  if(obj)
    fprintf(stderr,"--- %s tcl %p (%d) xotcl %p (%d) %s \n", string,
	    obj->cmdName, obj->cmdName->refCount, obj, obj->refCount, ObjStr(obj->cmdName));
  else
    fprintf(stderr,"--- No object: %s\n",string);
}
#else
# define objTrace(a,b)
#endif


/*
 * Conversion from CmdPtr to Class / Object
 */

static XOTCLINLINE XOTclClass*
GetClassFromCmdPtr(Command* cmdPtr) {
  if (cmdPtr && cmdPtr->objProc == XOTclObjDispatch &&
      !cmdPtr->cmdEpoch && cmdPtr->objClientData)
    return XOTclObjectToClass(cmdPtr->objClientData);
  else
    return 0;
}

static XOTCLINLINE XOTclObject*
GetObjectFromCmdPtr(Command* cmdPtr) {
  if (cmdPtr && cmdPtr->objProc == XOTclObjDispatch &&
      !cmdPtr->cmdEpoch) {
    return (XOTclObject*)cmdPtr->objClientData;
  }
  else
    return 0;
}

static void XOTclCleanupObject(XOTclObject* obj) {
  XOTclObjectRefCountDecr(obj);
  if (obj->refCount <= 0) {
    assert(obj->refCount == 0);
    assert(obj->flags & XOTCL_DESTROYED);
#if REFCOUNTED
    fprintf(stderr,"###CLNO %p flags %d rc %d destr %d dc %d\n",
	    obj, obj->flags,
	    (obj->flags & XOTCL_REFCOUNTED) != 0,
	    (obj->flags & XOTCL_DESTROYED) != 0,
	    (obj->flags & XOTCL_DESTROY_CALLED) != 0
	    );
#endif
    MEM_COUNT_FREE("XOTclObject/XOTclClass",obj);
#if defined(XOTCLOBJ_TRACE)
    fprintf(stderr, "CKFREE Object %p refcount=%d\n", obj, obj->refCount); 
#endif
#if !defined(NDEBUG)
    memset(obj, 0, sizeof(XOTclObject));
#endif
    ckfree((char *) obj);
  }
}

/*
 *  Tcl_Obj functions for objects
 */
static void RegisterObjTypes() {
  Tcl_RegisterObjType(&XOTclObjectType);
}

static void
FreeXOTclObjectInternalRep(register Tcl_Obj * objPtr) {
  XOTclObject* obj = (XOTclObject*) objPtr->internalRep.otherValuePtr;
#if defined(XOTCLOBJ_TRACE)
  if (obj) 
    fprintf(stderr,"FIP --- tcl %p (%d)\n",objPtr,objPtr->refCount);
#endif
  if (obj != NULL) {
#if REFCOUNTED
    fprintf(stderr,"FIP %p flags %d rc %d destr %d dc %d\n",
	    obj, obj->flags,
	    (obj->flags & XOTCL_REFCOUNTED) != 0,
	    (obj->flags & XOTCL_DESTROYED) != 0,
	    (obj->flags & XOTCL_DESTROY_CALLED) != 0
	    );
    if (obj->flags & XOTCL_REFCOUNTED &&
	!(obj->flags & XOTCL_DESTROYED)) {
      Tcl_Interp *in = obj->teardown;
      obj->teardown = NULL;
      fprintf(stderr, " ... teardown = %p\n", obj->teardown);
      /*obj->flags &= ~XOTCL_REFCOUNTED;*/
      callDestroyMethod((ClientData)obj, in, obj, 0);
      fprintf(stderr, " ... destroy method done \n");
    } else {
      fprintf(stderr, "BEFORE CLEANUPOBJ %d\n",
	      (obj->flags & XOTCL_REFCOUNTED));
#endif
      XOTclCleanupObject(obj);
#if REFCOUNTED
      fprintf(stderr, "AFTER CLEANUPOBJ\n");
    }
#endif
    objPtr->internalRep.otherValuePtr = NULL;
    objPtr->typePtr = NULL;
  }
}

static void 
DupXOTclObjectInternalRep(Tcl_Obj *src, Tcl_Obj *cpy) {
  XOTclObject* obj = (XOTclObject*)src->internalRep.otherValuePtr;
#if defined(XOTCLOBJ_TRACE)
  if (obj) fprintf(stderr,"DIP --- tcl %p (%d)\n", src, src->refCount);
#endif
  XOTclObjectRefCountIncr(obj);
  cpy->internalRep.otherValuePtr = src->internalRep.otherValuePtr;
  cpy->typePtr = src->typePtr;
}

static int
SetXOTclObjectFromAny(Tcl_Interp *interp, register Tcl_Obj * objPtr) {
  Tcl_ObjType *oldTypePtr = objPtr->typePtr;
  char *string = ObjStr(objPtr);
  XOTclObject* obj;
  int result = TCL_OK;

#ifdef XOTCLOBJ_TRACE
  fprintf(stderr,"SetXOTclObjectFromAny %p '%s' %p\n",
	  objPtr,string,objPtr->typePtr);
  if (oldTypePtr)
    fprintf(stderr,"   convert %s to XOTclObject\n", oldTypePtr->name);
#endif

  obj = GetObject(interp, string);
  if (obj) {
    if ((oldTypePtr != NULL) && (oldTypePtr->freeIntRepProc != NULL)) {
#ifdef XOTCLOBJ_TRACE
      fprintf(stderr,"   freeing type=%p, xottyp=%p\n",
	      objPtr->typePtr, &XOTclObjectType);
#endif
      oldTypePtr->freeIntRepProc(objPtr);
    }
    XOTclObjectRefCountIncr(obj);
#if defined(XOTCLOBJ_TRACE)
    fprintf(stderr, "SetXOTclObjectFromAny tcl %p (%d) xotcl %p (%d)\n",
	    objPtr, objPtr->refCount, obj, obj->refCount);
#endif
    objPtr->internalRep.otherValuePtr = (XOTclObject*) obj;
    objPtr->typePtr = &XOTclObjectType;
  } else
    result = TCL_ERROR;

  return result;
}

static void
UpdateStringOfXOTclObject(register Tcl_Obj *objPtr) {
  XOTclObject *obj = (XOTclObject*) objPtr->internalRep.otherValuePtr;
  char *nsFullName = NULL;
  register Command *cmdPtr;

#ifdef XOTCLOBJ_TRACE
  fprintf(stderr,"UpdateStringOfXOTclObject %p refCount %d\n",
	  objPtr,objPtr->refCount);
  fprintf(stderr,"    teardown %p id %p destroyCalled %d\n",
	  obj->teardown, obj->id, (obj->flags & XOTCL_DESTROY_CALLED));
#endif

  /* Here we use GetCommandName, because it doesnt need
     Interp*, but Tcl_GetCommandFullName(in,obj->id,ObjName); does*/
  if (obj && !(obj->flags & XOTCL_DESTROY_CALLED)) {
    Tcl_DString ds, *dsp = &ds;
    unsigned l;
    DSTRING_INIT(dsp);
    cmdPtr = (Command *) obj->id;
    nsFullName = cmdPtr->nsPtr->fullName;
    if (!(*nsFullName==':' && *(nsFullName+1)==':' &&
	  *(nsFullName+2)=='\0')) {
      Tcl_DStringAppend(dsp, nsFullName, -1);
    }
    Tcl_DStringAppend(dsp, "::", 2);
    Tcl_DStringAppend(dsp, Tcl_GetCommandName(NULL, obj->id), -1);

    l = (unsigned) Tcl_DStringLength(dsp)+1;
    objPtr->bytes = (char*) ckalloc(l);
    memcpy(objPtr->bytes, Tcl_DStringValue(dsp), l);
    objPtr->length = Tcl_DStringLength(dsp);
    DSTRING_FREE(dsp);
  } else if (obj) {
    /*fprintf(stderr,"try to read string of deleted command\n");*/
    FreeXOTclObjectInternalRep(objPtr);
    objPtr->bytes = NULL;
    objPtr->length = 0;
  } else {
    objPtr->bytes = NULL;
    objPtr->length = 0;
  }
  /*
    fprintf(stderr, "+++UpdateStringOfXOTclObject bytes='%s',length=%d\n",
	    objPtr->bytes,objPtr->length);
  */
}

#ifdef NOTUSED
static Tcl_Obj *
NewXOTclObjectObj(register XOTclObject* obj) {
    register Tcl_Obj *objPtr = 0;
    XOTclNewObj(objPtr);
    objPtr->bytes = NULL;
    objPtr->internalRep.otherValuePtr = obj;
    objPtr->typePtr = &XOTclObjectType;
#ifdef XOTCLOBJ_TRACE
  fprintf(stderr,"NewXOTclObjectObj %p\n",objPtr);
#endif
  return objPtr;
}
#endif

static Tcl_Obj *
NewXOTclObjectObjName(register XOTclObject* obj, char* name, int l)
{
  register Tcl_Obj *objPtr = 0;
  
  XOTclNewObj(objPtr);
  
  objPtr->length = l;
  objPtr->bytes = ckalloc(l+1);
  memcpy(objPtr->bytes, name, l);
  *(objPtr->bytes+l) = 0;
  objPtr->internalRep.otherValuePtr = obj;
  objPtr->typePtr = &XOTclObjectType;

#ifdef XOTCLOBJ_TRACE
    fprintf(stderr,"NewXOTclObjectObjName tcl %p (%d) xotcl %p (%d) %s\n",
	    objPtr, objPtr->refCount, obj, obj->refCount, objPtr->bytes);
#endif
  XOTclObjectRefCountIncr(obj);

  return objPtr;
}

#ifdef NOTUSED
static void
SetXOTclObjectObj(objPtr, obj)
    register Tcl_Obj *objPtr;
    register XOTclObject* obj;
{
    register Tcl_ObjType *oldTypePtr = objPtr->typePtr;

    if (Tcl_IsShared(objPtr)) {
  	panic("Tcl_SetXOTclObjectObj called with shared object");
    }

    Tcl_InvalidateStringRep(objPtr);
    if ((oldTypePtr != NULL) && (oldTypePtr->freeIntRepProc != NULL)) {
	oldTypePtr->freeIntRepProc(objPtr);
    }

    objPtr->internalRep.otherValuePtr = obj;
    objPtr->typePtr = &XOTclObjectType;
}
#endif


static int
GetXOTclObjectFromObj(Tcl_Interp *interp, register Tcl_Obj *objPtr,
		      XOTclObject **obj)
{
  int result;
  register Tcl_ObjType *cmdType = objPtr->typePtr;
#ifdef KEEP_TCL_CMD_TYPE
  static Tcl_ObjType *tclCmdNameType = NULL;

  if (tclCmdNameType == NULL && cmdType && cmdType != &XOTclObjectType) {
    if (!strcmp(cmdType->name,"cmdName")) {
      tclCmdNameType = cmdType;
    }
  }
#endif

  /* 
   * Only really share the "::x" Tcl_Objs but not "x" because we so not have 
   * references upon object kills and then will get dangling 
   * internalRep references to killed XOTclObjects
   */
  if (cmdType == &XOTclObjectType) {
    if (obj) {
      XOTclObject* o = (XOTclObject*) objPtr->internalRep.otherValuePtr;
      int refetch = 0;
      if (o->flags & XOTCL_DESTROYED) {
	/*fprintf(stderr,"????? calling free by hand\n");*/
	FreeXOTclObjectInternalRep(objPtr);
	refetch=1;
	result = SetXOTclObjectFromAny(interp, objPtr);
	if (result == TCL_OK) {
	  o = (XOTclObject*) objPtr->internalRep.otherValuePtr;
	  assert(o && !(o->flags & XOTCL_DESTROYED));
	}
      } else {
	result = TCL_OK;
      }

      *obj = o;

#ifdef XOTCLOBJ_TRACE
      if (result == TCL_OK)
	fprintf(stderr,"GetXOTclObjectFromObj tcl %p (%d) xotcl %p (%d) r=%d %s\n",
		objPtr, objPtr->refCount, o, o->refCount, refetch, objPtr->bytes);
      else
	fprintf(stderr,"GetXOTclObjectFromObj tcl %p (%d) **** rc=%d r=%d %s\n",
		objPtr, objPtr->refCount, result,refetch, objPtr->bytes);
#endif
    } else {
      result = TCL_OK;
    }
#ifdef KEEP_TCL_CMD_TYPE
  } else if (cmdType == tclCmdNameType) {
    Command *cmdPtr = (Command*)Tcl_GetCommandFromObj(interp, objPtr);
    if (cmdPtr) {
      XOTclObject *o = GetObjectFromCmdPtr(cmdPtr);

      /*
      fprintf(stderr,"Yes '%s' %p\n",objPtr->bytes,o);
      fprintf(stderr,"cmdPtr->objProc %p == %p, proc=%p\n",
	      cmdPtr->objProc, XOTclObjDispatch,cmdPtr->proc
	      );
      */
      if (o) {
	if (obj) *obj = o;
	result = TCL_OK;
      } else {
	goto convert_to_xotcl_object;
      }
    } else goto convert_to_xotcl_object;
#endif
  } else {
#ifdef KEEP_TCL_CMD_TYPE
  convert_to_xotcl_object:
#endif
    result = SetXOTclObjectFromAny(interp, objPtr);
    if (result == TCL_OK) {
      if (obj) *obj = (XOTclObject*) objPtr->internalRep.otherValuePtr;
    }
  }
  return result;
}

static int
GetXOTclClassFromObj(Tcl_Interp *interp, register Tcl_Obj *objPtr,
		     XOTclClass **cl) {
  XOTclObject *obj;
  XOTclClass *cls;
  int result = GetXOTclObjectFromObj(interp, objPtr, &obj);
  if (result == TCL_OK) {
    cls = XOTclObjectToClass(obj);
    if (cls) {
      if (cl) *cl = cls;
    } else
      result = TCL_ERROR;
  }
  return result;
}

/*
 * precedence ordering functions
 */

enum colors { WHITE, GRAY, BLACK };

static int
TopoSort(XOTclClass* cl, XOTclClass* base, XOTclClasses* (*next)(XOTclClass*)) {
  XOTclClasses* sl = (*next)(cl);
  XOTclClasses* pl;

  /*
   * careful to reset the color of unreported classes to
   * white in case we unwind with error, and on final exit
   * reset color of reported classes to white
   */

  cl->color = GRAY;
  for (; sl != 0; sl = sl->next) {
    XOTclClass* sc = sl->cl;
    if (sc->color==GRAY) { cl->color = WHITE; return 0; }
    if (sc->color==WHITE && !TopoSort(sc, base, next)) {
      cl->color=WHITE;
      if (cl == base) {
        XOTclClasses* pc = cl->order;
        while (pc != 0) { pc->cl->color = WHITE; pc = pc->next; }
      }
      return 0;
    }
  }
  cl->color = BLACK;
  pl = NEW(XOTclClasses);
  pl->cl = cl;
  pl->next = base->order;
  base->order = pl;
  if (cl == base) {
    XOTclClasses* pc = cl->order;
    while (pc != 0) { pc->cl->color = WHITE; pc = pc->next; }
  }
  return 1;
}

extern void
XOTclRemoveClasses(XOTclClasses* sl) {
  XOTclClasses *n;
  for (; sl; sl=n) {
    n = sl->next;
    FREE(XOTclClasses,sl);
  }
}

extern void
XOTclAddClass(XOTclClasses** cList, XOTclClass* cl) {
  XOTclClasses *l = *cList, *new = NEW(XOTclClasses);
  new->cl = cl;
  new->next = NULL;
  if (l) {
    while (l->next) l = l->next;
    l->next = new;
  } else
    *cList = new;
}

static XOTclClasses* Super(XOTclClass* cl) { return cl->super; }
static XOTclClasses* Sub(XOTclClass* cl) { return cl->sub; }

static XOTclClasses*
TopoOrder(XOTclClass* cl, XOTclClasses* (*next)(XOTclClass*)) {
  if (TopoSort(cl, cl, next))
    return cl->order;
  XOTclRemoveClasses(cl->order);
  cl->order = 0;
  return 0;
}

static XOTCLINLINE XOTclClasses*
ComputeOrder(register XOTclClass *cl, 
	     XOTclClasses* (*direction)(XOTclClass*)) {
  if (cl->order)
    return cl->order;
  return (cl->order = TopoOrder(cl, direction));
}

extern XOTclClasses*
XOTclComputePrecedence(XOTclClass* cl) {
  return ComputeOrder(cl, Super);
}

extern XOTclClasses*
XOTclComputeDependents(XOTclClass* cl) {
  return ComputeOrder(cl, Sub);
}


static void
FlushPrecedences(XOTclClass* cl) {
  XOTclClasses* pc;
  XOTclRemoveClasses(cl->order); cl->order = 0;
  pc = ComputeOrder(cl, Sub);

  /*
   * ordering doesn't matter here - we're just using toposort
   * to find all lower classes so we can flush their caches
   */

  if (pc) pc = pc->next;
  while (pc != 0) {
    XOTclRemoveClasses(pc->cl->order); pc->cl->order = 0;
    pc = pc->next;
  }
  XOTclRemoveClasses(cl->order); cl->order = 0;
}

static void
AddInstance(XOTclObject* obj, XOTclClass* cl) {
  obj->cl = cl;
  if (cl != 0) {
    int nw;
    (void) Tcl_CreateHashEntry(&cl->instances, (char*)obj, &nw);
  }
}

static int
RemoveInstance(XOTclObject* obj, XOTclClass* cl) {
  if (cl) {
    Tcl_HashEntry *hPtr = Tcl_FindHashEntry(&cl->instances, (char*)obj);
    if (hPtr) {
	Tcl_DeleteHashEntry(hPtr);
	return 1;
    }
  }
  return 0;
}

/*
 * superclass/subclass list maintenance
 */

static void
AS(XOTclClass* cl, XOTclClass* s, XOTclClasses** sl) {
  XOTclClasses* l = *sl;
  while (l &&  l->cl != s) l = l->next;
  if (!l) {
    XOTclClasses* sc = NEW(XOTclClasses);
    sc->cl = s; 
    sc->next = *sl; 
    *sl = sc;
  }
}

static void
AddSuper(XOTclClass* cl, XOTclClass* super) {
  if (cl && super) {
    /*
     * keep corresponding sub in step with super
     */
    AS(cl, super, &cl->super);
    AS(super, cl, &super->sub);
  }
}

static int
RemoveSuper1(XOTclClass* cl, XOTclClass* s, XOTclClasses** sl) {
  XOTclClasses* l = *sl;
  if (!l) return 0;
  if (l->cl == s) {
    *sl = l->next;
    FREE(XOTclClasses,l);
    return 1;
  }
  while (l->next && l->next->cl != s) l = l->next;
  if (l->next) {
    XOTclClasses* n = l->next->next;
    FREE(XOTclClasses,l->next);
    l->next = n;
    return 1;
  }
  return 0;
}

static int
RemoveSuper(XOTclClass* cl, XOTclClass* super) {
  /*
   * keep corresponding sub in step with super
   */
  int sp = RemoveSuper1(cl, super, &cl->super);
  int sb = RemoveSuper1(super, cl, &super->sub);

  return sp && sb;
}

/*
 * internal type checking
 */

extern  XOTclClass*
XOTclIsClass(Tcl_Interp* in, ClientData cd) {
  XOTclObject* obj = (XOTclObject*)cd;
  if (obj && XOTclObjectIsClass(obj))
    return (XOTclClass*) cd;
  return 0;
}

/*
 * methods lookup
 */

static XOTCLINLINE Command*
FindMethod (char* methodName, Tcl_Namespace* nsPtr) {
  Namespace *nsp = (Namespace *)nsPtr;
  Tcl_HashEntry* entryPtr;
  assert(nsp);

  if ((entryPtr = Tcl_FindHashEntry(&nsp->cmdTable, methodName))) {
    Command *cp = (Command *) Tcl_GetHashValue(entryPtr);
    if (cp->objProc != XOTclObjDispatch)
      return cp;
  }
  return NULL;
}

static XOTclClass*
SearchCMethod(XOTclClass* cl,
	      XOTclClasses* pl,
	      char* nm,
	      Command **procInfo) {
  assert(cl == NULL || pl == NULL);
  if (cl) {
    if (!cl->order) {
      cl->order = ComputeOrder(cl, Super);
    }
    pl = cl->order;
  }

  /* Search the class hierarchy */
  while (pl) {
    Command *pi = FindMethod(nm, pl->cl->nsPtr);
      
    if (pi) {
       
      *procInfo = pi;
      return pl->cl;
    }
    pl = pl->next;
  }
  *procInfo = NULL;
  return NULL;
}

static int
callDestroyMethod(ClientData cd, Tcl_Interp *in,
		  XOTclObject* obj, int flags) {
  int result;

  /* don't call destroy after exit handler started physical
     destruction */
  if (RUNTIME_STATE(in)->exitHandlerDestroyRound ==
      XOTCL_EXITHANDLER_ON_PHYSICAL_DESTROY)
    return TCL_OK;

  /* we don't call destroy, if we're in the exit handler
     during destruction of Object and Class */
  if (!RUNTIME_STATE(in)->callDestroy) {
    obj->flags |= XOTCL_DESTROY_CALLED;

    /* return TCL_ERROR so that clients know we haven't deleted the
       associated command yet */
    return TCL_ERROR;
  }
  /*fprintf(stderr, "+++ calldestroy flags=%d\n",flags);*/
  if (obj->flags & XOTCL_DESTROY_CALLED)
    return TCL_OK;

#if !defined(NDEBUG)
  {char* cmdName = ObjStr(obj->cmdName);
  assert(cmdName != NULL);
  assert(((Command *)Tcl_FindCommand(in, cmdName, NULL, 0)) != NULL);
  }
#endif


#ifdef OBJDELETION_TRACE
  fprintf(stderr, "   command found\n");
  PRINTOBJ("callDestroy", obj);
#endif
  result = callMethod(cd, in, global_objects[DESTROY], 2, 0, flags);
  if (result != TCL_OK) {
    static char cmd[] =
	"puts stderr \"[self]: Error in instproc destroy\n\
	 $::errorCode $::errorInfo\"";
    Tcl_VarEval(in, cmd, 0);
    if (++RUNTIME_STATE(in)->errorCount > 20)
      panic("too many destroy errors occured. Endless loop?", NULL);
  } else {
    if (RUNTIME_STATE(in)->errorCount > 0)
      RUNTIME_STATE(in)->errorCount--;
  }

#ifdef OBJDELETION_TRACE
  fprintf(stderr, "callDestroyMethod for %p exit\n", obj);
#endif
  return result;
}

/*
 * conditional memory allocations of optional storage
 */

static void
initObjectOpt(XOTclObject* obj) {
  obj->opt = NEW(XOTclObjectOpt);
  obj->opt->mixins = 0;
  obj->opt->filters = 0;
#ifdef XOTCL_METADATA
  XOTclMetaDataInit(obj);
#endif
  obj->opt->assertions = NULL;
  obj->opt->checkoptions = CHECK_NONE;
  obj->opt->clientData = 0;
}

static void
initClassOpt(XOTclClass* cl) {
  cl->opt = NEW(XOTclClassOpt);
  cl->opt->instfilters = 0;
  cl->opt->instmixins = 0;
  cl->opt->assertions = 0;
  cl->opt->parameterClass = 0;
#ifdef XOTCL_OBJECTDATA
  cl->opt->objectdata = 0;
#endif
  cl->opt->clientData = 0;
}

#define requireObjectOpt(obj) \
  if (!obj->opt) initObjectOpt(obj);
extern void
XOTclRequireObjectOpt(XOTclObject* obj) {
  requireObjectOpt(obj);
}

#define requireClassOpt(cl) \
  if (!cl->opt) initClassOpt(cl);
extern void
XOTclRequireClassOpt(XOTclClass* cl) {
  requireClassOpt(cl);
}

#define freeObjectOpt(obj) \
  if (obj->opt) {\
    FREE(XOTclObjectOpt,obj->opt); \
    obj->opt = 0; }

#define freeClassOpt(cl) \
  if (cl->opt) { \
    FREE(XOTclClassOpt, cl->opt); \
    cl->opt = 0; }


static Tcl_Namespace*
NSGetFreshNamespace(Tcl_Interp* in, ClientData cd, char *name);

static void
makeObjNamespace(Tcl_Interp* in, XOTclObject* obj) {
#ifdef NAMESPACE_TRACE
  fprintf(stderr, "+++ Make Namespace for %s\n", ObjStr(obj->cmdName));
#endif
  if (!obj->nsPtr) {
    Namespace* nsp;
    char* cmdName =  ObjStr(obj->cmdName);
    obj->nsPtr = NSGetFreshNamespace(in, (ClientData)obj, cmdName);
    if (!obj->nsPtr)
      panic("makeObjNamespace: Unable to make namespace",0);
    nsp = (Namespace*) obj->nsPtr;

    /*
     * Copy all obj variables to the newly created namespace
     */
    if (obj->varTable) {
      Tcl_HashSearch search;
      Tcl_HashEntry *hPtr, *newHPtr;
      register Var *varPtr;

      for (hPtr = Tcl_FirstHashEntry(obj->varTable, &search);  hPtr != NULL;
	   hPtr = Tcl_NextHashEntry(&search)) {
	int new = 0;
	char* name = Tcl_GetHashKey(obj->varTable, hPtr);
	varPtr = (Var *) Tcl_GetHashValue(hPtr);
	
	if (!name) {
	  panic("Can't copy: Hash Entry with no name", NULL);
	  continue;
	}

	newHPtr = Tcl_CreateHashEntry(&nsp->varTable, name, &new);
	if (new) {
	  /*
	   * put var into new hashtable entry
	   */
	  varPtr->flags |= VAR_IN_HASHTABLE;
	  Tcl_SetHashValue(newHPtr, varPtr);
	  varPtr->hPtr = newHPtr;
	  /*
	   * Mark the variable as a namespace variable
	
	  if (!(varPtr->flags & VAR_NAMESPACE_VAR)) {
	    varPtr->flags |= VAR_NAMESPACE_VAR;
	    varPtr->refCount++;
	  }
	  */
	  /*
	   * and correct the namespace information
	   */
	  varPtr->nsPtr = nsp;
	} else {
	  panic("Can't copy varTable variable to new namespace", NULL);
	}
      }
      MEM_COUNT_FREE("obj->varTable",obj->varTable);
      Tcl_DeleteHashTable(obj->varTable);
      ckfree((char *) obj->varTable);
      obj->varTable = 0;
    }
  }
}

#define requireObjNamespace(in,obj) \
  if (!obj->nsPtr) makeObjNamespace(in,obj);
extern void
XOTclRequireObjNamespace(Tcl_Interp* in, XOTclObject* obj) {
  requireObjNamespace(in,obj);
}


/*
 * Namespace related commands
 */

static int
NSDeleteCmd(Tcl_Interp* in, Tcl_Namespace* ns, char* name) {
  /* a simple deletion would delete a global command with
     the same name, if the command is not existing, so
     we use the CmdToken */
  Command* token = FindMethod (name, ns);
  /*fprintf(stderr, "NSDeleteCmd for %s\n",name);*/
  if (token) {
    return Tcl_DeleteCommandFromToken(in, (Tcl_Command) token);
  }
  return 0;
}

static void
CallStackDestroyObject(Tcl_Interp* in, XOTclObject* obj);
static void
PrimitiveCDestroy(ClientData cd);
static void
PrimitiveODestroy(ClientData cd);

static void
NSDeleteChildren(Tcl_Interp* in,  Tcl_Namespace* ns) {
  Namespace *nsPtr = (Namespace *) ns;
  if (&nsPtr->childTable) {
    Tcl_HashSearch hSrch;
    Tcl_HashEntry* hPtr = Tcl_FirstHashEntry(&nsPtr->cmdTable, &hSrch);
#ifdef OBJDELETION_TRACE
      fprintf(stderr, "NSDeleteChildren %s\n", nsPtr->fullName);
#endif
    for (; hPtr != 0; hPtr = Tcl_NextHashEntry(&hSrch)) {
      Command * cmdPtr = (Command *) Tcl_GetHashValue(hPtr);
      if (!cmdPtr->cmdEpoch) {
	char *oname = Tcl_GetHashKey(&nsPtr->cmdTable, hPtr);
	Tcl_DString name;
	XOTclObject* obj;
	/*
	  fprintf(stderr, " ... child %s\n", oname);
	*/
	ALLOC_NAME_NS(&name, nsPtr->fullName, oname);
	obj = GetObject(in, Tcl_DStringValue(&name));

	if (obj) {
	  /*
	    fprintf(stderr, " ... obj= %s\n", ObjStr(obj->cmdName));
	  */
	  /* in the exit handler physical destroy --> directly call destroy */
	  if (RUNTIME_STATE(in)->exitHandlerDestroyRound
	      == XOTCL_EXITHANDLER_ON_PHYSICAL_DESTROY) {
	    if (XOTclObjectIsClass(obj))
	      PrimitiveCDestroy((ClientData) obj);
	    else
	      PrimitiveODestroy((ClientData) obj);
	  } else {
	    if (obj->teardown != 0 && obj->id &&
		!(obj->flags & XOTCL_DESTROY_CALLED)) {
	      /*         Tcl_Command oid = obj->id;*/
	      if (callDestroyMethod((ClientData)obj, in, obj, 0) != TCL_OK) {
		/* destroy method failed, but we have to remove the command
		   anyway. */
		obj->flags |= XOTCL_DESTROY_CALLED;

		if (obj->teardown) {
		  CallStackDestroyObject(in, obj);
		}
		/*(void*) Tcl_DeleteCommandFromToken(in, oid);*/
	      }
	    }
	  }
	}
	DSTRING_FREE(&name);
      }
    }
  }
}

/*
 * ensure that a variable exists on object varTable or nsPtr->varTable,
 * if necessary create it. Return Var* if successful, otherwise 0
 */
static Var*
NSRequireVariableOnObj(Tcl_Interp* in, XOTclObject* obj, char* name,
		       int flgs) {

  VarFrameDecls;
  Var *varPtr, *arrayPtr;


  InitFakeFrame(in,obj,frame);
  VarFrameSwitchToObj(in, obj);
  varPtr = TclLookupVar(in, name, 0, flgs, "obj vwait",
			/*createPart1*/ 1, /*createPart2*/ 0, &arrayPtr);
  VarFrameRestore(in, obj);
  return varPtr;
}

/* delete a namespace recursively, but check whether the
   namespace is an object or not */
static void
NSDeleteNamespace(Tcl_Interp* in, Tcl_Namespace* ns) {
  Namespace* nsPtr = (Namespace *)ns;
  Tcl_HashSearch hSrch;
  Tcl_HashEntry* hPtr;

  NSDeleteChildren(in, ns);

  hPtr = &nsPtr->childTable ?
    Tcl_FirstHashEntry(&nsPtr->childTable, &hSrch) : 0;

  for (; hPtr != 0; hPtr = Tcl_NextHashEntry(&hSrch)) {
    Tcl_Namespace *child = ((Tcl_Namespace*)Tcl_GetHashValue(hPtr));
    NSDeleteNamespace(in, child);
  }
  /*	
  fprintf(stderr, "NSDeleteNamespace deleting %s\n", nsPtr->fullName);
  */
  MEM_COUNT_FREE("TclNamespace",ns);
  Tcl_DeleteNamespace(ns);
}

/*
 * delete all vars & procs in a namespace
 */
static void
NSCleanupNamespace(Tcl_Interp* in, Tcl_Namespace* ns) {
  Namespace* nsPtr = (Namespace *)ns;
  Interp *iPtr = (Interp *) in;
  Tcl_HashSearch hSrch;
  Tcl_HashEntry* hPtr;
  Command* procInfo;
  /*
   * Delete all variables and initialize var table again
   * (deletevars frees the vartable)
   */
  TclDeleteVars(iPtr, &nsPtr->varTable);
  Tcl_InitHashTable(&nsPtr->varTable, TCL_STRING_KEYS);

  /*
   * Delete all user-defined procs in the namespace
   */
  hPtr = &nsPtr->cmdTable ?
    Tcl_FirstHashEntry(&nsPtr->cmdTable, &hSrch) : 0;
  while (hPtr != NULL) {
    procInfo = (Command *) Tcl_GetHashValue(hPtr);
    /* objects should not be deleted here to preseve children deletion order*/
    if (!GetObjectFromCmdPtr(procInfo))
      Tcl_DeleteCommandFromToken(in, (Tcl_Command)procInfo);
    hPtr = Tcl_NextHashEntry(&hSrch);
  }
}

static void
NSNamespaceDeleteProc(ClientData clientData) {
  /* dummy for ns identification by pointer comparison */
}



void
XOTcl_DeleteNamespace(Tcl_Interp* in, Tcl_Namespace *nsPtr) {
  int activationCount = 0;
  Interp *iPtr = (Interp *) in;
  CallFrame *f = iPtr->framePtr;
  Namespace* nsp = (Namespace*) nsPtr;


  /*
  fprintf(stderr, "  ... correcting ActivationCount for %s was %d ",
	  nsPtr->fullName, nsp->activationCount);
  */
  while (f) {
    if (f->nsPtr == nsp)
      activationCount++;
    f = f->callerPtr;
  }

  nsp->activationCount = activationCount;
  /*
  fprintf(stderr, "to %d. \n",  nsp->activationCount);
  */
  MEM_COUNT_FREE("TclNamespace",nsPtr);
  Tcl_DeleteNamespace(nsPtr);
}

static Tcl_Namespace*
NSGetFreshNamespace(Tcl_Interp* in, ClientData cd, char *name) {
  Tcl_Namespace *namespacePtr =
    Tcl_FindNamespace(in, name, NULL, 0);
  if (namespacePtr) {
    /*  obj->nsPtr = namespacePtr;*/
    XOTcl_DeleteNamespace(in, namespacePtr);
  }
  namespacePtr = Tcl_CreateNamespace(in, name, cd,
			     (Tcl_NamespaceDeleteProc*)
			     NSNamespaceDeleteProc);
  MEM_COUNT_ALLOC("TclNamespace",namespacePtr);
  return namespacePtr;
}

/* cuts preceding ::xotcl::classes away */
#define NSCutXOTclClasses(n) (\
	(strncmp((n), "::xotcl::classes", 16) == 0) ?((n) + 16) : (n))

/*
 * check colons for illegal object/class names
 */
static XOTCLINLINE int
NSCheckColons(char* name, int l) {
  register char* n = name;
  if (*n == '\0') return 0; /* empty name */
  if (*(n+l-1) == ':')  return 0; /* name ends with : */
  if (*n == ':' && *(n+1) != ':') return 0; /* name begins with single : */
  for (; *n != '\0'; n++) {
    if (*n == ':' && *(n+1) == ':' && *(n+2) == ':')
      return 0;   /* more than 2 colons in series in a name */
  }
  return 1;
}

/*
 * check for parent namespace existance (used before commands are created)
 */

static XOTCLINLINE int
NSCheckForParent(Tcl_Interp* in, char* name, int l) {
  register char *n = name+l;
  int result = 1;

  /*search for last '::'*/
  while ((*n != ':' || *(n-1) != ':') && n-1 > name) {n--; }

  if (*n == ':' && n > name && *(n-1) == ':') {n--;}

  if ((n-name)>0) {
    Tcl_DString parentNSName, *dsp=&parentNSName;
    char *parentName;
    DSTRING_INIT(dsp);

    Tcl_DStringAppend(dsp, name, (n-name));
    parentName = Tcl_DStringValue(dsp);

    if (Tcl_FindNamespace(in, parentName, (Tcl_Namespace *) NULL, 
			  TCL_GLOBAL_ONLY) == 0) {
      XOTclObject* parentObj = XOTclGetObject(in, parentName);
      if (parentObj) {
	requireObjNamespace(in, parentObj);
      } else {
	result = 0;
      }
    }
    DSTRING_FREE(dsp);
  }
  return result;
}

/*
 * Find the "real" command belonging eg. to an XOTcl class or object.
 * Do not return cmds produced by Tcl_Import, but the "real" cmd
 * to which they point.
 */
static XOTCLINLINE Command*
NSFindCommand(Tcl_Interp* in, char* name) {
  Tcl_Command importedCmd;
  Command *cmdPtr;
  cmdPtr = (Command*) Tcl_FindCommand(in, name, NULL, 0);
  if (cmdPtr) {
    importedCmd = TclGetOriginalCommand((Tcl_Command) cmdPtr);
    if (importedCmd != 0)
      cmdPtr = (Command *) importedCmd;
  }
  return cmdPtr;
}


/*
 * Generic Tcl_Obj List
 */

static void
TclObjListDeleteList(XOTclTclObjList* list) {
  XOTclTclObjList* del;
  while (list) {
    del = list;
    list = list->next;
    DECR_REF_COUNT(del->content);
    FREE(XOTclTclObjList, del);
  }
}

static Tcl_Obj*
TclObjListNewElement(XOTclTclObjList** list, Tcl_Obj* ov) {
  XOTclTclObjList* elt = NEW(XOTclTclObjList);
  INCR_REF_COUNT(ov);
  elt->content = ov;
  elt->next = *list;
  *list = elt;
  return ov;
}

/*
 * Autonaming
 */

static Tcl_Obj*
AutonameIncr(Tcl_Interp* in, Tcl_Obj* name, XOTclObject* obj,
	     int instanceOpt, int resetOpt) {
  int valueLength, mustCopy=1, format = 0;
  char *valueString, *c;
  Tcl_Obj* valueObject, *result=NULL, *savedResult=NULL;
  VarFrameDecls;
  int flgs = TCL_LEAVE_ERR_MSG;

  VarFrameSwitchToObj(in, obj);
  if (obj->nsPtr)
    flgs |= TCL_NAMESPACE_ONLY;

#ifndef PRE83
  valueObject = TclIncrVar2(in, global_objects[AUTONAMES],
			    name, 1, flgs);
#else
  valueObject = TclIncrVar2(in, global_objects[AUTONAMES],
			    name, 1, /*part1NotParsed*/ 0);
#endif

  if (resetOpt) {
    if (valueObject != NULL) { /* we have an entry */
      Tcl_UnsetVar2(in, global_strings[AUTONAMES], ObjStr(name),
		    flgs);
    }
    result = global_objects[EMPTY];
    INCR_REF_COUNT(result);
  } else {
    if (valueObject == NULL) {
      valueObject = Tcl_ObjSetVar2(in, global_objects[AUTONAMES],
				   name, global_objects[ZERO], flgs);
    }
    if (instanceOpt) {
      char buffer[1], firstChar, *nextChars;
      nextChars = ObjStr(name);
      firstChar = *(nextChars ++);
      if (isupper((int)firstChar)) {
	buffer[0] = tolower((int)firstChar);
	result = Tcl_NewStringObj(buffer,1);
	INCR_REF_COUNT(result);
	Tcl_AppendToObj(result, nextChars, -1);
	mustCopy = 0;
      }
    }
    if (mustCopy) {
      result = Tcl_DuplicateObj(name);
      INCR_REF_COUNT(result);
      /*
      fprintf(stderr,"*** copy %p %s = %p\n",
	      name,ObjStr(name),result);
      */
    } 
    valueString = Tcl_GetStringFromObj(valueObject,&valueLength);

    /* if we find a % in the autoname -> We use Tcl_FormatObjCmd
       to let the autoname string be formated, like Tcl "format"
       command, with the value. E.g.:
	    autoname a%06d --> a000000, a000001, a000002, ...
    */
    for (c = ObjStr(result); *c != '\0'; c++) {
      if (*c == '%') {
	if (*(c+1) != '%') {
	  format = 1; 
	  break;
	} else {
	  /* when we find a %% we format and then append autoname, e.g.
	     autoname a%% --> a%1, a%2, ... */
	  c++;
	}
      }
    }
    if (format) {
      DEFINE_NEW_TCL_OBJECTS_ON_STACK(3, ov);
      savedResult = Tcl_GetObjResult(in);
      INCR_REF_COUNT(savedResult);
      ov[0] = global_objects[FORMAT];
      ov[1] = result;
      ov[2] = valueObject;
      if (Tcl_EvalObjv(in, 3, ov, 0) != TCL_OK) {
	VarFrameRestore(in, obj);
	DECR_REF_COUNT(savedResult);
	FREE_TCL_OBJECTS_ON_STACK(ov);
	return 0;
      }
      DECR_REF_COUNT(result);
      result = Tcl_DuplicateObj(Tcl_GetObjResult(in));
      INCR_REF_COUNT(result);
      Tcl_SetObjResult(in, savedResult);
      DECR_REF_COUNT(savedResult);
      FREE_TCL_OBJECTS_ON_STACK(ov);
    } else {
      /* append the value string, if not
	 formated or if only %% occurs */
      /*fprintf(stderr,"+++ append to obj %p %s : %s\n",
	result,ObjStr(result), valueString);*/
      Tcl_AppendToObj(result, valueString, valueLength);
      /*fprintf(stderr,"+++ append to obj done\n");*/
    }
  }
  
  VarFrameRestore(in, obj);
  assert(resetOpt && result->refCount>=1 || result->refCount == 1);
  return result;
}

/*
 * XOTcl CallStack
 */

static XOTCLINLINE int
CallStackPush(Tcl_Interp* in, XOTclObject* obj, XOTclClass* cl,
	      char* procName, Command* cmdPtr,
	      int objc, Tcl_Obj *CONST objv[],
	      int isFilterEntry, int isMixinEntry) {
  XOTclCallStack *cs;
  register XOTclCallStackContent *csc;

  cs = &RUNTIME_STATE(in)->cs;
  if (cs->top >= &cs->content[MAX_NESTING_DEPTH-1]) {
    Tcl_SetResult(in,
		  "too many nested calls to Tcl_EvalObj (infinite loop?)",
		  TCL_STATIC);
    return TCL_ERROR;
  }
  csc = ++cs->top;
  csc->self          = obj;

  /*fprintf(stderr, "PUSH=obj %s proc %s\n", ObjStr(obj->cmdName), procName);*/

  csc->cl            = cl;
  csc->cmdPtr        = cmdPtr;
  csc->destroyedCmd  = 0;
  csc->filterStackEntry = 0;
  if (isFilterEntry) {
    csc->cscType = XOTCL_CSC_TYPE_ACTIVE_FILTER;
    csc->filterStackEntry = obj->filterStack;
  } else if (isMixinEntry)
    csc->cscType = XOTCL_CSC_TYPE_MIXIN;
  else
    csc->cscType = XOTCL_CSC_TYPE_PLAIN;
  csc->frameCut      = 0;
  csc->objv          = objv;
  csc->objc          = objc;

  cs->topFrameCache = 0;
  MEM_COUNT_ALLOC("CallStack",NULL);
  return TCL_OK;
}

static XOTCLINLINE void
CallStackDoDestroy(Tcl_Interp* in, XOTclObject* obj) {
  Tcl_Command oid;

  oid = obj->id;
  PRINTOBJ("CallStackDoDestroy", obj);

  obj->id = 0;
  if (obj->teardown && oid) {
    Tcl_DeleteCommandFromToken(in, oid);
  }
}

static void
CallStackDestroyObject(Tcl_Interp* in, XOTclObject* obj) {
  XOTclCallStack *cs = &RUNTIME_STATE(in)->cs;
  XOTclCallStackContent *csc;
  int countSelfs = 0;

  Tcl_Command oid = obj->id;

  for (csc = &cs->content[1]; csc <= cs->top; csc++) {
    if (csc->self == obj) {
      csc->destroyedCmd = (Command *) oid;
      if (csc->destroyedCmd) {
	csc->destroyedCmd->refCount++;
	MEM_COUNT_ALLOC("command refCount",csc->destroyedCmd);
      }
      countSelfs++;
    }
  }
  /* if the object is not referenced at the callstack anymore
     we have to directly destroy it, because CallStackPop won't
     find the object destroy */
  if (countSelfs == 0) {
    CallStackDoDestroy(in, obj);
  } else
    /* to prevail the deletion order call delete children now
       -> children destructors are called before parent's
       destructor */
    if (obj->teardown && obj->nsPtr) {
      NSDeleteChildren(in, obj->nsPtr);
    }
}

static XOTCLINLINE int
CallStackIsDestroyed(Tcl_Interp* in) {
  return (RUNTIME_STATE(in)->cs.top->destroyedCmd == NULL) ? 0 : 1;
}

static XOTCLINLINE void
CallStackPop(Tcl_Interp* in) {
  XOTclCallStack *cs = &RUNTIME_STATE(in)->cs;
  XOTclCallStackContent *csc;
  XOTclCallStackContent *h = cs->top;

  assert(cs->top > cs->content);
  csc = cs->top;
  if (csc->destroyedCmd != 0) {
    int destroy = 1;
    TclCleanupCommand(csc->destroyedCmd);
    MEM_COUNT_FREE("command refCount", csc->destroyedCmd);
    /* do not physically destroy, when callstack still contains "self"
       entries of the object */
    while (--h > cs->content) {
      if (h->self == csc->self) {
	destroy = 0;
	break;
      }
    }
    if (destroy) {
      CallStackDoDestroy(in, csc->self);
    }
  }

  cs->topFrameCache = 0;
  cs->top--;
  MEM_COUNT_FREE("CallStack",NULL);
}

#define checkIsXOTclProcFrame(frame) \
  (frame->isProcCallFrame && frame->procPtr && \
  frame->procPtr->cmdPtr && !frame->procPtr->cmdPtr->cmdEpoch && \
  frame->procPtr->cmdPtr->nsPtr && \
  frame->procPtr->cmdPtr->nsPtr->deleteProc == NSNamespaceDeleteProc)


static XOTclCallStackContent*
CallStackGetFrame(Tcl_Interp* in) {
  Interp *iPtr = (Interp *) in;
  CallFrame *v = iPtr->varFramePtr, *f = iPtr->framePtr;
  XOTclCallStack *cs = &RUNTIME_STATE(in)->cs;
  XOTclCallStackContent *top = cs->top;

  if (cs->topFrameCache)
    return cs->topFrameCache;

  /* the xotcl call stack contains entries for filters, these are cut out
     of the tcl stack -> determine the entry in the xotcl stack that
     corresponds to the varFramePtr of the tcl stack */
  if (f && v) {
    int i, diff = f->level - v->level;
    XOTclCallStackContent *bot =  cs->content;
    CallFrame* d = iPtr->framePtr;

    for (i = diff; i > 0; i--, d = d->callerPtr) {
      if (!checkIsXOTclProcFrame(d)) {
	diff--;
      }
    }

    /* step through the diff, overstep inactive filters */
    while (diff > 0 && top > bot) {
      if (top->cscType != XOTCL_CSC_TYPE_INACTIVE_FILTER)
        diff--;
      do
	top--;
      while (top->cscType == XOTCL_CSC_TYPE_INACTIVE_FILTER);
    }

  }
  cs->topFrameCache = top;
  return top;
}

/*
 * cmd list handling
 */

/*
 * Cmd List Add/Remove ... returns the new element
 */
static XOTclCmdList*
CmdListAdd(XOTclCmdList** cList, Command* c, int noDuplicates) {
  XOTclCmdList *l = *cList, *new;

  /*
   * check for duplicates, if necessary
   */
  if (noDuplicates) {
    XOTclCmdList *h = l;
    while (h) {
      if (h->cmdPtr == c)
	return h;
      h = h->next;
    }
  }

  /*
   * ok, we have no duplicates -> append "new"
   * to the end of the list
   */
  new = NEW(XOTclCmdList);
  new->cmdPtr = c;
  new->cmdPtr->refCount++;
  MEM_COUNT_ALLOC("command refCount",new->cmdPtr);
  new->clientData = 0;
  new->next = NULL;

  if (l) {
    while (l->next)
      l = l->next;
    l->next = new;
  } else
    *cList = new;
  return new;
}

/*
 * Replace a command in a list entry without changing the client data
 * entry
 */
static void
CmdListReplaceCmd(XOTclCmdList* replace, Command* cmdPtr) {
  Command *del = replace->cmdPtr;
  replace->cmdPtr = cmdPtr;
  replace->cmdPtr->refCount++;
  MEM_COUNT_ALLOC("command refCount", replace->cmdPtr);
  TclCleanupCommand(del);
  MEM_COUNT_FREE("command refCount",replace->cmdPtr);
}


/*
 * physically delete an entry 'del'
 */
static void
CmdListDeleteCmdListEntry(XOTclCmdList* del,
			  XOTclFreeCmdListClientData* freeFct) {
  if (freeFct)
    (*freeFct)(del);
  MEM_COUNT_FREE("command refCount",del->cmdPtr);
  TclCleanupCommand(del->cmdPtr);
  FREE(XOTclCmdList, del);
}

/*
 * remove a command 'delCL' from a command list, but do not
 * free it ... returns the removed XOTclCmdList*
 */
static XOTclCmdList*
CmdListRemoveFromList(XOTclCmdList** cmdList, XOTclCmdList* delCL) {
  register XOTclCmdList* c = *cmdList, *del = 0;
  if (c == 0)
    return NULL;
  if (c == delCL) {
    *cmdList = c->next;
    del = c;
  } else {
    while (c->next && c->next != delCL) {
      c = c->next;
    }
    if (c->next == delCL) {
      del = delCL;
      c->next = delCL->next;
    }
  }
  return del;
}

/*
 * remove all command pointers from a list that have a bumped epoch
 */
static void
CmdListRemoveEpoched(XOTclCmdList** cmdList,
		     XOTclFreeCmdListClientData* freeFct) {
   XOTclCmdList *f = *cmdList, *del;
   while (f) {
     if (f->cmdPtr->cmdEpoch) {
       del = f;
       f = f->next;
       del = CmdListRemoveFromList(cmdList, del);
       CmdListDeleteCmdListEntry(del, freeFct);
     } else
       f = f->next;
   }
}

/*
 * delete all entries from a given namespace
 */
static void
CmdListRemoveNSFromList(XOTclCmdList** cmdList, Namespace* nsPtr,
			XOTclFreeCmdListClientData* freeFct) {
  XOTclCmdList* c, *del = 0;
  /*
  CmdListRemoveEpoched(cmdList, freeFct);
  */
  c = *cmdList;
  while (c && c->cmdPtr->nsPtr == nsPtr) {
    del = c;
    *cmdList = c->next;
    CmdListDeleteCmdListEntry(del, freeFct);
    c = *cmdList;
  }
  while (c) {
    if (c->cmdPtr->nsPtr == nsPtr) {
      del = c;
      c = *cmdList;
      while (c->next && c->next != del)
	c = c->next;
      if (c->next == del)
	c->next = del->next;
      CmdListDeleteCmdListEntry(del, freeFct);
    }
    c = c->next;
  }
}

/*
 * free the memory of a whole 'cmdList'
 */
static void
CmdListRemoveList(XOTclCmdList** cmdList,
		  XOTclFreeCmdListClientData* freeFct) {
  XOTclCmdList *del;
  while (*cmdList) {
    del = *cmdList;
    *cmdList = (*cmdList)->next;
    CmdListDeleteCmdListEntry(del, freeFct);
  }
}

/*
 * simple list search proc to search a list of cmds
 * for a command ptr
 */
static XOTCLINLINE XOTclCmdList*
CmdListFindCmdInList(Command* cmdPtr, XOTclCmdList* l) {
  register XOTclCmdList* h = l;
  while (h != 0) {
    if (h->cmdPtr == cmdPtr)
      return h;
    h = h->next;
  }
  return 0;
}

/*
 * simple list search proc to search a list of cmds
 * for a simple Name
 */
static XOTCLINLINE XOTclCmdList*
CmdListFindNameInList(Tcl_Interp* in, char* name, XOTclCmdList* l) {
  register XOTclCmdList* h = l;
  while (h != 0) {
    CONST84 char* cmdName = Tcl_GetCommandName(in, (Tcl_Command)h->cmdPtr);
    if (cmdName[0] == name[0] && !strcmp(cmdName, name))
      return h;
    h = h->next;
  }
  return 0;
}

/*
 * Assertions
 */
static XOTclTclObjList*
AssertionNewList(Tcl_Interp* in, Tcl_Obj* aObj) {
  Tcl_Obj **ov; int oc;
  XOTclTclObjList *last=NULL;

  if (Tcl_ListObjGetElements(in, aObj, &oc, &ov) == TCL_OK) {
    if (oc > 0) {
      int i;
      for (i=oc-1; i>=0; i--) {
	TclObjListNewElement(&last, ov[i]);
      }
    }
  }
  return last;
}

static Tcl_Obj*
AssertionList(Tcl_Interp* in, XOTclTclObjList* alist) {
  Tcl_Obj* newAssStr = Tcl_NewStringObj("",0);
  for (; alist!=NULL; alist = alist->next) {
    Tcl_AppendStringsToObj(newAssStr, "{",  ObjStr(alist->content),
			   "}", (char *) NULL);
    if (alist->next != NULL)
      Tcl_AppendStringsToObj(newAssStr, " ", (char*) NULL);
  }
  return newAssStr;
}

/* append a string of pre and post assertions to a proc
   or instproc body */
static void
AssertionAppendPrePost(Tcl_Interp* in, Tcl_DString *dsPtr,
		       XOTclProcAssertion* procs) {
  if (procs) {
    Tcl_Obj* preAss = AssertionList(in, procs->pre);
    Tcl_Obj* postAss = AssertionList(in, procs->post);
    INCR_REF_COUNT(preAss); INCR_REF_COUNT(postAss);
    Tcl_DStringAppendElement(dsPtr, ObjStr(preAss));
    Tcl_DStringAppendElement(dsPtr, ObjStr(postAss));
    DECR_REF_COUNT(preAss); DECR_REF_COUNT(postAss);
  }
}

static int
AssertionListCheckOption(Tcl_Interp* in, XOTclObject* obj) {
  if (!obj->opt)
    return TCL_OK;
  if (obj->opt->checkoptions & CHECK_OBJINVAR)
    Tcl_AppendElement(in, "invar");
  if (obj->opt->checkoptions & CHECK_CLINVAR)
    Tcl_AppendElement(in, "instinvar");
  if (obj->opt->checkoptions & CHECK_PRE)
    Tcl_AppendElement(in, "pre");
  if (obj->opt->checkoptions & CHECK_POST)
    Tcl_AppendElement(in, "post");
  return TCL_OK;
}

static XOTclProcAssertion*
AssertionFindProcs(XOTclAssertionStore* aStore, char* name) {
  Tcl_HashEntry *hPtr;
  if (aStore == NULL) return NULL;
  hPtr = Tcl_FindHashEntry(&aStore->procs, name);
  if (hPtr == NULL) return NULL;
  return (XOTclProcAssertion*) Tcl_GetHashValue(hPtr);
}

static void
AssertionRemoveProc(XOTclAssertionStore* aStore, char* name) {
  Tcl_HashEntry *hPtr;
  if (aStore) {
    hPtr = Tcl_FindHashEntry(&aStore->procs, name);
    if (hPtr) {
      XOTclProcAssertion* procAss =
	(XOTclProcAssertion*) Tcl_GetHashValue(hPtr);
      TclObjListDeleteList(procAss->pre);
      TclObjListDeleteList(procAss->post);
      FREE(XOTclProcAssertion, procAss);
      Tcl_DeleteHashEntry(hPtr);
    }
  }
}


static void
AssertionAddProc(Tcl_Interp* in, char* name, XOTclAssertionStore* aStore,
		 Tcl_Obj* pre, Tcl_Obj* post) {
  int nw = 0; 
  Tcl_HashEntry* hPtr = NULL;
  XOTclProcAssertion *procs = NEW(XOTclProcAssertion);

  AssertionRemoveProc(aStore, name);
  procs->pre = AssertionNewList(in, pre);
  procs->post = AssertionNewList(in, post);
  hPtr = Tcl_CreateHashEntry(&aStore->procs, name, &nw);
  if (nw) Tcl_SetHashValue(hPtr, (ClientData)procs);
}

static XOTclAssertionStore*
AssertionCreateStore() {
  XOTclAssertionStore* aStore = NEW(XOTclAssertionStore);
  aStore->invariants = NULL;
  Tcl_InitHashTable(&aStore->procs, TCL_STRING_KEYS);
  MEM_COUNT_ALLOC("Tcl_InitHashTable",&aStore->procs);
  return aStore;
}

static void
AssertionRemoveStore(XOTclAssertionStore* aStore) {
  Tcl_HashSearch hSrch;
  Tcl_HashEntry* hPtr;

  if (aStore) {
    hPtr = Tcl_FirstHashEntry(&aStore->procs, &hSrch);
    while (hPtr) {
      /*
       * AssertionRemoveProc calls Tcl_DeleteHashEntry(hPtr), thus
       * we get the FirstHashEntry afterwards again to proceed
       */
      AssertionRemoveProc(aStore, Tcl_GetHashKey(&aStore->procs, hPtr));
      hPtr = Tcl_FirstHashEntry(&aStore->procs, &hSrch);
    }
    Tcl_DeleteHashTable(&aStore->procs);
    MEM_COUNT_FREE("Tcl_InitHashTable",&aStore->procs);
    TclObjListDeleteList(aStore->invariants);
    FREE(XOTclAssertionStore, aStore);
  }
}

/*
 * check a given condition in the current callframe's scope
 * it's the responsiblity of the caller to push the intended callframe
 */
static int
checkConditionInScope(Tcl_Interp* in, Tcl_Obj *condition) {
  int result, success;
  Tcl_Obj *ov[2];
  ov [1] = condition;
  INCR_REF_COUNT(condition);
  result = XOTcl_ExprObjCmd(NULL, in, 2, ov);
  DECR_REF_COUNT(condition);
  if (result == TCL_OK) {
    result = Tcl_GetIntFromObj(in,Tcl_GetObjResult(in),&success);
    if (result == TCL_OK && success == 0)
      result = XOTCL_CHECK_FAILED;
  }
  return result;
}

static int
AssertionCheckList(Tcl_Interp* in, XOTclObject* obj,
		   XOTclTclObjList* alist, char* methodName) {
  XOTclTclObjList* checkFailed = NULL;
  Tcl_Obj* savedObjResult = Tcl_GetObjResult(in);
  int savedCheckoptions, acResult = TCL_OK;

  /*
   * no obj->opt -> checkoption == CHECK_NONE
   */
  if (!obj->opt)
    return TCL_OK;

  /* we do not check assertion modifying methods, otherwise
     we can not react in catch on a runtime assertion check failure */
  if (isCheckString(methodName) || isInfoString(methodName) ||
      isInvarString(methodName) || isInstinvarString(methodName) ||
      isProcString(methodName) || isInstprocString(methodName))
    return TCL_OK;

  INCR_REF_COUNT(savedObjResult);

  Tcl_ResetResult(in);

  while (alist != NULL) {
    /* Eval instead of IfObjCmd => the substitutions in the
       conditions will be done by Tcl */
    char* assStr = ObjStr(alist->content), *c = assStr;
    int comment = 0;

    while (c != 0 && *c != '\0') {
      if (*c == '#') {
	comment = 1; break;
      }
      c++;
    }
    if (!comment) {
      ObjFrameDecls;
      XOTclPushObjFrame(in,obj);
      CallStackPush(in, obj, 0, 0, 0, 0, 0, 0, 0);

      /* don't check assertion during assertion check */
      savedCheckoptions = obj->opt->checkoptions;
      obj->opt->checkoptions = CHECK_NONE;

      /*fprintf(stderr, "Checking Assertion %s ", assStr);*/

      /*
       * now check the assertion in the pushed callframe's scope
       */
      acResult = checkConditionInScope(in, alist->content);
      if (acResult != TCL_OK)
	checkFailed = alist;

      obj->opt->checkoptions = savedCheckoptions;
      /*
	if (checkFailed)
	fprintf(stderr, "...failed\n");
	 else
	 fprintf(stderr, "...ok\n");
      */
      CallStackPop(in);
      XOTclPopObjFrame(in,obj);
      /*
      Tcl_PopCallFrame(in);
      */
    }
    if (checkFailed)
      break;
    alist = alist->next;
  }

  if (checkFailed != NULL) {
    DECR_REF_COUNT(savedObjResult);
    if (acResult == TCL_ERROR) {
      Tcl_Obj* sr = Tcl_GetObjResult(in);
      INCR_REF_COUNT(sr);	
      XOTclVarErrMsg(in, "Error in Assertion: {",
			  ObjStr(checkFailed->content), "} in proc '",
			  GetSelfProc(in), "'\n\n", ObjStr(sr),(char*)NULL);
      DECR_REF_COUNT(sr);
      return TCL_ERROR;
    }
    return XOTclVarErrMsg(in,
			  "Assertion failed check: {",
			  ObjStr(checkFailed->content), "} in proc '",
			  GetSelfProc(in), "'",(char*)NULL);
  }

  Tcl_SetObjResult(in, savedObjResult);
  DECR_REF_COUNT(savedObjResult);
  return TCL_OK;
}

static int
AssertionCheckInvars(Tcl_Interp* in, XOTclObject* obj, char* method) {
  int result = TCL_OK;
  int checkoptions = CHECK_NONE;

  if (obj->opt) {
    checkoptions = obj->opt->checkoptions;
    if (checkoptions & CHECK_OBJINVAR && obj->opt->assertions)
      result = AssertionCheckList(in, obj, obj->opt->assertions->invariants,
				  method);
  }

  if (result != TCL_ERROR && checkoptions & CHECK_CLINVAR) {
    XOTclClasses* clPtr;
    if (!obj->cl->order)
      obj->cl->order = ComputeOrder(obj->cl, Super);
    clPtr = obj->cl->order;
    while (clPtr != 0 && result != TCL_ERROR) {
      XOTclAssertionStore* aStore = (clPtr->cl->opt) ?
	clPtr->cl->opt->assertions : 0;
      if (aStore)
	result = AssertionCheckList(in, obj, aStore->invariants, method);

      clPtr = clPtr->next;
    }
  }
  return result;
}

static int
AssertionCheck(Tcl_Interp* in, XOTclObject* obj, XOTclClass* cl,
	       char* method, int checkOption) {
  XOTclProcAssertion* procs;
  int result = TCL_OK;
  XOTclAssertionStore* aStore;
  int checkoptions = CHECK_NONE;

  if (cl)
    aStore = cl->opt ? cl->opt->assertions : 0;
  else
    aStore = obj->opt ? obj->opt->assertions : 0;

  if (obj->opt)
    checkoptions = obj->opt->checkoptions;

  if (checkOption & checkoptions) {
    procs = AssertionFindProcs(aStore, method);
    if (procs) {
      switch (checkOption) {
      case CHECK_PRE:
	result = AssertionCheckList(in, obj, procs->pre, method);
	break;
      case CHECK_POST:
	result = AssertionCheckList(in, obj, procs->post, method);
	break;
      }
    }
    if (result != TCL_ERROR)
      result = AssertionCheckInvars(in, obj, method);
  }
  return result;
}

/*
 * Per-Object-Mixins
 */

/*
 * push a mixin stack information on this object
 */
static void
MixinStackPush(XOTclObject* obj) {
  XOTclMixinStack* h = NEW(XOTclMixinStack);
  h->currentCmdPtr = 0;
  h->mixinChainOn = 0;
  h->next = obj->mixinStack;
  obj->mixinStack = h;
}

/*
 * pop a mixin stack information on this object
 */
static void
MixinStackPop(XOTclObject* obj) {
  XOTclMixinStack *h = obj->mixinStack;
  obj->mixinStack = h->next;
  FREE(XOTclMixinStack,h);
}

/*
 * Appends XOTclClasses* containing the mixin classes and their
 * superclasses to 'mixinClasses' list from a given mixinList
 */
static void
MixinComputeOrderFullList(Tcl_Interp* in,
			  XOTclCmdList** mixinList,
			  XOTclClasses** mixinClasses) {
  XOTclCmdList *m;
  XOTclClasses *pl;

  CmdListRemoveEpoched(mixinList, 0);
  m = *mixinList;

  while (m) {
    XOTclClass* mCl = GetClassFromCmdPtr(m->cmdPtr);
    if (mCl) {
      if (!mCl->order)
	mCl->order = ComputeOrder(mCl, Super);
      pl = mCl->order;
      while (pl) {
	/*fprintf(stderr, " %s, ", ObjStr(pl->cl->object.cmdName));*/
	XOTclAddClass(mixinClasses, pl->cl);
	pl = pl->next;
      }
    }
    m = m->next;
  }
}

static void
MixinResetOrder(XOTclObject* obj) {
  CmdListRemoveList(&obj->mixinOrder,0);
  obj->mixinOrder = 0;
}

/*
 * Computes a linearized order of per-object and per-class mixins. Then
 * duplicates in the full list and with the class inheritance list of
 * 'obj' are eliminated.
 * The precendence rule is that the last occurence makes it into the
 * final list.
 */
static void
MixinComputeOrder(Tcl_Interp* in, XOTclObject* obj) {
  XOTclClasses *mixinClasses = 0, *nextCl, *pl, *checker;
  MixinResetOrder(obj);
  /*fprintf(stderr, "Mixin Order:\n First List: ");*/

  /* append per-obj mixins */
  if (obj->opt)
    MixinComputeOrderFullList(in, &obj->opt->mixins, &mixinClasses);

  /* append per-class mixins */
  if (!obj->cl->order)
    obj->cl->order = ComputeOrder(obj->cl, Super);
  pl = obj->cl->order;
  while (pl) {
    if (pl->cl->opt && pl->cl->opt->instmixins != 0)
      MixinComputeOrderFullList(in, &pl->cl->opt->instmixins, &mixinClasses);
    pl = pl->next;
  }

  /* use no duplicates & no classes of the precedence order
     on the resulting list */
  while (mixinClasses) {
    checker = nextCl = mixinClasses->next;
    while (checker) {
      if (checker->cl == mixinClasses->cl) break;
      checker = checker->next;
    }
    /* if mixinClasses has no duplicate on mixinList ->
       check obj->cl hierachy */
    if (checker == 0) {
      if (!obj->cl->order)
	obj->cl->order = ComputeOrder(obj->cl, Super);
      checker = obj->cl->order;
      while (checker) {
	if (checker->cl == mixinClasses->cl) break;
	checker = checker->next;
      }
    }
    /* if class is also not found on precedence order ->
       add it to mixinOrder list (otherwise free the memory) */
    if (checker == 0) {
      CmdListAdd(&obj->mixinOrder, (Command*) mixinClasses->cl->object.id,
		 /*noDuplicates*/ 0);
      /*
      fprintf(stderr, "  Adding %s,\n", ObjStr(mixinClasses->cl->object.cmdName));
      */
    }
    FREE(XOTclClasses, mixinClasses);

    mixinClasses = nextCl;
  }
}

/*
 * add a mixin class to 'mixinList' by appending it
 */
static int
MixinAdd(Tcl_Interp* in, XOTclCmdList** mixinList, Tcl_Obj* ov) {
  XOTclClass* mixin;

  if (GetXOTclClassFromObj(in, ov, &mixin) != TCL_OK)
    return XOTclErrBadVal(in, "a class as mixin", ObjStr(ov));

  CmdListAdd(mixinList, (Command *) mixin->object.id,
	     /*noDuplicates*/ 1);

  return TCL_OK;
}

/*
 * if the class hierarchy or class mixins have changed ->
 * invalidate mixin entries in all dependent instances
 */
static void
MixinInvalidateObjOrders(XOTclClass* cl) {
  XOTclClasses *saved = cl->order, *clPtr;
  cl->order = 0;
  clPtr = ComputeOrder(cl, Sub);

  while (clPtr != 0) {
    Tcl_HashSearch hSrch;
    Tcl_HashEntry* hPtr = &clPtr->cl->instances ?
      Tcl_FirstHashEntry(&clPtr->cl->instances, &hSrch) : 0;

    for (; hPtr != 0; hPtr = Tcl_NextHashEntry(&hSrch)) {
      XOTclObject* obj = (XOTclObject*)
	Tcl_GetHashKey(&clPtr->cl->instances, hPtr);
      MixinResetOrder(obj);
      obj->flags &= ~XOTCL_MIXIN_ORDER_VALID;
    }
    clPtr = clPtr->next;
  }

  XOTclRemoveClasses(cl->order);
  cl->order = saved;
}

/*
 * remove a mixin from the mixin stack of object 'obj'
 */
static void
MixinRemoveFromMixinStack(XOTclCmdList* del, XOTclObject* obj) {
  /* if the current mixin is to be deleted, we make its predecessor
     current mixin */
  if (obj->mixinStack->currentCmdPtr == del->cmdPtr) {
    if (obj->mixinOrder && del->cmdPtr != obj->mixinOrder->cmdPtr) {
      XOTclCmdList* predecessor = obj->mixinOrder;
      while (predecessor->next  &&
	     predecessor->next->cmdPtr != del->cmdPtr)
	predecessor = predecessor->next;
      if (predecessor->next)
	obj->mixinStack->currentCmdPtr = predecessor->next->cmdPtr;
    } else {
      obj->mixinStack->currentCmdPtr = 0;
    }
  }
}

/*
 * remove mixin with a command ptr from the mixin list and delete it
 * if 'obj' is specified, we also remove it from the objects mixin stack
 */
static void
MixinRemoveFromCmdPtr(XOTclCmdList** mixinList, Command* c, XOTclObject* obj) {
  XOTclCmdList* m = *mixinList, *del;
  while (m) {
    if (m->cmdPtr == c) {
      del = CmdListRemoveFromList(mixinList, m);
      if (del) {
	if (obj)
	  MixinRemoveFromMixinStack(del, obj);
	/* actually delete the mixin entry */
	CmdListDeleteCmdListEntry(del, 0);
      }
      break;
    }
    m = m->next;
  }
}

/*
 * remove mixin from cmd ptr with obj ... search per-object and
 * per-class mixins for the mixin command 'c'
 */
static void
MixinRemoveOnObjFromCmdPtr(XOTclObject* obj, Command* c) {
  XOTclClasses *pl;
  obj->flags &= ~XOTCL_MIXIN_ORDER_VALID;
  if (obj->opt)
    MixinRemoveFromCmdPtr(&obj->opt->mixins, c, obj);
  if (!obj->cl->order)
    obj->cl->order = ComputeOrder(obj->cl, Super);
  pl = obj->cl->order;
  while (pl) {
    if (pl->cl->opt) {
      MixinRemoveFromCmdPtr(&pl->cl->opt->instmixins, c, obj);
      MixinInvalidateObjOrders(pl->cl);
    }
    pl = pl->next;
  }
}

/*
 * the mixin order is either
 *   DEFINED (there are mixins on the instance),
 *   NONE    (there are no mixins for the instance),
 *   or INVALID (a class re-strucuturing has occured, thus it is not clear
 *               whether mixins are defined or not).
 * If it is INVALID MixinComputeDefined can be used to compute the order
 * and set the instance to DEFINE or NONE
 */
static void
MixinComputeDefined(Tcl_Interp* in, XOTclObject* obj) {
  /*MixinResetOrder(obj);*/
  MixinComputeOrder(in, obj);
  obj->flags |= XOTCL_MIXIN_ORDER_VALID;
  if (obj->mixinOrder)
    obj->flags |= XOTCL_MIXIN_ORDER_DEFINED;
  else
    obj->flags &= ~XOTCL_MIXIN_ORDER_DEFINED;
}

/*
 * walk through the mixin order until the current mixin is reached.
 * then use the next mixin as current mixin.
 *
 * precondition: obj->mixinStack is not NULL
 */
static void
MixinSeekCurrent(Tcl_Interp* in, XOTclObject* obj,
		 XOTclCmdList** cmdList) {
  Command* currentCmdPtr = obj->mixinStack->currentCmdPtr;

  /* ensure that the mixin order is not invalid, otherwise compute order */
  if (!(obj->flags & XOTCL_MIXIN_ORDER_VALID))
    MixinComputeDefined(in, obj);

  *cmdList = obj->mixinOrder;

  /* go forward to current class */
  while (*cmdList && currentCmdPtr) {
    if ((*cmdList)->cmdPtr == currentCmdPtr) currentCmdPtr = 0;
    *cmdList = (*cmdList)->next;
  }
}

/*
 * before we can perform a mixin dispatch, MixinSearchProc seeks the
 * current mixin and the relevant calling information
 */
static Command*
MixinSearchProc(Tcl_Interp* in, XOTclObject* obj, char* methodName,
		XOTclClass** cl, Tcl_ObjCmdProc** proc, ClientData* cp,
		Command** currentCmdPtr) {
  Command* procInfo = NULL;
  XOTclCmdList *cmdList;
  XOTclClass* cls;

  *currentCmdPtr = 0;
  *proc = 0;
  *cp = 0;

  if (!obj || !obj->mixinStack)
    return 0;

  MixinSeekCurrent(in, obj, &cmdList);
  while (cmdList) {
    if(cmdList->cmdPtr->cmdEpoch) {
      MixinRemoveOnObjFromCmdPtr(obj, cmdList->cmdPtr);
      /*
       * invalidate the order & seek start class
       */
      MixinResetOrder(obj);
      MixinSeekCurrent(in, obj, &cmdList);
    } else {
      cls = GetClassFromCmdPtr(cmdList->cmdPtr);
      if (cls) {
	procInfo = FindMethod(methodName, cls->nsPtr);
	/*
	 * on success: compute mixin call data
	 */
	if (procInfo) {
	  *cl = cls;
	  *proc = procInfo->objProc;
	  *cp   = procInfo->objClientData;
	  *currentCmdPtr = cmdList->cmdPtr;
	  break;
	} else {
	  cmdList = cmdList->next;
	}
      }
    }
  }

  return procInfo;
}

/*
 * info option for mixins and instmixins
 */
static int
MixinInfo(Tcl_Interp* in, XOTclCmdList* m, char* pattern) {
  Tcl_Obj *list = Tcl_NewListObj(0, NULL);
  XOTclClass* mixinClass;
  while (m) {
    mixinClass = GetClassFromCmdPtr(m->cmdPtr);
    if (mixinClass && (!pattern ||
	Tcl_StringMatch(ObjStr(mixinClass->object.cmdName), pattern)))
	Tcl_ListObjAppendElement(in, list, mixinClass->object.cmdName);
    m = m->next;
  }
  Tcl_SetObjResult(in, list);
  return TCL_OK;
}

static Command*
MixinSearchMethodByName(Tcl_Interp* in, XOTclCmdList* mixinList, char* name) {
  Command* cmdPtr;
  while (mixinList) {
    XOTclClass* mcl = GetClass(in,
      (char*) Tcl_GetCommandName(in, (Tcl_Command)mixinList->cmdPtr));
    SearchCMethod(mcl, NULL, name, &cmdPtr);
    if (cmdPtr) {
      return cmdPtr;
    }
    mixinList = mixinList->next;
  }
  return 0;
}


/*
 *  Filter-Commands
 */

/*
 * The search method implements filter search order for filter
 * and instfilter: first a given name is interpreted as fully
 * qualified instproc name. If no instproc is found, a proc is
 * search with fully name. Otherwise the simple name is searched
 * on the heritage order: object (only for
 * per-object filters), class, meta-class
 */

static Command*
FilterSearch(Tcl_Interp* in, char* name,
	     XOTclObject* startingObj, XOTclClass* startingCl) {
  Command* cmdPtr = NULL;


#ifdef SEARCH_FULLY_QUALIFIED_NAMES
  /*
  while (*name == ':') name++;
  */
  /*
   * first check whether the given (full) name is an XOTcl instproc
   */
  Tcl_PushCallFrame(in, &frame,
		    (Tcl_Namespace*)RUNTIME_STATE(in)->XOTclClassesNS, 0);
  cmdPtr = (Command *) Tcl_FindCommand(in, name, 0, 0);
  if (cmdPtr && TclIsProc(cmdPtr) &&
      cmdPtr->nsPtr->deleteProc == NSNamespaceDeleteProc) {
    /* ok, it is an XOTcl proc or instproc */
    Tcl_PopCallFrame(in);
    return cmdPtr;
  }
  Tcl_PopCallFrame(in);

  /*
   * now check whether the given (full) name is an XOTcl proc
   */
  cmdPtr = (Command *) Tcl_FindCommand(in, name, 0, 0);
  if (cmdPtr && TclIsProc(cmdPtr) &&
      cmdPtr->nsPtr->deleteProc == NSNamespaceDeleteProc)
    /* ok, it is an XOTcl proc or instproc */
    return cmdPtr;

  /* it is not a fully qualified name -> search on heritage order */
#endif


  if (startingObj) {
    /*
     * the object-specific filter can also be defined on the object's
     * class, its hierarchy, or the respective instmixins; thus use the
     * object's class as start point for the class-specific search then ...
     */
    startingCl = startingObj->cl;

    /*
     * search for filters on object mixins
     */
    if (startingObj->opt && startingObj->opt->mixins) {
      cmdPtr = MixinSearchMethodByName(in, startingObj->opt->mixins, name);
      if (cmdPtr)
	return cmdPtr;
    }
  }

  /*
   * search for instfilters on instmixins
   */
  if (startingCl) {
    if (startingCl->opt && startingCl->opt->instmixins) {
      cmdPtr = MixinSearchMethodByName(in, startingCl->opt->instmixins, name);
      if (cmdPtr)
	return cmdPtr;
    }
  }

  /*
   * seach for object procs that are used as filters
   */
  if (startingObj && startingObj->nsPtr) {
    cmdPtr = FindMethod(name, startingObj->nsPtr);
    if (cmdPtr) {
      return cmdPtr;
    }
  }

  /*
   * ok, no filter on obj or mixins -> search class
   */
  if (startingCl) {
    SearchCMethod(startingCl, NULL, name, &cmdPtr);
    if (!cmdPtr)
      /*
       * If no filter is found yet -> search the meta-class
       */
      SearchCMethod(startingCl->object.cl, NULL, name, &cmdPtr);
  }
  return cmdPtr;
}

/*
 * Filter Guards
 */


/* check a filter guard, return 1 if ok */
static int
FilterGuardCheck(Tcl_Interp* in, ClientData filterGuards) {
  XOTclTclObjList* fr = (XOTclTclObjList*) filterGuards;
  int frc;
  XOTclCallStack *cs = &RUNTIME_STATE(in)->cs;


  while (fr) {
    /*
     * if there are more than one filter guard for this filter
     * (i.e. they are inherited), than they are OR combined
     * -> if one check succeeds => return 1
     */
    /*
    fprintf(stderr, "checking guard **%s**\n", ObjStr(fr->content));
    */
    cs->filtersForbidden++;
    frc = checkConditionInScope(in, fr->content);
    cs->filtersForbidden--;

    if (frc == TCL_OK) {
      /*
      fprintf(stderr, " +++ OK\n");
      */
      return TCL_OK;
    } else if (frc == TCL_ERROR) {
      Tcl_Obj* sr = Tcl_GetObjResult(in);
      INCR_REF_COUNT(sr);
      /*
      fprintf(stderr, " +++ ERROR\n");
      */
      XOTclVarErrMsg(in, "Filter Guard Error: '",
		     ObjStr(fr->content), "'\n\n",
		     ObjStr(sr), 0);
      DECR_REF_COUNT(sr);
      return TCL_ERROR;
    }
    fr = fr->next;
  }
  /*
  fprintf(stderr, " +++ FAILED\n");
  */
  return XOTCL_CHECK_FAILED;
}

/*
static void
FilterGuardPrint(Tcl_Interp* in, ClientData filterGuards) {
  XOTclTclObjList* fr = (XOTclTclObjList*) filterGuards;
  fprintf(stderr, " +++ <GUARDS> \n");
  while (fr) {

    fprintf(stderr, "   *     %s \n", ObjStr(fr->content));
    fr = fr->next;
  }
  fprintf(stderr, " +++ </GUARDS>\n");
}
*/

static void
FilterGuardDelList(XOTclCmdList* filterCL) {
  if (filterCL && filterCL->clientData) {
    TclObjListDeleteList ((XOTclTclObjList*)filterCL->clientData);
    filterCL->clientData = 0;
  }
}

static void
FilterGuardAdd(Tcl_Interp* in, XOTclCmdList* filterCL, Tcl_Obj* filterGuard) {
  if (filterGuard) {
    TclObjListNewElement((XOTclTclObjList**) &filterCL->clientData,
			 filterGuard);
  }
}
static void
FilterGuardAddList(Tcl_Interp* in, XOTclCmdList* dest, ClientData source) {
  XOTclTclObjList* s = (XOTclTclObjList*) source;
  while (s) {
    FilterGuardAdd(in, dest, (Tcl_Obj*) s->content);
    s = s->next;
  }
}

static int
FilterGuardAddGuardsOfFilterList(Tcl_Interp* in, XOTclCmdList* dest,
				 XOTclObject* obj, Command *filterCmd,
				 XOTclCmdList* filterList) {
  XOTclCmdList* h;
  if (filterList != 0) {
    h = CmdListFindCmdInList(filterCmd, filterList);
    if (h) {
      if (!h->clientData) {
	/*
	 * if the filter is defined somewhere without filter guards
	 * we do not have to apply any guards -> they are OR combined
	 * that means with one filter w/o guards check is always true
	 */
	  FilterGuardDelList(dest);
	  /*
	   * -1 means we're done
	   */
	  return -1;
      }
      FilterGuardAddList(in, dest, h->clientData);
      /*
       * 1 means "filterCmd" is a registered filter
       */
      return 1;
    }
  }
  /*
   * 0 means "filterCmd" is not a registered filter
   */
  return 0;
}

static void
FilterGuardAddInheritedGuards(Tcl_Interp* in, XOTclCmdList* dest,
			     XOTclObject* obj, Command *filterCmd) {
  XOTclClasses* pl;
  int isRegisteredFilter = 0, regClassOnMixin = 0, gr;

  /* search guards for instfilters registered on mixins */
  if (!(obj->flags & XOTCL_MIXIN_ORDER_VALID))
    MixinComputeDefined(in, obj);
    if ((obj->flags & XOTCL_MIXIN_ORDER_VALID) &&
	(obj->flags & XOTCL_MIXIN_ORDER_DEFINED)) {
    XOTclCmdList *ml = obj->mixinOrder;
    XOTclClass *mixin;
    while (ml) {
      mixin = GetClassFromCmdPtr(ml->cmdPtr);
      if (mixin) {
	if (mixin->opt) {
	  gr = FilterGuardAddGuardsOfFilterList(in, dest, obj, filterCmd,
						mixin->opt->instfilters);
	  
	  if (gr == -1) return; else if (gr == 1) isRegisteredFilter = 1;
	}

      }
      ml = ml->next;
    }
  }

  /* search per-object filters */
  if (obj->opt && obj->opt->filters) {
    gr = FilterGuardAddGuardsOfFilterList(in, dest, obj,
					  filterCmd, obj->opt->filters);
    if (gr == -1) return; else if (gr == 1) isRegisteredFilter = 1;
  }

  
    /* search per-class filters */
    if (!obj->cl->order)
      obj->cl->order = ComputeOrder(obj->cl, Super);
    pl = obj->cl->order;

    while (pl) {
      if (pl->cl->opt) {
	gr = FilterGuardAddGuardsOfFilterList(in, dest, obj, filterCmd,
					      pl->cl->opt->instfilters);
	if (gr == -1) return; else if (gr == 1) isRegisteredFilter = 1;
      }

      pl = pl->next;
    }


  /*
   * if this is not a registered filter, it is an inherited filter, like:
   *   Class A
   *   A instproc f ...
   *   Class B -superclass A
   *   B instproc {{f {<guard>}}}
   *   B instfilter f
   * -> get the guard from the filter that inherits it (here B->f)
   */
  if (!isRegisteredFilter) {
    XOTclCmdList* registeredFilter =
      CmdListFindNameInList(in,
	 (char*) Tcl_GetCommandName(in, (Tcl_Command) filterCmd),
	 obj->filterOrder);
    if (registeredFilter) {
      FilterGuardAddList(in, dest, registeredFilter->clientData);
    }
  }
}

static int
FilterGuardList(Tcl_Interp* in, XOTclCmdList* frl, char* filterName) {
  XOTclCmdList* h;
  if (frl) {
    h = CmdListFindNameInList(in, filterName, frl);
    if (h) {
      Tcl_ResetResult(in);
      if (h->clientData) {
	XOTclTclObjList* g = (XOTclTclObjList*) h->clientData;
	Tcl_SetObjResult(in, g->content);
      }
      return TCL_OK;
    }
  }
  return XOTclVarErrMsg(in, "info (inst)filterguard: can't find filter ",
			filterName, 0);
}

/*
 * append a filter command to the 'filterList' of an obj/class
 */
static int
FilterAdd(Tcl_Interp* in, XOTclCmdList** filterList, Tcl_Obj* name,
	  XOTclObject* startingObj, XOTclClass* startingCl) {
  Command* cmdPtr;
  int ocName; Tcl_Obj** ovName;
  Tcl_Obj* filterGuard = 0;
  XOTclCmdList* new;

  if (Tcl_ListObjGetElements(in, name, &ocName, &ovName) == TCL_OK
	&& ocName > 1) {
    if (ocName == 2) {
      name = ovName[0];
      filterGuard = ovName[1];
    } else return XOTclVarErrMsg(in,
				 "filter registration '",
				 ObjStr(name), "' has too many elements.",
				 (char*) NULL);
  }

  cmdPtr = FilterSearch(in, ObjStr(name),
			startingObj, startingCl);

  if (!cmdPtr) {
    if (startingObj)
      return XOTclVarErrMsg(in,
			    "filter: can't find filterproc on: ",
			    ObjStr(startingObj->cmdName), " - proc: ",
			    ObjStr(name), (char*) NULL);
    else
      return XOTclVarErrMsg(in,
			    "instfilter: can't find filterproc on: ",
			    ObjStr(startingCl->object.cmdName), " - proc: ",
			    ObjStr(name), (char*) NULL);
  }
  /*
  fprintf(stderr, " +++ adding filter %s \n", ObjStr(name));
  */
  new = CmdListAdd(filterList, cmdPtr, /*noDuplicates*/ 1);
  FilterGuardAdd(in, new, filterGuard);

  return TCL_OK;
}

/*
 * reset the filter order cached in obj->filterOrder
 */
static void
FilterResetOrder(XOTclObject* obj) {
  CmdListRemoveList(&obj->filterOrder, FilterGuardDelList);
  obj->filterOrder = 0;
}

/*
 * search the filter in the hierarchy again with FilterSearch, e.g. 
 * upon changes in the class hierarchy or mixins that carry the filter
 * command, so that we can be sure it is still reachable.
 */
static void
FilterSearchAgain(Tcl_Interp* in, XOTclCmdList **filters, 
		  XOTclObject* startingObj, XOTclClass* startingCl) {
  char *simpleName;
  Command *cmdPtr;
  XOTclCmdList* cmdList, *del;
  CmdListRemoveEpoched(filters, FilterGuardDelList);
  cmdList = *filters;
  while (cmdList) {
    simpleName = (char*) Tcl_GetCommandName(in, 
					    (Tcl_Command)cmdList->cmdPtr);
    cmdPtr = FilterSearch(in, simpleName, startingObj, startingCl);
    if (cmdPtr == NULL) {
      del = cmdList;
      cmdList = cmdList->next;
      del = CmdListRemoveFromList(filters, del);
      CmdListDeleteCmdListEntry(del, FilterGuardDelList);
    } else {
      if (cmdPtr != cmdList->cmdPtr)
	CmdListReplaceCmd(cmdList, cmdPtr);
      cmdList = cmdList->next;
    }
  }
  /* some entries might be NULL now, if they are not found anymore
     -> delete those 
  CmdListRemoveNulledEntries(filters, FilterGuardDelList);
  */

}

/*
 * if the class hierarchy or class filters have changed ->
 * invalidate filter entries in all dependent instances
 *
 */
static void
FilterInvalidateObjOrders(Tcl_Interp* in, XOTclClass* cl) {
  XOTclClasses *saved = cl->order, *clPtr, *savePtr;

  cl->order = 0;
  savePtr = clPtr = ComputeOrder(cl, Sub);
  cl->order = saved;

  while (clPtr != 0) {
    Tcl_HashSearch hSrch;
    Tcl_HashEntry* hPtr = &clPtr->cl->instances ?
      Tcl_FirstHashEntry(&clPtr->cl->instances, &hSrch) : 0;

    /* recalculate the commands of all instfilter registrations */
    if (clPtr->cl->opt)
      FilterSearchAgain(in, &clPtr->cl->opt->instfilters, 0, clPtr->cl);

    for (; hPtr != 0; hPtr = Tcl_NextHashEntry(&hSrch)) {
      XOTclObject* obj = (XOTclObject*)
	Tcl_GetHashKey(&clPtr->cl->instances, hPtr);
      FilterResetOrder(obj);
      obj->flags &= ~XOTCL_FILTER_ORDER_VALID;

      /* recalculate the commands of all object filter registrations */
      if (obj->opt)
	FilterSearchAgain(in, &obj->opt->filters, obj, 0);
    }
    clPtr = clPtr->next;
  }
  XOTclRemoveClasses(savePtr);
}

/*
 * from cl on down the hierarchy we remove all filters
 * the refer to "removeClass" namespace. E.g. used to
 * remove filters defined in superclass list from dependent
 * class cl
 */
static void
FilterRemoveDependentFilterCmds(XOTclClass* cl, XOTclClass* removeClass) {
  XOTclClasses *saved = cl->order, *clPtr;
  cl->order = 0;
  clPtr = ComputeOrder(cl, Sub);

  while (clPtr != 0) {
    Tcl_HashSearch hSrch;
    Tcl_HashEntry* hPtr = &clPtr->cl->instances ?
      Tcl_FirstHashEntry(&clPtr->cl->instances, &hSrch) : 0;
    if (clPtr->cl->opt) {
      CmdListRemoveNSFromList(&clPtr->cl->opt->instfilters,
			      (Namespace*)removeClass->nsPtr,
			      FilterGuardDelList);
    }
    for (; hPtr != 0; hPtr = Tcl_NextHashEntry(&hSrch)) {
      XOTclObject* obj = (XOTclObject*)
	Tcl_GetHashKey(&clPtr->cl->instances, hPtr);
      if (obj->opt) {
	CmdListRemoveNSFromList(&obj->opt->filters,
				(Namespace*)removeClass->nsPtr,
				FilterGuardDelList);
      }
    }
    clPtr = clPtr->next;
  }

  XOTclRemoveClasses(cl->order); cl->order = saved;
}

/*
 * build up a qualifier of the form <obj/cl> proc/instproc <procName>
 * if cl is not NULL, we build an instproc identifier for cl, else a proc
 * with obj
 */
static Tcl_Obj*
getFullProcQualifier(Tcl_Interp *in, CONST84 char* cmdName,
		     XOTclObject* obj, XOTclClass* cl) {
  Tcl_Obj *list = Tcl_NewListObj(0, NULL);
  Tcl_Obj* procObj = Tcl_NewStringObj(cmdName, -1);
  if (cl) {
    Tcl_ListObjAppendElement(in, list, cl->object.cmdName);
    Tcl_ListObjAppendElement(in, list, global_objects[INSTPROC]);
  } else {
    Tcl_ListObjAppendElement(in, list, obj->cmdName);
    Tcl_ListObjAppendElement(in, list, global_objects[PROC]);
  }
  Tcl_ListObjAppendElement(in, list, procObj);
  return list;
}

/*
 * info option for filters and instfilters
 * withGuards -> if not 0 => append guards
 * fullProcQualifiers -> if not 0 => full names with obj/class proc/instproc
 */
static int
FilterInfo(Tcl_Interp* in, XOTclCmdList* f, char* pattern,
	   int withGuards, int fullProcQualifiers) {
  CONST84 char* simpleName;
  Tcl_Obj *list = Tcl_NewListObj(0, NULL);

  while (f) {
    simpleName = Tcl_GetCommandName(in, (Tcl_Command)f->cmdPtr);
    if (!pattern || Tcl_StringMatch(simpleName, pattern)) {
      if (withGuards && f->clientData) {
	Tcl_Obj* innerList = Tcl_NewListObj(0,0);
	XOTclTclObjList* g = (XOTclTclObjList*) f->clientData;
	Tcl_ListObjAppendElement(in, innerList,
				 Tcl_NewStringObj(simpleName, -1));
	Tcl_ListObjAppendElement(in, innerList, g->content);
	Tcl_ListObjAppendElement(in, list, innerList);
      } else {
	if (fullProcQualifiers) {
	  XOTclClass* fcl = GetClass(in,
				     NSCutXOTclClasses(f->cmdPtr->nsPtr->fullName));
	  XOTclObject* fobj = fcl ? 0 : GetObject(in, f->cmdPtr->nsPtr->fullName);
	  Tcl_ListObjAppendElement(in, list,
				   getFullProcQualifier(in, simpleName,
							fobj, fcl));
	} else {
	  Tcl_ListObjAppendElement(in, list,
				   Tcl_NewStringObj(simpleName, -1));
	}
      }
    }
    f = f->next;
  }
  Tcl_SetObjResult(in, list);
  return TCL_OK;
}

/*
 * Appends XOTclCmdPtr* containing the filter cmds and their
 * superclass specializations to 'filterList'
 */
static void
FilterComputeOrderFullList(Tcl_Interp* in,
			   XOTclCmdList** filters,
			   XOTclCmdList** filterList) {
  XOTclCmdList *f ;
  char* simpleName;
  XOTclClass* fCl;
  XOTclClasses *pl;

  /*
   * ensure that no epoched command is in the filters list
   */
  CmdListRemoveEpoched(filters, FilterGuardDelList);
  f = *filters;

  while (f) {
    simpleName = (char*) Tcl_GetCommandName(in, (Tcl_Command)f->cmdPtr);
    fCl = GetClass(in,
		   NSCutXOTclClasses(f->cmdPtr->nsPtr->fullName));

    CmdListAdd(filterList, f->cmdPtr, /*noDuplicates*/ 0);

    if (!fCl) {
      /* try to find the object for per-object filter */
      XOTclObject* fObj = GetObject(in, f->cmdPtr->nsPtr->fullName);
      /* and then seek class/inherited filters */
      if (fObj) fCl = fObj->cl;
      /*
      fprintf(stderr, " *%s::%s*, ",  fObj->nsPtr->fullName, simpleName);
      */
    } else {
      /*
      fprintf(stderr, " *%s::%s*, ",  ObjStr(fCl->object.cmdName), simpleName);
     */
    }

    /* if we have a filter class -> search up the inheritance hierarchy*/
    if (fCl) {
      if (!fCl->order)
	fCl->order = ComputeOrder(fCl, Super);
      pl = fCl->order;
      if (pl && pl->next) {
	/* don't search on the start class again */
	pl = pl->next;
	/* now go up the hierarchy */
	while (pl) {
	  Command *pi = FindMethod(simpleName, pl->cl->nsPtr);
	  if (pi) {
	    CmdListAdd(filterList, pi, /*noDuplicates*/ 0);
	    /*
	    fprintf(stderr, " %s::%s, ", ObjStr(pl->cl->object.cmdName), simpleName);
	    */
	  }
	  pl = pl->next;
	}
      }
    }
    f = f->next;
  }
}

/*
 * Computes a linearized order of filter and instfilter. Then
 * duplicates in the full list and with the class inheritance list of
 * 'obj' are eliminated.
 * The precendence rule is that the last occurence makes it into the
 * final list.
 */
static void
FilterComputeOrder(Tcl_Interp* in, XOTclObject* obj) {
  XOTclCmdList *filterList = 0, *next, *checker, *newlist;
  XOTclClasses *pl;

  FilterResetOrder(obj);
  /*
  fprintf(stderr, "<Filter Order obj=%s> List: ", ObjStr(obj->cmdName));
  */

  /* append instfilters registered for mixins */
  if (!(obj->flags & XOTCL_MIXIN_ORDER_VALID))
    MixinComputeDefined(in, obj);

  if ((obj->flags & XOTCL_MIXIN_ORDER_VALID) &&
      (obj->flags & XOTCL_MIXIN_ORDER_DEFINED)) {
    XOTclCmdList *ml = obj->mixinOrder;
    XOTclClass *mixin;
    while (ml) {
      mixin = GetClassFromCmdPtr(ml->cmdPtr);
      if (mixin && mixin->opt && mixin->opt->instfilters)
	FilterComputeOrderFullList(in, &mixin->opt->instfilters, &filterList);
      ml = ml->next;
    }
  }

  /* append per-obj filters */
  if (obj->opt)
    FilterComputeOrderFullList(in, &obj->opt->filters, &filterList);

  /* append per-class filters */
  if (!obj->cl->order)
    obj->cl->order = ComputeOrder(obj->cl, Super);
  pl = obj->cl->order;
  while (pl) {
    if (pl->cl->opt && pl->cl->opt->instfilters) {
      FilterComputeOrderFullList(in, &pl->cl->opt->instfilters, &filterList);
    }
    pl = pl->next;
  }

  /*
  fprintf(stderr, "\n");
  */
  /* use no duplicates & no classes of the precedence order
     on the resulting list */
  while (filterList) {
    checker = next = filterList->next;
    while (checker) {
      if (checker->cmdPtr == filterList->cmdPtr) break;
      checker = checker->next;
    }
    if (checker == 0) {
      newlist = CmdListAdd(&obj->filterOrder, filterList->cmdPtr,
			   /*noDuplicates*/ 0);
      FilterGuardAddInheritedGuards(in, newlist, obj, filterList->cmdPtr);
      /*
      fprintf(stderr, "  Adding %s::%s,\n", filterList->cmdPtr->nsPtr->fullName, Tcl_GetCommandName(in, (Tcl_Command)filterList->cmdPtr));
      FilterGuardPrint(in, newlist->clientData);
      */
      
    }

    CmdListDeleteCmdListEntry(filterList, FilterGuardDelList);

    filterList = next;
  }
  /*
  fprintf(stderr, "</Filter Order>\n");
  */
}

/*
 * the filter order is either
 *   DEFINED (there are filter on the instance),
 *   NONE    (there are no filter for the instance),
 *   or INVALID (a class re-strucuturing has occured, thus it is not clear
 *               whether filters are defined or not).
 * If it is INVALID FilterComputeDefined can be used to compute the order
 * and set the instance to DEFINE or NONE
 */
static void
FilterComputeDefined(Tcl_Interp* in, XOTclObject* obj) {
  FilterComputeOrder(in, obj);
  obj->flags |= XOTCL_FILTER_ORDER_VALID;
  if (obj->filterOrder)
    obj->flags |= XOTCL_FILTER_ORDER_DEFINED;
  else
    obj->flags &= ~XOTCL_FILTER_ORDER_DEFINED;
}

/*
 * push a filter stack information on this object
 */
static void
FilterStackPush(Tcl_Interp* in, XOTclObject* obj, Tcl_Obj* calledProc) {
  XOTclCallStackContent *csc = CallStackGetFrame(in);
  XOTclFilterStack* h = NEW(XOTclFilterStack);

  h->currentCmdPtr = 0;
  h->filterChainOn = 0;
  h->calledProc = calledProc;
  INCR_REF_COUNT(h->calledProc);
  h->callingProc =
    Tcl_NewStringObj(Tcl_GetCommandName(in, (Tcl_Command) csc->cmdPtr), -1);
  INCR_REF_COUNT(h->callingProc);

  h->callingClass =
    csc->cl ? csc->cl->object.cmdName : global_objects[EMPTY];
  INCR_REF_COUNT(h->callingClass);

  h->callingObject =
    csc->self ? csc->self->cmdName : global_objects[EMPTY];
  INCR_REF_COUNT(h->callingObject);

  h->next = obj->filterStack;
  obj->filterStack = h;
}

/*
 * pop a filter stack information on this object
 */
static void
FilterStackPop(XOTclObject* obj) {
  XOTclFilterStack *h = obj->filterStack;
  obj->filterStack = h->next;

  /* free stack entry */
  DECR_REF_COUNT(h->calledProc);
  DECR_REF_COUNT(h->callingProc);
  DECR_REF_COUNT(h->callingClass);
  DECR_REF_COUNT(h->callingObject);
  FREE(XOTclFilterStack, h);
}

/*
 * walk through the filter order until the current filter is reached.
 * then use the next filter as current filter.
 *
 * precondition: obj->filterStack is not NULL
 */
static void
FilterSeekCurrent(Tcl_Interp* in, XOTclObject* obj,
		  XOTclCmdList** cmdList) {
  Command* currentCmdPtr = obj->filterStack->currentCmdPtr;

  /* ensure that the filter order is not invalid, otherwise compute order */
  if (!(obj->flags & XOTCL_FILTER_ORDER_VALID))
    FilterComputeDefined(in, obj);

  *cmdList = obj->filterOrder;

  /* go forward to current class */
  while (*cmdList && currentCmdPtr) {
    if ((*cmdList)->cmdPtr == currentCmdPtr) currentCmdPtr = 0;
    *cmdList = (*cmdList)->next;
  }
}

/*
 * seek through the filters active for "obj" and check whether cmdPtr
 * is among them
 */
static XOTCLINLINE int
FilterActiveOnObj(Tcl_Interp* in, XOTclObject* obj, Command* cmdPtr) {
  XOTclCallStack *cs = &RUNTIME_STATE(in)->cs;
  XOTclCallStackContent *bot = cs->content;
  register XOTclCallStackContent *csc = cs->top;
  while (csc > bot) {
    /* only check the callstack entries for this object &&
       only check the callstack entries for the given cmdPtr */
    if (obj == csc->self && cmdPtr == csc->cmdPtr &&
	csc->cscType == XOTCL_CSC_TYPE_ACTIVE_FILTER) {
      return 1;
    }
    csc--;
  }
  return 0;
}

/*
 * search through the filter list on obj and class hierarchy
 * for registration of a command ptr as filter
 *
 * returns a tcl obj list with the filter registration, like:
 * "<obj> filter <filterName>,
 * "<class> instfilter <filterName>,
 * or an empty list, if not registered
 */
static Tcl_Obj*
FilterFindReg(Tcl_Interp* in, XOTclObject* obj, Command* cmdPtr) {
  Tcl_Obj *list = Tcl_NewListObj(0, NULL);
  XOTclClasses* pl;

  /* search per-object filters */
  if (obj->opt && CmdListFindCmdInList(cmdPtr, obj->opt->filters)) {
    Tcl_ListObjAppendElement(in, list, obj->cmdName);
    Tcl_ListObjAppendElement(in, list, global_objects[FILTER]);
    Tcl_ListObjAppendElement(in, list,
      Tcl_NewStringObj(Tcl_GetCommandName(in, (Tcl_Command)cmdPtr), -1));
    return list;
  }

  /* search per-class filters */
  if (!obj->cl->order)
    obj->cl->order = ComputeOrder(obj->cl, Super);
  pl = obj->cl->order;
  while (pl) {
    if (pl->cl->opt && pl->cl->opt->instfilters) {
      if (CmdListFindCmdInList(cmdPtr, pl->cl->opt->instfilters)) {
	Tcl_ListObjAppendElement(in, list, pl->cl->object.cmdName);
	Tcl_ListObjAppendElement(in, list, global_objects[INSTFILTER]);
	Tcl_ListObjAppendElement(in, list, 
	  Tcl_NewStringObj(Tcl_GetCommandName(in, (Tcl_Command)cmdPtr), -1));
	return list;
      }
    }
    pl = pl->next;
  }
  return list;
}

/*
 * remove a filter from the filter stack of object 'obj'
 */
static void
FilterRemoveFromFilterStack(XOTclCmdList* del, XOTclObject* obj) {
  /* if the current filter is to be deleted, we make its predecessor
     current filter */
  if (obj->filterStack->currentCmdPtr == del->cmdPtr) {
    if (obj->filterOrder && del->cmdPtr != obj->filterOrder->cmdPtr) {
      XOTclCmdList* predecessor = obj->filterOrder;
      while (predecessor->next  &&
	     predecessor->next->cmdPtr != del->cmdPtr)
	predecessor = predecessor->next;
      if (predecessor->next)
	obj->filterStack->currentCmdPtr = predecessor->next->cmdPtr;
    } else {
      obj->filterStack->currentCmdPtr = 0;
    }
  }
}
/*
 * remove filter with a command ptr from the filter list and delete it
 * if 'obj' is specified, we also remove it from the object's filter stack
 */
static void
FilterRemoveFromCmdPtr(XOTclCmdList** filterList, Command* c, XOTclObject* obj) {
  XOTclCmdList* f = *filterList, *del;
  while (f) {
    if (f->cmdPtr == c) {
      del = CmdListRemoveFromList(filterList, f);
      if (del) {
	if (obj)
	  FilterRemoveFromFilterStack(del, obj);
	/* actually delete the filter entry */
	CmdListDeleteCmdListEntry(del, FilterGuardDelList);
      }
      break;
    }
    f = f->next;
  }
}

/*
 * remove filter from cmd ptr with obj ... search per-object and
 * per-class filters for the filter command 'c'
 */
static void
FilterRemoveOnObjFromCmdPtr(Tcl_Interp* in, XOTclObject* obj, Command* c) {
  XOTclClasses *pl;
  obj->flags &= ~XOTCL_FILTER_ORDER_VALID;

  if (obj->opt)
    FilterRemoveFromCmdPtr(&obj->opt->filters, c, obj);

  if (!(obj->flags & XOTCL_MIXIN_ORDER_VALID))
    MixinComputeDefined(in, obj);
  if ((obj->flags & XOTCL_MIXIN_ORDER_VALID) &&
      (obj->flags & XOTCL_MIXIN_ORDER_DEFINED)) {
    XOTclCmdList *ml = obj->mixinOrder;
    XOTclClass *mixin;
    while (ml) {
      mixin = GetClassFromCmdPtr(ml->cmdPtr);
      if (mixin->opt) {
	FilterRemoveFromCmdPtr(&mixin->opt->instfilters, c, obj);
	FilterInvalidateObjOrders(in, mixin);
      }
      ml = ml->next;
    }
  }

  if (!obj->cl->order)
    obj->cl->order = ComputeOrder(obj->cl, Super);
  pl = obj->cl->order;
  while (pl) {
    if (pl->cl->opt) {
      FilterRemoveFromCmdPtr(&pl->cl->opt->instfilters, c, obj);
      FilterInvalidateObjOrders(in, pl->cl);
    }
    pl = pl->next;
  }
}

/*
 * before we can perform a filter dispatch, FilterSearchProc seeks the
 * current filter and the relevant calling information
 */
static Command*
FilterSearchProc(Tcl_Interp* in, XOTclObject* obj,
		 Tcl_ObjCmdProc** proc, ClientData* cp,
		 Command** currentCmdPtr) {
  XOTclCmdList *cmdList;

  *currentCmdPtr = 0;
  *proc = 0;
  *cp = 0;

  if (!obj || !obj->filterStack)
    return 0;

  FilterSeekCurrent(in, obj, &cmdList);
  while (cmdList) {
    if(cmdList->cmdPtr->cmdEpoch) {
      FilterRemoveOnObjFromCmdPtr(in, obj, cmdList->cmdPtr);
      /*
       * invalidate the order & seek start class
       */
      FilterResetOrder(obj);
      FilterSeekCurrent(in, obj, &cmdList);
    } else if (FilterActiveOnObj(in, obj, cmdList->cmdPtr)) {
      /*
      fprintf(stderr, "Filter <%s> -- Active on: %s\n",
	     Tcl_GetCommandName(in, (Tcl_Command)cmdList->cmdPtr), ObjStr(obj->cmdName));
      */
      obj->filterStack->currentCmdPtr = cmdList->cmdPtr;
      FilterSeekCurrent(in, obj, &cmdList);
    } else {
      /* ok. we' ve found it */
      *proc = cmdList->cmdPtr->objProc;
      *cp   = cmdList->cmdPtr->objClientData;
      *currentCmdPtr = cmdList->cmdPtr;
      /*
      fprintf(stderr, "FilterSearchProc - found: %s, %p\n",
	     Tcl_GetCommandName(in, (Tcl_Command)cmdList->cmdPtr), cmdList->cmdPtr);
      */
      return cmdList->cmdPtr;
    }
  }
  return 0;
}

/*
 * cut filter frames on Tcl callstack, so that filters seem to be transparent
 * for upvar, uplevel, etc.
 *
 * return 1, if frames are cutted, otherwise 0
 */
static int
FilterCutFrames(Tcl_Interp* in, CallFrame **savedFramePtr,
		CallFrame **savedVarFramePtr) {
  Interp *iPtr = (Interp *) in;
  CallFrame *l, *newFrame, *oldFrame, *lastFrame, *lastVarFrame;
  int firstFrame = 1;
  int isObj;
  int level = (iPtr->framePtr) ? iPtr->framePtr->level : 0;
  XOTclCallStack *cs = &RUNTIME_STATE(in)->cs;
  XOTclCallStackContent
    *bot = cs->content,
    *csc = cs->top;

  /*
   * if the topmost frame is a filter -> instvar, uplevel, upvar, etc.
   * should function on the filter -> we do not cut anything
   */

  /* we increment the frameCutRound in order to be able to determine
     which frames have been cut in this cutting round */
  cs->frameCutRound++;

  /* for the time of the actual proc we create an enviroment
     without the filter calls (e.g. for the reason that uplevel/upvar
     function correctly) */

  *savedFramePtr = oldFrame = iPtr->framePtr;
  *savedVarFramePtr = iPtr->varFramePtr;
  iPtr->framePtr = 0;
  iPtr->varFramePtr = 0;
  lastFrame = 0; lastVarFrame = 0;

  /*
   * copy all frames that aren't inactive filters
   */
  while (oldFrame != 0) {
    while (csc > bot && csc->frameCut > 0) {
      csc--;
    }

    isObj = checkIsXOTclProcFrame(oldFrame);

    /*
     * we only want to ignore inactive filters
     */
    if (!isObj  /* if it is not an obj namespace => it can't be a filter */
	|| csc <= bot /* no more XOTcl entries => not a filter */
	|| !(oldFrame->procPtr->cmdPtr == csc->cmdPtr &&
	     csc->cscType == XOTCL_CSC_TYPE_INACTIVE_FILTER)
	/* if it is an XOTcl proc, it may only be a filter if the
	   XOTCL_CSC_TYPE_INACTIVE_FILTER cscType value is set. */
	) {
      CallFrame *f;

      f = NEW(CallFrame);
      memcpy(f, oldFrame, sizeof(CallFrame));

      /* if the var frame ptr is pointing to the current (old)
	 callframe => correct it to the new (copied) one */
      if (oldFrame == *savedVarFramePtr)
	 iPtr->varFramePtr = f;

      if (lastFrame)
        lastFrame->callerPtr = f;
      lastFrame = f;

      if (*savedVarFramePtr == oldFrame) {
        iPtr->varFramePtr = f;
      }
      if (firstFrame) {
        iPtr->framePtr = f;
        firstFrame = 0;
      }
    } else {
      level--;
      csc->frameCut = cs->frameCutRound;
    }

    /* determine corresponding callstack entry */
    if (isObj) {
      csc--;
    }
    oldFrame = oldFrame->callerPtr;
  }

  /* the old frame ptr was pointing somewhere, but this frame was cut,
     thus, the new var frame ptr is still 0 -> this happens, when an filter
     on top of the callstack is cut -> then the topmost frame is the target
     frame
  */
  if (*savedVarFramePtr != 0 && iPtr->varFramePtr == 0)
    iPtr->varFramePtr = iPtr->framePtr;

  if (lastFrame)
    lastFrame->callerPtr = 0;

  /*
   * correct callerVarPtr information
   */
  newFrame = iPtr->framePtr;
  while (newFrame) {
    if (newFrame->callerVarPtr && newFrame->callerPtr) {
      /* a caller var ptr is set, we have to correct it into the
	 new address space */
      if (newFrame->callerPtr->level == newFrame->callerVarPtr->level)
	/* both levels are the same => point to callerPtr */
	newFrame->callerVarPtr = newFrame->callerPtr;
      else {
	/* levels are different -> we have to search for it */
	CallFrame* hNew = newFrame;
	while (hNew) {
	  if (hNew->level == newFrame->callerVarPtr->level)
	    /* ok, it is there */
	    break;
	  hNew = hNew->callerPtr;
	}
	if (hNew && hNew->level == newFrame->callerVarPtr->level)
	  /* correct it to the found frame ptr */
	  newFrame->callerVarPtr = hNew;
	else {
	  /* oops, it is cut -> reset to the default (callerPtr) */
	  newFrame->callerVarPtr = newFrame->callerPtr;
	}
      }
    } else newFrame->callerVarPtr = 0;

    newFrame = newFrame->callerPtr;
  }

  /*
   * correct level information
   */
  level = 0;

  for(l = iPtr->framePtr; l != 0; l = l->callerVarPtr)
    level++;
  for(l = iPtr->framePtr; l != 0; l = l->callerVarPtr) {
    l->level = level;
    level--;
  }

  return 1;
}

static void
FilterRestoreFrames(Tcl_Interp* in, CallFrame *savedFramePtr,
		    CallFrame *savedVarFramePtr) {
  Interp *iPtr = (Interp *) in;
  CallFrame* save = iPtr->framePtr, *copy = save, *original, *l;
  int level = 0;
  XOTclCallStack *cs = &RUNTIME_STATE(in)->cs;
  XOTclCallStackContent *csc = cs->top;

  /* reset to old callframes */
  iPtr->framePtr = savedFramePtr;
  original = savedFramePtr;
  iPtr->varFramePtr = savedVarFramePtr;

  while (original != 0 && copy != 0) {
    int isObj = checkIsXOTclProcFrame(original);

    if (!(isObj && csc->frameCut)) {
      if (copy->procPtr && copy->procPtr->cmdPtr
	  && original->procPtr && original->procPtr->cmdPtr
	  && copy->procPtr->cmdPtr == original->procPtr->cmdPtr) {
	/* if local var table is created on copy -> let the original
	   use the created one */
	if (original->varTablePtr == 0)
	  original->varTablePtr = copy->varTablePtr;
	copy = copy->callerPtr;
      }
    }
    original = original->callerPtr;
    if (isObj)
      csc--;
  }
  /*
   * correct level information in original frames
   */
  level = 0;
  for(l = iPtr->framePtr; l != 0; l = l->callerVarPtr)
    level++;

  for(l = iPtr->framePtr; l != 0; l = l->callerVarPtr) {
    l->level = level;
    level--;
  }

  /* free memory */
  while (save != 0) {
    CallFrame *f= save;
    save = save->callerPtr;
    FREE(CallFrame, f);
  }

  /* Update the XOTcl CallStack cutted frames of this round */
  for (csc = &cs->content[1]; csc <= cs->top; csc++) {
    if (csc->frameCut == cs->frameCutRound) {
      csc->frameCut = 0;
    }
  }
  cs->frameCutRound--;
}

/* determine whether a filter on the current callstack or not */
static int
FilterEntryOnCallStack(Tcl_Interp* in) {
  XOTclCallStack *cs = &RUNTIME_STATE(in)->cs;
  XOTclCallStackContent *csc;
  for (csc = &cs->content[1]; csc <= cs->top; csc++) {
    if (isFilterEntry(csc)) {
      return 1;
    }
  }
  return 0;
}


/*
 * Search default values specified through 'parameter' on one class
 */
static int
SearchDefaultValuesOnClass(Tcl_Interp* in, XOTclObject* obj,
                           XOTclClass* cmdCl, XOTclClass* targetClass) {
  int result = TCL_OK;
  register Tcl_HashEntry* entryPtr = 0;
  Var* defaults;
  Namespace *nsp = (Namespace *)targetClass->object.nsPtr;

  if (nsp) {
    entryPtr =
      Tcl_FindHashEntry(&nsp->varTable, "__defaults");
  } else {
    if (targetClass->object.varTable)
      entryPtr =
	Tcl_FindHashEntry(targetClass->object.varTable, "__defaults");
  }
  if (entryPtr) {
    defaults = (Var*) Tcl_GetHashValue(entryPtr);
    /*
    fprintf(stderr, "+++ we have defaults for <%s>\n",
	    className(targetClass));
    */
    if (TclIsVarArray(defaults)) {
      Tcl_HashTable* table = defaults->value.tablePtr;
      Tcl_HashSearch hSrch;
      Tcl_HashEntry* hPtr = table ? Tcl_FirstHashEntry(table, &hSrch) : 0;

      for (; hPtr != 0; hPtr = Tcl_NextHashEntry(&hSrch)) {
	char *varName        = Tcl_GetHashKey(table, hPtr);
	Tcl_Obj *varNameObj  = Tcl_NewStringObj(varName,-1);
	Var  *val            = (Var*)Tcl_GetHashValue(hPtr);
	INCR_REF_COUNT(varNameObj);
	if (TclIsVarScalar(val)) {
	  Tcl_Obj *oldValue;
	  oldValue = XOTclOGetInstVar2(obj, in, varNameObj, NULL,
				       TCL_LEAVE_ERR_MSG|TCL_PARSE_PART1);
	  /** we check whether the variable is already set.
	      if so, we do not set it again */
	  if (oldValue == NULL) {
	    char *value = ObjStr(val->value.objPtr), *v;
	    Tcl_Obj *valueObj = val->value.objPtr;
	    int doSubst = 0;
	    for (v=value; *v; v++) {
	      if (*v == '[' && doSubst == 0)
		doSubst = 1;
	      else if ((doSubst == 1 && *v == ']') || *v == '$') {
		doSubst = 2;
		break;
	      }
	    }
	    if (doSubst == 2) { /* we have to subst */
	      Tcl_Obj *ov[2];
	      int rc = CallStackPush(in, obj, cmdCl, varName, 0, 1,
				     &varNameObj, 0, 0);
	      if (rc != TCL_OK) {
		DECR_REF_COUNT(varNameObj);
		return rc;
	      }
	      ov[1] = valueObj;
	      Tcl_ResetResult(in);
	      rc = callCommand(in, SUBST, 2, ov);
	      CallStackPop(in);
	      if (rc == TCL_OK) {
		valueObj = Tcl_GetObjResult(in);
	      } else {
		DECR_REF_COUNT(varNameObj);
		return rc;
	      }
	    }
	    INCR_REF_COUNT(valueObj);
	    result = callMethodWithArg((ClientData)obj, in,
				       varNameObj, valueObj, 3, 0, 0);
	    DECR_REF_COUNT(valueObj);
	  }
	}
	DECR_REF_COUNT(varNameObj);
      }
    }
  }
  return result;
}

/*
 * Search default values specified through 'parameter' on
 * mixin and class hierarchy
 */
static int
SearchDefaultValues(Tcl_Interp* in,
                    XOTclObject* obj, XOTclClass* cmdCl) {
  ObjFrameDecls;
  XOTclClass *cl = obj->cl, *mixin;
  XOTclClasses *pl = 0;
  XOTclCmdList *ml = 0;
  int result = TCL_OK;

  if (!(obj->flags & XOTCL_MIXIN_ORDER_VALID))
    MixinComputeDefined(in, obj);
  if ((obj->flags & XOTCL_MIXIN_ORDER_VALID) &&
      (obj->flags & XOTCL_MIXIN_ORDER_DEFINED))
    ml = obj->mixinOrder;

  if (cl) {
    if (!cl->order)
      cl->order = ComputeOrder(cl, Super);
    pl = cl->order;
  }

  XOTclPushObjFrame(in,obj);

  while (ml) {
    mixin = GetClassFromCmdPtr(ml->cmdPtr);
    result = SearchDefaultValuesOnClass(in, obj, cmdCl, mixin);
    if (result != TCL_OK)
      break;
    ml = ml->next;
  }
  while (pl) {
    result = SearchDefaultValuesOnClass(in, obj, cmdCl, pl->cl);
    if (result != TCL_OK)
      break;
    pl = pl->next;
  }

  XOTclPopObjFrame(in,obj);

  return result;
}

static int
ParameterSearchDefaultsMethod(ClientData cd, Tcl_Interp* in, int objc,
			      Tcl_Obj *objv[]) {
  XOTclClass *cl = XOTclObjectToClass(cd);
  XOTclObject *defaultObj;

  if (!cl) return XOTclObjErrType(in, objv[0], "Class");
  if (objc != 2)
    return XOTclObjErrArgCnt(in,
                             cl->object.cmdName,
                             "searchDefaults obj");
  if (GetXOTclObjectFromObj(in, objv[1], &defaultObj) != TCL_OK)
    return XOTclVarErrMsg(in, "Can't find default object ",
                          ObjStr(objv[1]), 0);

  /*
   *  Search for default values for vars on superclasses
   */
  return SearchDefaultValues(in, defaultObj, defaultObj->cl);
}

static int
callParameterMethodWithArg(XOTclObject* obj, Tcl_Interp* in,
			   Tcl_Obj* method, Tcl_Obj* arg,
			   int objc, Tcl_Obj *CONST objv[], int flags) {
  Tcl_Obj *pcl = (obj->cl->opt && obj->cl->opt->parameterClass) ?
    obj->cl->opt->parameterClass : global_objects[PARAM_CL];
  XOTclClass *paramCl;
  int result;

  if (GetXOTclClassFromObj(in,pcl,&paramCl) == TCL_OK) {
    result = callMethodWithArg((ClientData)paramCl, in,
			       method, arg, objc, objv, flags);
  }
  else
    result = XOTclVarErrMsg(in, "create: can't find parameter class", 0);
  return result;
}


/*
 * method dispatch
 */

/* actually call a method (with assertion checking) */
static int
callProcCheck(ClientData cp, Tcl_Interp* in, int objc, Tcl_Obj *CONST objv[],
	      Command *procInfo,
	      XOTclObject* obj, XOTclClass *cl, char *methodName,
	      int isFilterEntry , int isMixinEntry, int* isDestroyed) {
  int result, checkoptions;
  Interp *iPtr = (Interp *) in;

#if defined(PROFILE)
  long int startUsec, startSec;
  struct timeval trt;

  gettimeofday(&trt, NULL);
  startSec  = trt.tv_sec;
  startUsec = trt.tv_usec;
#endif
  assert(obj);

  *isDestroyed = 0;

  if (!TclIsProc(procInfo)) {
    if (obj->teardown) {
      checkoptions = obj->opt ? obj->opt->checkoptions : CHECK_NONE;
      if (checkoptions & CHECK_INVAR &&
	  AssertionCheckInvars(in, obj, methodName) == TCL_ERROR)
	return TCL_ERROR;
    }
    RUNTIME_STATE(in)->callIsDestroy = 0;

    result = (*procInfo->objProc)(cp, in, objc, objv);

    /*
    if (obj && obj->teardown && cl && !(obj->flags & XOTCL_DESTROY_CALLED)) {
      fprintf(stderr, "Obj= %s ", ObjStr(obj->cmdName));
      fprintf(stderr, "CL= %s ", ObjStr(cl->object.cmdName));
      fprintf(stderr, "method=%s\n", methodName);
      }
    */
    if (RUNTIME_STATE(in)->callIsDestroy) {
      *isDestroyed = 1;
    } else {
      checkoptions = obj->opt ? obj->opt->checkoptions : CHECK_NONE;
      if (checkoptions & CHECK_INVAR &&
	  AssertionCheckInvars(in, obj, methodName) == TCL_ERROR)
	return TCL_ERROR;
    }
  } else
    if (CallStackPush(in, obj, cl, methodName, procInfo, objc, objv,
		      isFilterEntry, isMixinEntry) == TCL_OK) {
      /*
       * if this is a filter, check whether its filterguard applies,
       * if not: just step forward to the next filter
       */
      if (isFilterEntry) {
	XOTclCmdList *cmdList;
	/*
	 * seek procInfo in obj's filterOrder
	 */
	if (!(obj->flags & XOTCL_FILTER_ORDER_VALID))
	  FilterComputeDefined(in, obj);
	cmdList = obj->filterOrder;
	while (cmdList && cmdList->cmdPtr != procInfo)
	  cmdList = cmdList->next;
	
	/*
	 * when it is found, check whether it has a filter guard
	 */
	if (cmdList && cmdList->clientData) {
	  ObjFrameDecls;
	  int rc;

	  /*
	  FilterGuardPrint(in, cmdList->clientData);
	  */
	  /*
	   * ok, there is a filter guard ... we have to push a
	   * fake callframe on the tcl stack so that uplevel
	   * is in sync with the XOTcl callstack, and we can uplevel
	   * into the above pushed CallStack entry of the filter
	   */

	  XOTclPushObjFrame(in,obj);

	  rc = FilterGuardCheck(in, cmdList->clientData);

	  XOTclPopObjFrame(in,obj);

	  if (rc != TCL_OK) {
	    CallStackPop(in);
	    if (rc == TCL_ERROR)
	      return TCL_ERROR;
	    /*
	     * call next, use the given objv's, not the callstack objv
	     * we may not be in a method, thus there may be wrong or
	     * no callstackobjs
	     */
	    result = XOTclNextMethod(obj, in, cl, methodName,
				     objc, objv, /*useCallStackObjs*/ 0,
				     isDestroyed);
	    return result;
	  }
	}
      }

      if (obj->teardown && !(obj->flags & XOTCL_DESTROY_CALLED)) {
	checkoptions = obj->opt ? obj->opt->checkoptions : CHECK_NONE;
	if (checkoptions != CHECK_NONE &&
	    AssertionCheck(in, obj, cl, methodName, CHECK_PRE) == TCL_ERROR) {
	  CallStackPop(in);
	  return TCL_ERROR;
	}
      }
      if (iPtr->numLevels <= 2)
	RUNTIME_STATE(in)->returnCode=TCL_OK;
#ifdef DISPATCH_TRACE
      printCall(in,"callProcCheck", objc,objv);
      fprintf(stderr, " returnCode %d xotcl rc %d\n",
	      iPtr->returnCode, RUNTIME_STATE(in)->returnCode);
#endif

      result = (*procInfo->objProc)(cp, in, objc, objv);

#ifdef DISPATCH_TRACE
      printExit(in,"callProcCheck", objc,objv, result);
      fprintf(stderr, " returnCode %d xotcl rc %d\n",
	      iPtr->returnCode, RUNTIME_STATE(in)->returnCode);
#endif
      if (iPtr->numLevels <= 2 && RUNTIME_STATE(in)->returnCode == TCL_BREAK)
	result = TCL_BREAK;
      else if (result == TCL_BREAK && RUNTIME_STATE(in)->returnCode == TCL_OK)
	RUNTIME_STATE(in)->returnCode = result;

      /* we give the information whether the call has destroyed the
	 object back to the caller, because after CallStackPop it
	 cannot be retrieved via the call stack */
      *isDestroyed = CallStackIsDestroyed(in);

      /* if the object is destroyed -> the assertion structs's are already
	 destroyed */
      if (!*isDestroyed && obj->teardown) {
	checkoptions = obj->opt ? obj->opt->checkoptions : CHECK_NONE;
	if (checkoptions != CHECK_NONE &&
	    AssertionCheck(in, obj, cl, methodName, CHECK_POST) == TCL_ERROR) {
	  CallStackPop(in);
	  return TCL_ERROR;
	}
      }

      CallStackPop(in);
    } else
      result = TCL_ERROR;

#if defined(PROFILE)
  if (*isDestroyed == 0) {
    XOTclProfileEvaluateData(in, startSec, startUsec, obj, cl, methodName);
  }
#endif

  return result;
}

static XOTCLINLINE int
DoDispatch(ClientData cd, Tcl_Interp* in, int objc, Tcl_Obj *CONST objv[],
	   int flags) {
  XOTclObject *obj = (XOTclObject*)cd;
  int result = TCL_OK,
    isDestroyed = 0,
    mixinStackPushed = 0,
    filterStackPushed = 0,
    isMixinEntry = 0,
    isFilterEntry = 0;
  char *methodName = ObjStr(objv[1]),
    *callMethod = methodName;
  XOTclClass *cl = 0;
  ClientData cp = 0;
  Tcl_ObjCmdProc *proc = 0;
  Command *procInfo = 0;
  Tcl_Obj *objName = obj->cmdName, *clName = 0;
  XOTclCallStack *cs = &RUNTIME_STATE(in)->cs;

#ifdef AUTOVARS
  int isNext = isNextString(methodName);
#endif

#ifdef DISPATCH_TRACE
  printCall(in,"DISPATCH", objc,objv);
#endif

  INCR_REF_COUNT(objName);

  if (!(obj->flags & XOTCL_FILTER_ORDER_VALID))
    FilterComputeDefined(in, obj);

  if (!(obj->flags & XOTCL_MIXIN_ORDER_VALID))
    MixinComputeDefined(in, obj);

#ifdef AUTOVARS
    if(!isNext) {
#endif
      if (!(flags & XOTCL_CM_NO_FILTERS) && !cs->filtersForbidden) {
	/* Only start new filter chain, if
	   (a) filters are defined and
	   (b) the toplevel csc entry is not an filter on self
	*/
	if ((obj->flags & XOTCL_FILTER_ORDER_VALID) &&
	    (obj->flags & XOTCL_FILTER_ORDER_DEFINED)) {
	  XOTclObject* self = GetSelfObj(in);
	  if (obj != self ||
	      cs->top->cscType != XOTCL_CSC_TYPE_ACTIVE_FILTER) {
	    Tcl_ObjCmdProc *fproc = 0;
	    Command        *fProcInfo =0;
	    ClientData     fcp;
	    Command* currentCmdPtr;
	
	    FilterStackPush(in, obj, objv[1]);
	    filterStackPushed = 1;
	    fProcInfo = FilterSearchProc(in, obj, &fproc,
					 &fcp, &currentCmdPtr);
	
	    obj->filterStack->currentCmdPtr = currentCmdPtr;
	
	    if (fproc) {
	      isFilterEntry = 1;
	      cl = GetClass(in, NSCutXOTclClasses(fProcInfo->nsPtr->fullName));
	      proc = fproc;
	      cp = fcp;
	      procInfo = fProcInfo;
	      callMethod =
		(char*) Tcl_GetCommandName(in, (Tcl_Command)fProcInfo);
	      obj->filterStack->filterChainOn = 1;
	    } else {
	      FilterStackPop(obj);
	      filterStackPushed = 0;
	    }
	  } /*else XOTclCallStackTrace(in);*/
	}
      }

      /* check if a mixin is to be called.
	 don't use mixins on next method calls, since normally it is not
	 intercepted (it is used as a primitive command).
	 don't use mixins on init calls, since init is invoked on mixins
	 during mixin registration (in XOTclOMixinMethod)
      */

      if ((obj->flags & XOTCL_MIXIN_ORDER_VALID) &&
	  (obj->flags & XOTCL_MIXIN_ORDER_DEFINED)) {
	XOTclClass     *mcl=0;
	Tcl_ObjCmdProc *mproc = 0;
	Command        *mProcInfo =0;
	ClientData     mcp;
	Command* currentCmdPtr;
	MixinStackPush(obj);	
	mixinStackPushed = 1;

	mProcInfo = MixinSearchProc(in, obj, methodName,
				    &mcl, &mproc, &mcp, &currentCmdPtr);

	if (mproc) {
	  if (!isFilterEntry) {
	    obj->mixinStack->currentCmdPtr = currentCmdPtr;
	    isMixinEntry = 1;
	    cl = mcl;
	    proc = mproc;
	    cp = mcp;
	    procInfo = mProcInfo;
	    obj->mixinStack->mixinChainOn = 1;
	  }
	} else {
	  MixinStackPop(obj);
	  mixinStackPushed = 0;
	}
      }
#ifdef AUTOVARS
    }
#endif

    /* if no filter/mixin is found => do ordinary method lookup */
    if (proc == 0) {
      if (obj->nsPtr)
	procInfo = FindMethod(methodName, obj->nsPtr);

      cl = procInfo ? 0 : SearchCMethod(obj->cl, NULL, methodName, &procInfo);
      if (procInfo) {
	proc = procInfo->objProc;
	cp   = procInfo->objClientData;
      } else {
	assert(cp == 0);
      }
    }

#if 0
    if (cl) {
      clName = cl->object.cmdName;
      INCR_REF_COUNT(clName);
    }
#endif

    if (proc) {
      if (cp == 0 || TclIsProc(procInfo)) {
	if (cp == 0)
	  cp = cd;
	result = callProcCheck(cp, in, objc-1, objv+1, procInfo,
			       obj, cl, callMethod, isFilterEntry,
			       isMixinEntry, &isDestroyed);
      } else {
	/* there is method specific client data for the method =>
	   another dispatch mechanism handles this call */

	/*
	fprintf(stderr, "Propagating Dispatch of :\n");
	printObjv(objc, objv);
	*/

	result = callProcCheck(cp, in, objc, objv, procInfo,
			       obj, cl, callMethod, isFilterEntry,
			       isMixinEntry, &isDestroyed);
      }

      if (result == TCL_ERROR) 
	XOTclErrInProc(in, objName, cl?cl->object.cmdName:NULL, callMethod);
      
    } else if (XOTclObjectIsClass(obj) &&
	       (flags & XOTCL_CM_NO_UNKNOWN)) {
      Tcl_AppendResult(in, ObjStr(objv[0]),
		       ": unable to dispatch method '",
		       callMethod, "'", 0);
      result = TCL_ERROR;
    } else {
      result = XOTCL_UNKNOWN;
    }
    DECR_REF_COUNT(objName);
#if 0
    if (clName) {
      DECR_REF_COUNT(clName);
    }
#endif

#ifdef DISPATCH_TRACE
    printExit(in,"DISPATCH", objc,objv, result);
#endif

  if (!isDestroyed && obj && mixinStackPushed && obj->mixinStack) {
    MixinStackPop(obj);
  }
  if (!isDestroyed && obj && filterStackPushed && obj->filterStack) {
    FilterStackPop(obj);
  }

  return result;
}

static int
ObjDispatch(ClientData cd, Tcl_Interp* in,
	    int objc, Tcl_Obj *CONST objv[],
	    int flags) {
  int result;

#ifdef STACK_TRACE
  XOTclStackTrace(in);
#endif

#ifdef CALLSTACK_TRACE
  XOTclCallStackTrace(in);
#endif

  if (objc < 2) {
    XOTclObject *obj = (XOTclObject*)cd;
    Tcl_SetObjResult(in, obj->cmdName);
    return TCL_OK;
  /*return XOTclObjErrArgCnt(in, objv[0], "message ?args...?");*/
  }

  /* try normal dispatch */
  result = DoDispatch(cd, in, objc, objv, flags);
  if (result != XOTCL_UNKNOWN) {
    return result;
  } else {
    /*
     * back off and try unknown;
     */
    XOTclObject *obj = (XOTclObject*)cd;
    DEFINE_NEW_TCL_OBJECTS_ON_STACK(objc+1, tov);

    /*
    fprintf(stderr,"calling unknwon for %s %s ...\n",
	    ObjStr(obj->cmdName), ObjStr(objv[1]));
    */

    tov[0] = obj->cmdName;
    tov[1] = global_objects[UNKNOWN];
    if (objc>1)
      memcpy(tov+2, objv+1, sizeof(Tcl_Obj *)*(objc-1));

    result = DoDispatch(cd, in, objc+1, tov, flags);
    FREE_TCL_OBJECTS_ON_STACK(tov);

    if (result != XOTCL_UNKNOWN)
      return result;
  }
  /*
   * and if that fails too, error out
   */
  Tcl_ResetResult(in);
  Tcl_AppendResult(in, ObjStr(objv[0]),
                   ": unable to dispatch method '", ObjStr(objv[1]), "'", 0);
  return TCL_ERROR;
}

int
XOTclDirectSelfDispatch(ClientData cd, Tcl_Interp* in, int objc, Tcl_Obj *CONST objv[]) {
  int result;
#ifdef XOTCLOBJ_TRACE
  XOTclObject *obj;
#endif
  objTrace("BEFORE SELF DISPATCH", obj);
  result =  ObjDispatch((ClientData)GetSelfObj(in), in, objc, objv, 0);
  objTrace("AFTER SELF DISPATCH", obj);
  return result;
}

int
XOTclObjDispatch(ClientData cd, Tcl_Interp* in, int objc, Tcl_Obj *CONST objv[]) {
  return ObjDispatch(cd, in, objc, objv, 0);
}

/*
 *  Proc-Creation
 */

static int
MakeProc(Tcl_Namespace* ns, XOTclAssertionStore* aStore,
	 Tcl_Interp* in, int objc, Tcl_Obj* objv[]) {
    int result, oc = objc;
    Tcl_CallFrame frame;
    Tcl_Obj *oldBody;
    char *body;
#ifdef AUTOVARS
    char *p;
#endif
    oldBody = objv[3];
    body = ObjStr(oldBody);

    objv[3] = Tcl_NewStringObj("", 0);
    INCR_REF_COUNT(objv[3]);

    Tcl_AppendStringsToObj(objv[3], "::xotcl::initProcNS\n", NULL);

#ifdef AUTOVARS
    if ((p = strstr(body, "self")) && p != body && *(p-1) != '[')
      Tcl_AppendStringsToObj(objv[3], "::set self [self]\n", NULL);
    if (strstr(body, "proc"))
      Tcl_AppendStringsToObj(objv[3], "::set proc [self proc]\n", NULL);
    if (strstr(body, "class"))
      Tcl_AppendStringsToObj(objv[3], "::set class [self class]\n", NULL);
#endif

    Tcl_AppendStringsToObj(objv[3], body, NULL);

    Tcl_PushCallFrame(in,&frame,ns,0);

    if (objc > 4) oc = 4;
    result = Tcl_ProcObjCmd(0, in, oc, objv) != TCL_OK;

    Tcl_PopCallFrame(in);

    if (objc == 6)
      AssertionAddProc(in, ObjStr(objv[1]), aStore, objv[4], objv[5]);

    /*#ifdef AUTOVARS*/
    DECR_REF_COUNT(objv[3]);
    objv[3] = oldBody;
    /*#endif*/

    return result;
}

/*
 * List-Functions for Info
 */
static int
ListInfo(Tcl_Interp* in, int isclass) {
  Tcl_ResetResult(in);
  Tcl_AppendElement(in, "vars"); Tcl_AppendElement(in, "body");
  Tcl_AppendElement(in, "default"); Tcl_AppendElement(in, "args");
  Tcl_AppendElement(in, "procs"); Tcl_AppendElement(in, "commands");
  Tcl_AppendElement(in, "class"); Tcl_AppendElement(in, "children");
  Tcl_AppendElement(in, "filter"); Tcl_AppendElement(in, "filterguard");
  Tcl_AppendElement(in, "info");
  Tcl_AppendElement(in, "invar"); Tcl_AppendElement(in, "mixin");
  Tcl_AppendElement(in, "parent");
  Tcl_AppendElement(in, "pre"); Tcl_AppendElement(in, "post");
  if (isclass) {
    Tcl_AppendElement(in, "superclass"); Tcl_AppendElement(in, "subclass");
    Tcl_AppendElement(in, "heritage"); Tcl_AppendElement(in, "instances");
    Tcl_AppendElement(in, "instcommands"); Tcl_AppendElement(in, "instprocs");
    Tcl_AppendElement(in, "instdefault"); Tcl_AppendElement(in, "instbody");
    Tcl_AppendElement(in, "instmixin");
    Tcl_AppendElement(in, "classchildren"); Tcl_AppendElement(in, "classparent");
    Tcl_AppendElement(in, "instfilter"); Tcl_AppendElement(in, "instfilterguard");
    Tcl_AppendElement(in, "instinvar");
    Tcl_AppendElement(in, "instpre"); Tcl_AppendElement(in, "instpost");
    Tcl_AppendElement(in, "parameter");
  }
  return TCL_OK;
}

static XOTCLINLINE int
noMetaChars(char *pattern) {
  register char c, *p = pattern;
  assert(pattern);
  for (c=*p; c; c = *++p) {
    if (c == '*' || c == '[') {
      return 0;
    }
  }
  return 1;
}

static int
ListKeys(Tcl_Interp* in, Tcl_HashTable* table, char* pattern) {
  Tcl_HashEntry* hPtr;
  char *key;

  if (pattern && noMetaChars(pattern)) {
    hPtr = table ? Tcl_FindHashEntry(table, pattern) : 0;
    if (hPtr) {
      key = Tcl_GetHashKey(table, hPtr);
      Tcl_SetResult(in, key, TCL_VOLATILE);
    } else {
      Tcl_SetObjResult(in, global_objects[EMPTY]);
    }
  } else {
    Tcl_Obj *list = Tcl_NewListObj(0, NULL);
    Tcl_HashSearch hSrch;
    hPtr = table ? Tcl_FirstHashEntry(table, &hSrch) : 0;
    for (; hPtr; hPtr = Tcl_NextHashEntry(&hSrch)) {
      key = Tcl_GetHashKey(table, hPtr);
      if (!pattern || Tcl_StringMatch(key, pattern)) {
	Tcl_ListObjAppendElement(in, list, Tcl_NewStringObj(key,-1));
      }
    }
    Tcl_SetObjResult(in, list);
  }
  return TCL_OK;
}

static int
varExists(Tcl_Interp* in, XOTclObject* obj, char *varName) {
  VarFrameDecls;
  Var *varPtr;
  int result;
#if defined(PRE83)
  Var *arrayPtr;
#endif

  VarFrameSwitchToObj(in, obj);

#if defined(PRE83)
  varPtr = TclLookupVar(in, varName, (char *) NULL, TCL_PARSE_PART1, "access",
			/*createPart1*/ 0, /*createPart2*/ 0, &arrayPtr);
#else
  varPtr = TclVarTraceExists(in, varName);
#endif
  result = ((varPtr != NULL) && !TclIsVarUndefined(varPtr));

  VarFrameRestore(in, obj);

  return result;
}

static int
ListVars(Tcl_Interp* in, XOTclObject* obj, char* pattern) {
  Tcl_Obj *varlist, *okList, *element;
  int i, length;
  Namespace* nsp = (Namespace*) obj->nsPtr;

  if (nsp) {
    ListKeys(in, &nsp->varTable, pattern);
  } else if (obj->varTable) {
    ListKeys(in, obj->varTable, pattern);
  } else {
    return TCL_OK;
  }
  varlist = Tcl_GetObjResult(in);
  Tcl_ListObjLength(in, varlist, &length);
  okList = Tcl_NewListObj(0, (Tcl_Obj **) NULL);
  for (i=0; i<length; i++) {
    Tcl_ListObjIndex(in, varlist, i, &element);
    if (varExists(in, obj, ObjStr(element))) {
      Tcl_ListObjAppendElement(in, okList, element);
    } else {
      /*fprintf(stderr,"must ignore '%s' %d\n", ObjStr(element),i);*/
      /*Tcl_ListObjReplace(in, varlist, i, 1, 0, NULL);*/
    }
  }
  Tcl_SetObjResult(in, okList);
  return TCL_OK;
}

static int
ListObjPtrHashTable(Tcl_Interp* in, Tcl_HashTable* table, char* pattern) {
  Tcl_HashEntry* hPtr;
  if (pattern && noMetaChars(pattern)) {
    XOTclObject* childobj = GetObject(in,  pattern);
    hPtr = Tcl_FindHashEntry(table, (char*)childobj);
    if (hPtr) {
      Tcl_SetObjResult(in, childobj->cmdName);
    } else {
      Tcl_SetObjResult(in, global_objects[EMPTY]);
    }
  } else {
    Tcl_Obj *list = Tcl_NewListObj(0, NULL);
    Tcl_HashSearch hSrch;
    hPtr = table ? Tcl_FirstHashEntry(table, &hSrch) : 0;
    for (;  hPtr; hPtr = Tcl_NextHashEntry(&hSrch)) {
      XOTclObject* obj = (XOTclObject*)Tcl_GetHashKey(table, hPtr);
      if (!pattern || Tcl_StringMatch(ObjStr(obj->cmdName), pattern)) {
	Tcl_ListObjAppendElement(in, list, obj->cmdName);
      }
    }
    Tcl_SetObjResult(in, list);
  }
  return TCL_OK;
}

static int
ListMethodKeys(Tcl_Interp* in, Tcl_HashTable* table, char* pattern,
	       int noProcs, int noCmds) {
  Tcl_HashSearch hSrch;
  Tcl_HashEntry* hPtr = table ? Tcl_FirstHashEntry(table, &hSrch) : 0;
  for (; hPtr != 0; hPtr = Tcl_NextHashEntry(&hSrch)) {
    char* key = Tcl_GetHashKey(table, hPtr);
    Tcl_ObjCmdProc* proc = ((Command*)Tcl_GetHashValue(hPtr))->objProc;

    if (pattern && !Tcl_StringMatch(key, pattern)) continue;
    if (noCmds  && proc != RUNTIME_STATE(in)->objInterpProc) continue;
    if (noProcs && proc == RUNTIME_STATE(in)->objInterpProc) continue;
    Tcl_AppendElement(in, key);
  }
  return TCL_OK;
}

static int
ListMethods(Tcl_Interp* in, XOTclObject* obj, char* pattern,
	    int noProcs, int noCmds, int noMixins) {
  Namespace *nsp = (Namespace *)obj->nsPtr;
  XOTclClasses* pl;
  if (nsp)
    ListMethodKeys(in, &nsp->cmdTable, pattern, noProcs, noCmds);

  if (!noMixins) {
    if (!(obj->flags & XOTCL_MIXIN_ORDER_VALID))
      MixinComputeDefined(in, obj);
    if ((obj->flags & XOTCL_MIXIN_ORDER_VALID) &&
	(obj->flags & XOTCL_MIXIN_ORDER_DEFINED)) {
      XOTclCmdList *ml = obj->mixinOrder;
      XOTclClass *mixin;
      while (ml) {
	mixin = GetClassFromCmdPtr(ml->cmdPtr);
	if (mixin)
	  ListMethodKeys(in, &((Namespace *)mixin->nsPtr)->cmdTable, pattern,
		       noProcs, noCmds);
	ml = ml->next;
      }
    }
  }

  /* append per-class filters */
  if (!obj->cl->order)
    obj->cl->order = ComputeOrder(obj->cl, Super);
  pl = obj->cl->order;
  while (pl) {
    ListMethodKeys(in, &((Namespace *)pl->cl->nsPtr)->cmdTable, pattern,
		   noProcs, noCmds);
    pl = pl->next;
  }
  return TCL_OK;
}

static int XOTclCInfoMethod(ClientData d, Tcl_Interp* h, int i,
			    Tcl_Obj *CONST v[]);

static int
ListClass(Tcl_Interp* in, XOTclObject* obj, char* pattern,
	  int objc, Tcl_Obj* CONST objv[]) {
  if (pattern == 0) {
    Tcl_SetObjResult(in, obj->cl->object.cmdName);
    return TCL_OK;
  } else {
    int result;
    DEFINE_NEW_TCL_OBJECTS_ON_STACK(objc, ov);

    memcpy(ov, objv, sizeof(Tcl_Obj *)*objc);
    ov[1] = Tcl_NewStringObj("superclass", 10);
    INCR_REF_COUNT(ov[1]);
    result = XOTclCInfoMethod((ClientData)obj->cl, in, objc, ov);
    DECR_REF_COUNT(ov[1]);
    FREE_TCL_OBJECTS_ON_STACK(ov);
    return result;
  }
}

static int
ListSuperclasses(Tcl_Interp* in, XOTclClass* cl, char* pattern) {
  if (pattern == 0) {
    XOTclClasses* sl = cl->super;
    XOTclClasses* sc = 0;

    /*
     * reverse the list to obtain presentation order
     */

    Tcl_ResetResult(in);
    while (sc != sl) {
      XOTclClasses* nl = sl;
      while (nl->next != sc) nl = nl->next;
      Tcl_AppendElement(in, className(nl->cl));
      sc = nl;
    }
  } else {
    XOTclClass* isc = GetClass(in, pattern);
    XOTclClasses* pl;
    if (isc == 0) return XOTclErrBadVal(in, "a class", pattern);
    pl = ComputeOrder(cl, Super);

    /*
     * search precedence to see if we're related or not
     */
    while (pl != 0) {
      if (pl->cl == isc) {
	Tcl_SetIntObj(Tcl_GetObjResult(in), 1);
	break;
      }
      pl = pl->next;
    }
    if (pl == 0)
      Tcl_SetIntObj(Tcl_GetObjResult(in), 0);
  }
  return TCL_OK;
}

static int
ListSubclasses(Tcl_Interp* in, XOTclClass* cl, char* pattern) {
  if (pattern == 0) {
    XOTclClasses* sl = cl->sub;
    XOTclClasses* sc = 0;

    /*
     * order unimportant
     */
    Tcl_ResetResult(in);
    for (sc = sl; sc != 0; sc = sc->next)
      Tcl_AppendElement(in, className(sc->cl));
  } else {
    XOTclClass* isc = GetClass(in, pattern);
    XOTclClasses* pl;
    XOTclClasses* saved;
    if (isc == 0) return XOTclErrBadVal(in, "a class", pattern);
    saved = cl->order; cl->order = 0;
    pl = ComputeOrder(cl, Sub);

    /*
     * search precedence to see if we're related or not
     */
    while (pl != 0) {
      if (pl->cl == isc) {
	Tcl_SetIntObj(Tcl_GetObjResult(in), 1);
	break;
      }
      pl = pl->next;
    }
    if (pl == 0)
      Tcl_SetIntObj(Tcl_GetObjResult(in), 0);

    XOTclRemoveClasses(cl->order); cl->order = saved;
  }
  return TCL_OK;
}



static int
ListHeritage(Tcl_Interp* in, XOTclClass* cl, char* pattern) {
  XOTclClasses* pl = ComputeOrder(cl, Super);
  if (pl) pl = pl->next;
  Tcl_ResetResult(in);
  for (; pl != 0; pl = pl->next) {
    char* name = className(pl->cl);
    if (pattern && !Tcl_StringMatch(name, pattern)) continue;
    Tcl_AppendElement(in, name);
  }
  return TCL_OK;
}

static Proc*
FindProc(Tcl_Interp* in, Tcl_HashTable* table, char* name) {
  Tcl_HashEntry* hPtr = table ? Tcl_FindHashEntry(table, name) : 0;
  if (hPtr) {
    Command* co = (Command*)Tcl_GetHashValue(hPtr);
    if (co->objProc == RUNTIME_STATE(in)->objInterpProc)
      return (Proc*) co->objClientData;
    else if (co->proc == RUNTIME_STATE(in)->interpProc)
      return (Proc*) co->clientData;
  }
  return 0;
}

static int
ListProcArgs(Tcl_Interp* in, Tcl_HashTable* table, char* name) {
  Proc* proc = FindProc(in, table, name);
  if (proc) {
      CompiledLocal *args = proc->firstLocalPtr;
      Tcl_ResetResult(in);
      for (;args != NULL; args = args->nextPtr) {
          if TclIsVarArgument(args)
	      Tcl_AppendElement(in, args->name);

      }
      return TCL_OK;
  }
  return XOTclErrBadVal(in, "a tcl method name", name);
}

static int
GetProcDefault(Tcl_Interp* in, Tcl_HashTable* table,
	       char* name, char* arg, Tcl_Obj** resultObj) {
  Proc* proc = FindProc(in, table, name);
  *resultObj = 0;
  if (proc) {
    CompiledLocal *ap;
    for (ap = proc->firstLocalPtr; ap != 0; ap = ap->nextPtr) {
      if (!TclIsVarArgument(ap)) continue;
      if (strcmp(arg, ap->name) != 0) continue;
	
      if (ap->defValuePtr != NULL) {
	*resultObj = ap->defValuePtr;
	return TCL_OK;
      }
      return TCL_OK;
    }
  }
  return TCL_ERROR;
}

static int
ListProcDefault(Tcl_Interp* in, Tcl_HashTable* table,
		 char* name, char* arg, char* var) {
  Tcl_Obj* defVal;
  if (GetProcDefault(in, table, name, arg, &defVal) == TCL_OK) {
    if (defVal != 0) {
      if (Tcl_SetVar2(in, var, 0, ObjStr(defVal), 0) != NULL) {
	Tcl_SetIntObj(Tcl_GetObjResult(in), 1);
	return TCL_OK;
      }
    } else {
      if (Tcl_SetVar2(in, var, 0, "", 0) != NULL) {
	Tcl_SetIntObj(Tcl_GetObjResult(in), 0);
	return TCL_OK;
      }
    }
    Tcl_ResetResult(in);
    Tcl_AppendResult
      (in, "couldn't store default value in variable '",
       var, "'", (char *) 0);
    return TCL_ERROR;
  }
  Tcl_ResetResult(in);
  Tcl_AppendResult(in, "procedure '", name,
		   "' doesn't exist or doesn't have an argument '",
		   arg, "'", (char *) 0);
  return TCL_ERROR;
}

static int
ListProcBody(Tcl_Interp* in, Tcl_HashTable* table, char* name) {
  Proc* proc = FindProc(in, table, name);

  if (proc) {
    char* body = ObjStr(proc->bodyPtr);
    if (strncmp(body, "::xotcl::initProcNS\n",20) == 0)
      body+=20;
    Tcl_SetObjResult(in, Tcl_NewStringObj(body, -1));
    return TCL_OK;
  }
  return XOTclErrBadVal(in, "a tcl method name", name);
}


static int
ListObjChildren(Tcl_Interp* in, XOTclObject* obj, char* pattern) {
  XOTclObject* childobj;
  Namespace *nsp = (Namespace *)obj->nsPtr;
  Tcl_HashTable* table;
  VarFrameDecls;

  if (!nsp) return TCL_OK;

  table = &nsp->cmdTable;

  if (pattern && noMetaChars(pattern)) {
    VarFrameSwitchToObj(in, obj);
    if ((childobj = GetObject(in, pattern))) {
      Tcl_SetObjResult(in, childobj->cmdName);
    } else {
      Tcl_SetObjResult(in, global_objects[EMPTY]);
    }
    VarFrameRestore(in,obj);
  } else {
    Tcl_Obj *list = Tcl_NewListObj(0, NULL);
    Tcl_HashSearch hSrch;
    Tcl_HashEntry* hPtr = table ? Tcl_FirstHashEntry(table, &hSrch) : 0;
    VarFrameSwitchToObj(in, obj);
    for (; hPtr != 0; hPtr = Tcl_NextHashEntry(&hSrch)) {
      char* key = Tcl_GetHashKey(table, hPtr);
      if (!pattern || Tcl_StringMatch(key, pattern)) {
	if ((childobj = GetObject(in, key))) {
	  Tcl_ListObjAppendElement(in, list, childobj->cmdName);
	}
      }
    }
    VarFrameRestore(in,obj);
    Tcl_SetObjResult(in, list);
  }
  return TCL_OK;
}

static int
ListClassChildren(Tcl_Interp* in, XOTclClass* cl, char* pattern) {
  XOTclObject* obj = &cl->object, *childobj;
  Tcl_CallFrame frame;
  Tcl_HashSearch hSrch;
  Tcl_HashEntry* hPtr;
  Tcl_Obj *list;
  Namespace *nsp = (Namespace *)obj->nsPtr;

  if (!nsp) return TCL_OK;

  hPtr = &nsp->cmdTable ?
    Tcl_FirstHashEntry(&nsp->cmdTable, &hSrch) :
    0;
  list = Tcl_NewListObj(0, NULL);
  for (; hPtr != 0; hPtr = Tcl_NextHashEntry(&hSrch)) {
    char* key = Tcl_GetHashKey(&nsp->cmdTable, hPtr);
    if (!pattern || Tcl_StringMatch(key, pattern)) {
	Tcl_PushCallFrame(in, &frame, (Tcl_Namespace*)obj->nsPtr,0);
	childobj = GetObject(in, key);
	if (XOTclObjectIsClass(childobj)) {
	    Tcl_ListObjAppendElement(in, list, childobj->cmdName);
	}
	Tcl_PopCallFrame(in);
    }
  }
  Tcl_SetObjResult(in, list);
  return TCL_OK;
}

static int
ListParent(Tcl_Interp* in, XOTclObject* obj) {
  if (obj->id) {
    Command *cmdPtr = (Command *) obj->id;
    if (cmdPtr->nsPtr) {
      Tcl_SetResult(in, cmdPtr->nsPtr->fullName,
		    TCL_VOLATILE);
    }
  }
  return TCL_OK;
}

static int
XOTcl_UplevelObjCmd(ClientData cd, Tcl_Interp *in, int objc, Tcl_Obj *CONST objv[]);

static void
FindInCallStack(Tcl_Interp* in, Tcl_Obj* searchElt) {
  Interp *iPtr = (Interp *) in;
  int i, result = TCL_ERROR;
  Tcl_Obj* ov[2];
  ov[0] = Tcl_NewIntObj(1);
  ov[1] = searchElt;
  INCR_REF_COUNT(ov[0]);

  for (i=1; i < iPtr->numLevels && result != TCL_OK; i++) {
    Tcl_SetIntObj(ov[0], i);
    result = XOTcl_UplevelObjCmd(NULL, in, 2, ov);
  }
  DECR_REF_COUNT(ov[0]);

  if (result != TCL_OK)
    Tcl_SetObjResult(in, global_objects[EMPTY]);
}

static XOTclClass*
FindCalledClass(Tcl_Interp* in, XOTclObject* obj) {
  char* methodName = 0;
  XOTclClass *cl;
  Command *procInfo = NULL;
  XOTclCallStackContent *csc = CallStackGetFrame(in);

  if (csc && !isFilterEntry(csc) &&
      !csc->cscType == XOTCL_CSC_TYPE_MIXIN)
    return GetSelfClass(in);

  if (isFilterEntry(csc))
    methodName = ObjStr(csc->filterStackEntry->calledProc);
  else if (csc->cscType == XOTCL_CSC_TYPE_MIXIN && obj->mixinStack)
    methodName = (char*) GetSelfProc(in);

  if (!methodName) methodName = "";

  if (obj->nsPtr)
    procInfo = FindMethod(methodName, obj->nsPtr);

  if (procInfo) {
    cl = 0;
  } else {
    cl = SearchCMethod(obj->cl, NULL, methodName, &procInfo);
  }
  return cl;
}

/*
 * Next Primitive Handling
 */
static XOTCLINLINE void
NextSearchMethod(XOTclObject* obj, Tcl_Interp* in, XOTclClass **cl,
		 char **method, Tcl_ObjCmdProc **proc, Command **procInfo,
		 ClientData *cp,
		 int* isMixinEntry, int* turnMixinsOff,
		 int* isFilterEntry, int* turnFiltersOff,
		 Command** currentCmdPtr) {
  XOTclClasses *pl = 0;

  /*
   *  Next in filters
   */
  if (!(obj->flags & XOTCL_FILTER_ORDER_VALID))
    FilterComputeDefined(in, obj);

  if ((obj->flags & XOTCL_FILTER_ORDER_VALID) &&
      obj->filterStack &&
      obj->filterStack->currentCmdPtr) {
    *procInfo = FilterSearchProc(in, obj, proc, cp, currentCmdPtr);
    if (*proc == 0) {
      /*if (obj->filterStack->filterChainOn) fprintf(stderr, "chainon\n");*/
      if (obj->filterStack->filterChainOn) {
	*turnFiltersOff = 1;
	/* reset the information to the values of method, cl
	   to the values they had before calling the filters */
	*method = ObjStr(obj->filterStack->calledProc);
	*cl = 0;
      }
    } else {
      *method = (char*) Tcl_GetCommandName(in, (Tcl_Command)(*procInfo));
      *cl = GetClass(in, NSCutXOTclClasses((*procInfo)->nsPtr->fullName));
      *isFilterEntry = 1;
      return;
    }
  }
  /*
   *  Next in Mixins
   */
  if (!(obj->flags & XOTCL_MIXIN_ORDER_VALID))
    MixinComputeDefined(in, obj);

  if ((obj->flags & XOTCL_MIXIN_ORDER_VALID) &&
      obj->mixinStack) {
    *procInfo = MixinSearchProc(in, obj, *method,
				cl, proc, cp, currentCmdPtr);
    if (*proc == 0) {
      if (obj->mixinStack->mixinChainOn) {
	*turnMixinsOff = 1;
	*cl = 0;
      }
    } else {
      *isMixinEntry = 1;
      return;
    }
  }

  /*
   * otherwise: normal method dispatch
   *
   * if we are already in the precedence ordering, then advance
   * past our last point; otherwise (if cl==0) begin from the start
   */

  /* if a mixin or filter chain has ended -> we have to search
     the obj-specific methods as well */
  if (obj->nsPtr && (*turnMixinsOff || *turnFiltersOff)) {
    *procInfo = FindMethod(*method, obj->nsPtr);
  } else {
    *procInfo = 0;
  }

  if (!*procInfo) {
    pl = ComputeOrder(obj->cl, Super);
    while (pl && *cl) {
      if (pl->cl == *cl) *cl = 0;
      pl = pl->next;
    }

    /*
     * search for a further class method
     */
    *cl = SearchCMethod(NULL, pl, *method, procInfo);
  } else {
    *cl = 0;
  }

  if (*procInfo) {
    *proc = (*procInfo)->objProc;
    *cp   = (*procInfo)->objClientData;
  }

  return;
}

static int
XOTclNextMethod(XOTclObject* obj, Tcl_Interp* in, XOTclClass *givenCl,
		char *givenMethod, int objc, Tcl_Obj *CONST objv[],
		int useCallstackObjs,  int* isDestroyed) {
  XOTclCallStackContent *csc = CallStackGetFrame(in);
  Tcl_ObjCmdProc *proc = 0;
  Command *procInfo;
  ClientData cp = 0;
  int result = TCL_OK,
    isMixinEntry = 0, turnMixinsOff = 0,
    isFilterEntry = 0, turnFiltersOff = 0;
  int nobjc; Tcl_Obj **nobjv;
  Command* currentCmdPtr = 0;
  XOTclClass** cl = &givenCl;
  char** method = &givenMethod;

  /* if no args are given => use args from stack */
  if (objc < 2 && useCallstackObjs) {
    nobjc = csc->objc;
    nobjv = (Tcl_Obj **)csc->objv;
  } else {
    nobjc = objc;
    nobjv = (Tcl_Obj **)objv;
  }

  /*
   * Search the next method & compute its method data
   */
  NextSearchMethod(obj, in, cl, method, &proc, &procInfo, &cp,
		   &isMixinEntry, &turnMixinsOff,
		   &isFilterEntry, &turnFiltersOff,
		   &currentCmdPtr);
  /*
  fprintf(stderr, "NextSearchMethod -- RETURN: method=%s", *method);
  if (obj)
    fprintf(stderr, " obj=%s,", ObjStr(obj->cmdName));
  if ((*cl))
    fprintf(stderr, " cl=%s,", (*cl)->nsPtr->fullName);
  fprintf(stderr, " mixin=%d, filter=%d, proc=%p\n",
	  isMixinEntry, isFilterEntry, proc);
  */
  /*
   * change mixin state
   */
  if (obj->mixinStack) {
    /* if we're at the end of a mixin chain -> turn mixin chain off */
    if (turnMixinsOff && obj->mixinStack->mixinChainOn)
      obj->mixinStack->mixinChainOn = 0;
    /* otherwise move the command pointer forward */
    if (isMixinEntry) {
      obj->mixinStack->currentCmdPtr = currentCmdPtr;
      /* be sure that mixinChainOn is 1 ... mixins may be turned on
	 after the filter chain has ended (when mixins + filters are
	 defined)
      */
      obj->mixinStack->mixinChainOn = 1;
    }
  }
  /*
   * change filter state
   */
  if (obj->filterStack) {
    if (csc->cscType == XOTCL_CSC_TYPE_ACTIVE_FILTER)
      csc->cscType = XOTCL_CSC_TYPE_INACTIVE_FILTER;

    /* if we're at the end of a filter chain -> turn filter chain off */
    if (turnFiltersOff && obj->filterStack->filterChainOn) {
      obj->filterStack->filterChainOn = 0;
    }
    /* otherwise move the command pointer forward */
    if (isFilterEntry) {
      obj->filterStack->currentCmdPtr = currentCmdPtr;
    }
  }

  Tcl_ResetResult(in); /* needed for bytecode support */

  /*
   * now actually call the "next" method
   */
  if (proc != 0) {
    cp = (cp != 0) ? cp : (ClientData)obj;

    /* cut the flag, that no stdargs should be used, if it is there */
    if (nobjc > 1) {
      char* nobjv1 = ObjStr(nobjv[1]);
      if (nobjv1[0] == '-' && !strcmp(nobjv1, "--noArgs"))
	nobjc = 1;
    }
    result = callProcCheck(cp, in, nobjc, nobjv, procInfo,
			   obj, *cl, *method, isFilterEntry, isMixinEntry,
			   isDestroyed);

    if (csc->cscType == XOTCL_CSC_TYPE_INACTIVE_FILTER)
      csc->cscType = XOTCL_CSC_TYPE_ACTIVE_FILTER;
  } else if (result == TCL_OK && turnFiltersOff) {
    result = XOTCL_UNKNOWN;
  }
  return result;
}

int
XOTclNextObjCmd(ClientData cd, Tcl_Interp *in, int objc, Tcl_Obj *CONST objv[]) {
  XOTclCallStackContent *csc = CallStackGetFrame(in);
  int isDestroyed;

  if (!csc->self)
    return XOTclVarErrMsg(in, "next: can't find self", (char*) NULL);

  if (!csc->cmdPtr)
    return XOTclErrMsg(in, "next: no executing proc", TCL_STATIC);

  return XOTclNextMethod(csc->self, in, csc->cl,
			 (char*) Tcl_GetCommandName(in, (Tcl_Command) csc->cmdPtr),
			 objc, objv, 1, &isDestroyed);
}

/*
 * "self" object command
 */

static int
FindSelfNext(Tcl_Interp *in, XOTclObject* obj) {
  XOTclCallStackContent *csc = CallStackGetFrame(in);
  Tcl_ObjCmdProc *proc = 0;
  Command *procInfo;
  ClientData cp = 0;
  int isMixinEntry = 0, turnMixinsOff = 0,
    isFilterEntry = 0, turnFiltersOff = 0;
  Command* currentCmdPtr = 0;
  XOTclClass *cl = csc->cl;
  XOTclObject* o = csc->self;
  char *method = (char*)
    Tcl_GetCommandName(in, (Tcl_Command) csc->cmdPtr);

  Tcl_ResetResult(in);

  method = (char*) GetSelfProc(in);
  if (!(method))
    return TCL_OK;

  NextSearchMethod(o, in, &cl, &method, &proc, &procInfo, &cp,
		   &isMixinEntry, &turnMixinsOff,
		   &isFilterEntry, &turnFiltersOff,
		   &currentCmdPtr);

  if (procInfo) {
    Tcl_SetObjResult(in, getFullProcQualifier(in,
      Tcl_GetCommandName(in, (Tcl_Command)procInfo), o, cl));
  }
  return TCL_OK;
}

int
XOTclGetSelfObjCmd(ClientData cd, Tcl_Interp *in, int objc, Tcl_Obj *CONST objv[]) {
  XOTclObject* obj;
  char *option;
  CallFrame *savedFramePtr = 0, *savedVarFramePtr = 0;
  int filterFrameCutOff = 0;

  if (objc < 2) {
    if ((obj = GetSelfObj(in))) {
      /*
	fprintf(stderr, "obj=%p, cd=%p <%s> <%s>\n", obj,cd,
	  ObjStr((obj->cmdName)),
	  ObjStr(((XOTclObject*)cd)->cmdName));
      */
      Tcl_SetObjResult(in, obj->cmdName);
      return TCL_OK;
    }
    else
      return XOTclVarErrMsg(in, "self: no current object", (char*) NULL);
  } else if (objc != 2) {
    return XOTclVarErrMsg(in, "wrong # of args for self",
			  (char*)NULL);
  }

  option = ObjStr(objv[1]);
  if (isProcString(option)) { /* proc subcommand */
    char *procName = (char*) GetSelfProc(in);
    if (procName) {
      Tcl_SetResult(in, procName, TCL_VOLATILE);
      return TCL_OK;
    } else
      return XOTclVarErrMsg(in, "Can't find proc", (char*) NULL);
  } else if (isClassString(option)) { /* class subcommand */
    XOTclClass *cl = GetSelfClass(in);
    if (cl)
      Tcl_SetObjResult(in, cl->object.cmdName);
    else
      Tcl_SetObjResult(in, global_objects[EMPTY]);
    return TCL_OK;
  } else {
    XOTclCallStackContent *csc = CallStackGetFrame(in);
    obj = csc->self;
    if (!obj)
      return  XOTclVarErrMsg(in, "Can't find self during 'self ",
			     *option, "'", (char*) NULL);
    switch (*option) { /* other callstack information */
    case 'c':
      if (!strcmp(option, "calledproc")) {
	if (!isFilterEntry(csc))
	  return XOTclVarErrMsg(in,
		   "self calledproc called from outside of a filter",
		   (char*) NULL);
	Tcl_SetObjResult(in, csc->filterStackEntry->calledProc);
	return TCL_OK;
      } else if (!strcmp(option, "calledclass")) {
	Tcl_ResetResult(in);
	Tcl_AppendResult(in, className(FindCalledClass(in, obj)),
			 (char*) NULL);
	return TCL_OK;
      } else if (!strcmp(option, "callingproc")) {
	if (isFilterEntry(csc)) {
	  Tcl_SetObjResult(in, csc->filterStackEntry->callingProc);
	} else {
	  if (FilterEntryOnCallStack(in)) {
	    filterFrameCutOff =
	      FilterCutFrames(in, &savedFramePtr, &savedVarFramePtr);
	  }
	  FindInCallStack(in, global_objects[SELF_PROC]);
	  if (filterFrameCutOff)
	    FilterRestoreFrames(in, savedFramePtr, savedVarFramePtr);
	}
	return TCL_OK;
      } else if (!strcmp(option, "callingclass")) {
	if (isFilterEntry(csc)) {
	  Tcl_SetObjResult(in, csc->filterStackEntry->callingClass);
	} else {
	  if (FilterEntryOnCallStack(in)) {
	    filterFrameCutOff =
	      FilterCutFrames(in, &savedFramePtr, &savedVarFramePtr);
	  }
	  FindInCallStack(in, global_objects[SELF_CLASS]);
	  if (filterFrameCutOff)
	    FilterRestoreFrames(in, savedFramePtr, savedVarFramePtr);
	}
	return TCL_OK;
      } else if (!strcmp(option, "callingobject")) {
	if (isFilterEntry(csc)) {
	  Tcl_SetObjResult(in, csc->filterStackEntry->callingObject);
	} else {
	  if (FilterEntryOnCallStack(in)) {
	    filterFrameCutOff =
	      FilterCutFrames(in, &savedFramePtr, &savedVarFramePtr);
	  }
	  FindInCallStack(in, global_objects[SELF]);
	  if (filterFrameCutOff)
	    FilterRestoreFrames(in, savedFramePtr, savedVarFramePtr);
	}
	return TCL_OK;

      }
    case 'f':
      if (!strcmp(option, "filterreg")) {
	if (!(isFilterEntry(csc)))
	  return XOTclVarErrMsg(in,
		   "self filterreg called from outside of a filter",
		   (char*) NULL);
	Tcl_SetObjResult(in, FilterFindReg(in, obj, GetSelfProcCmdPtr(in)));
	return TCL_OK;
      case 'n':
	if (!strcmp(option, "next"))
	  return FindSelfNext(in, obj);
      default:
	return XOTclVarErrMsg(in, "unknown option for <self>",
			      (char*)NULL);
      }
    }
  }
  return TCL_ERROR;
}


/*
 * object creation & destruction
 */

/*
 * mark an obj on the existing callstack, as not destroyed
 */
static void
UndestroyObj(Tcl_Interp* in, XOTclObject* obj) {
  XOTclCallStack *cs = &RUNTIME_STATE(in)->cs;
  XOTclCallStackContent *csc;

  /*
   * mark the object on the whole callstack as not destroyed
   */
  for (csc = &cs->content[1]; csc <= cs->top; csc++) {
    if (obj == csc->self && csc->destroyedCmd != 0) {
      /*
       * The ref count was incremented, when csc->destroyedCmd
       * was set. We revert this first before forgetting the
       * destroyedCmd.
       */
      if (csc->destroyedCmd->refCount > 1) {
	csc->destroyedCmd->refCount--;
	MEM_COUNT_FREE("command refCount",csc->destroyedCmd);
      }
      csc->destroyedCmd  = 0;
    }
  }
  /*
   * mark obj->flags XOTCL_DESTROY_CALLED as NOT CALLED (0)
   */
  obj->flags &= ~XOTCL_DESTROY_CALLED;
}

/*
 * bring an object into a state, as after initialization
 */
static void
CleanupDestroyObject(Tcl_Interp* in, XOTclObject* obj) {
  XOTclClass *thecls, *theobj;

  thecls = RUNTIME_STATE(in)->theClass;
  theobj = RUNTIME_STATE(in)->theObject;
  /* remove the instance, but not for ::Class/::Object */
  if (obj != &(thecls->object)&& obj != &(theobj->object))
    (void)RemoveInstance(obj, obj->cl);

  if (obj->nsPtr) {
    NSCleanupNamespace(in, obj->nsPtr);
    NSDeleteChildren(in, obj->nsPtr);
  }

  if (obj->varTable) {
    TclDeleteVars(((Interp *)in), obj->varTable);
    FREE(obj->varTable, obj->varTable);
    obj->varTable = 0;
  }

  if (obj->opt) {
    AssertionRemoveStore(obj->opt->assertions);
    obj->opt->assertions = NULL;

#ifdef XOTCL_METADATA
    XOTclMetaDataDestroy(obj);
#endif

    CmdListRemoveList(&obj->opt->mixins, 0);
    CmdListRemoveList(&obj->opt->filters, FilterGuardDelList);
    freeObjectOpt(obj);
  }

  obj->flags &= ~XOTCL_MIXIN_ORDER_VALID;
  MixinResetOrder(obj);
  obj->flags &= ~XOTCL_FILTER_ORDER_VALID;
  FilterResetOrder(obj);
}

/*
 * do obj initialization & namespace creation
 */
static void
CleanupInitObject(Tcl_Interp* in, XOTclObject* obj,
		  XOTclClass* cl, Tcl_Namespace *namespacePtr) {
#ifdef OBJDELETION_TRACE
  fprintf(stderr,"+++ CleanupInitObject\n");
#endif

  obj->teardown = in;
  obj->nsPtr = namespacePtr;
  AddInstance(obj, cl);

  obj->opt = 0;
  obj->varTable = 0;
  obj->mixinOrder = 0;
  obj->filterOrder = 0;
  /*
   * flgs: DESTROY_CALLED is 0, filters & mixins are INVALID
   * and not DEFINED
   */
  obj->flags = 0;
}

/*
 * physical object destroy
 */
static void
PrimitiveODestroy(ClientData cd) {
  XOTclObject* obj = (XOTclObject*)cd;
  Tcl_Interp* in;
  Tcl_Command cmd;
  Command *cmdPtr;

  assert(obj && !(obj->flags & XOTCL_DESTROYED));

  /*
   * check and latch against recurrent calls with obj->teardown
   */
  PRINTOBJ("PrimitiveODestroy", obj);

  if (!obj || !obj->teardown) return;
  in = obj->teardown;
  obj->teardown = 0;

  /*
   * Don't destroy, if the interpreter is destroyed already
   * e.g. TK calls Tcl_DeleteInterp directly, if the window is killed
   */
  if (Tcl_InterpDeleted(in)) return;
  /*
   * call and latch user destroy with obj->id if we haven't
   */
  if (!(obj->flags & XOTCL_DESTROY_CALLED)) {
    callDestroyMethod(cd, in, obj, 0);
    obj->id = 0;
  }

#ifdef OBJDELETION_TRACE
  fprintf(stderr,"  physical delete of %p id=%p destroyCalled=%d '%s'\n",
	  obj, obj->id, (obj->flags & XOTCL_DESTROY_CALLED), ObjStr(obj->cmdName));
#endif

  CleanupDestroyObject(in, obj);

  while (obj->mixinStack != NULL)
    MixinStackPop(obj);
  while (obj->filterStack != NULL)
    FilterStackPop(obj);

  cmd = Tcl_FindCommand(in, ObjStr(obj->cmdName), 0, 0);
  cmdPtr = (Command *) cmd;
  if (cmdPtr != NULL)
    cmdPtr->deleteProc = 0;

  if (obj->nsPtr) {
    XOTcl_DeleteNamespace(in, (Tcl_Namespace *)obj->nsPtr);
    obj->nsPtr = 0;
  }

  /*fprintf(stderr, " +++ OBJ/CLS free: %s\n", ObjStr(obj->cmdName));*/

  obj->flags |= XOTCL_DESTROYED;
  objTrace("ODestroy", obj);
#if REFCOUNTED
  fprintf(stderr,"ODestroy %p flags %d rc %d destr %d dc %d\n",
	    obj, obj->flags,
	    (obj->flags & XOTCL_REFCOUNTED) != 0,
	    (obj->flags & XOTCL_DESTROYED) != 0,
	    (obj->flags & XOTCL_DESTROY_CALLED) != 0
	    );
  if (!(obj->flags & XOTCL_REFCOUNTED)) {
#endif

    DECR_REF_COUNT(obj->cmdName);

#if REFCOUNTED
  }
  fprintf(stderr,"ODestroy CALL CLEANUPO\n");
#endif
  XOTclCleanupObject(obj);
#if REFCOUNTED
  fprintf(stderr,"ODestroy DONE\n");
#endif

#if !defined(NDEBUG)
  if (obj != (XOTclObject*)RUNTIME_STATE(in)->theClass)
    checkAllInstances(in, RUNTIME_STATE(in)->theClass,0);
#endif
}

static void
PrimitiveOInit(void* mem, Tcl_Interp* in, char* name, XOTclClass* cl) {
  XOTclObject* obj = (XOTclObject*)mem;

#ifdef OBJDELETION_TRACE
  fprintf(stderr,"+++ PrimitiveOInit\n");
#endif

#ifdef XOTCLOBJ_TRACE
  fprintf(stderr, "OINIT %s = %p\n", name, obj); 
#endif
  XOTclObjectRefCountIncr(obj);

  /* if the command of the obj was used before, we have to clean
   * up the callstack from set "destroyedCmd" flags
   */
  UndestroyObj(in, obj);

  if (Tcl_FindNamespace(in, name, NULL, 0)) {
    CleanupInitObject(in, obj, cl,
		      NSGetFreshNamespace(in, (ClientData)obj, name));
  } else {
    CleanupInitObject(in, obj, cl, NULL);
  }

  obj->mixinStack = 0;
  obj->filterStack = 0;
}

/*
 * Object creation: create object name (full name) and Tcl command
 */
static XOTclObject*
PrimitiveOCreate(Tcl_Interp* in, char* name, XOTclClass* cl) {
  XOTclObject* obj = (XOTclObject*)ckalloc(sizeof(XOTclObject));
  Tcl_DString fullName, *fullNamePtr = &fullName;
  int length;
  char* fn;

#if defined(XOTCLOBJ_TRACE)
  fprintf(stderr, "CKALLOC Object %p %s\n", obj, name);
#endif
#ifdef OBJDELETION_TRACE
  fprintf(stderr,"+++ PrimitiveOCreate\n");
#endif

  memset(obj, 0, sizeof(XOTclObject));
  MEM_COUNT_ALLOC("XOTclObject/XOTclClass",obj);
  assert(obj); /* ckalloc panics, if malloc fails */

  if (*name == ':' && *(name+1) == ':') {
    fn = name;
    length = strlen(name);
  } else {
    Tcl_Namespace* currNsPtr = Tcl_GetCurrentNamespace(in);
    if (currNsPtr != Tcl_GetGlobalNamespace(in) &&
	!(currNsPtr->deleteProc == NSNamespaceDeleteProc)) {
      ALLOC_NAME_NS(fullNamePtr, currNsPtr->fullName, name);
    } else {
      ALLOC_TOP_NS(fullNamePtr, name);
    }
    fn = Tcl_DStringValue(fullNamePtr);
    length = Tcl_DStringLength(fullNamePtr);
  }

  if (!NSCheckForParent(in, fn, length)) {
    if (name != fn) {
      DSTRING_FREE(fullNamePtr);
    }
    ckfree((char*) obj);
    return 0;
  }
  obj->id = Tcl_CreateObjCommand(in, fn, XOTclObjDispatch,
				 (ClientData)obj, PrimitiveODestroy);

  PrimitiveOInit(obj, in, fn, cl);
  obj->cmdName = NewXOTclObjectObjName(obj, fn, length);
  INCR_REF_COUNT(obj->cmdName);

  objTrace("PrimitiveOCreate", obj);

  if (name != fn) {
    DSTRING_FREE(fullNamePtr);
  }
  return obj;
}

/*
 * Cleanup class: remove filters, mixins, assertions, instances ...
 * and remove class from class hierarchy
 */
static void
CleanupDestroyClass(Tcl_Interp* in, XOTclClass* cl) {
  Tcl_HashSearch hSrch;
  Tcl_HashEntry* hPtr;
  XOTclClass *theobj;
  XOTclObject* obj = (XOTclObject*)cl;
  theobj = RUNTIME_STATE(in)->theObject;

  if (cl->opt) {
    CmdListRemoveList(&cl->opt->instmixins, 0);
    MixinInvalidateObjOrders(cl);

    CmdListRemoveList(&cl->opt->instfilters, FilterGuardDelList);
    FilterInvalidateObjOrders(in, cl);
    /* remove dependent filters of this class from all subclasses*/
    FilterRemoveDependentFilterCmds(cl, cl);
    AssertionRemoveStore(cl->opt->assertions);
#ifdef XOTCL_OBJECTDATA
    XOTclFreeObjectData(cl);
#endif
  }

  NSCleanupNamespace(in, cl->nsPtr);
  NSDeleteChildren(in, cl->nsPtr);


  /* reset all instances to the class ::Object, that makes no sense
     for ::Object itself */
  if (cl != theobj) {
    hPtr = &cl->instances ? Tcl_FirstHashEntry(&cl->instances, &hSrch) : 0;
    for (; hPtr != 0; hPtr = Tcl_NextHashEntry(&hSrch)) {
      XOTclObject* inst = (XOTclObject*)Tcl_GetHashKey(&cl->instances, hPtr);
      if (inst && (inst != (XOTclObject*)cl) && inst->id) {
	if (inst != &(theobj->object)) {
	  (void)RemoveInstance(inst, obj->cl);
	  AddInstance(inst, theobj);
	}
      }
    }
  }

  Tcl_DeleteHashTable(&cl->instances);
  MEM_COUNT_FREE("Tcl_InitHashTable",&cl->instances);

  if (cl->parameters) {
    DECR_REF_COUNT(cl->parameters);
  }

  if (cl->opt) {
    if (cl->opt->parameterClass) {
      DECR_REF_COUNT(cl->opt->parameterClass);
    }
    freeClassOpt(cl);
  }

  /*
   * flush all caches, unlink superclasses
   */

  FlushPrecedences(cl);
  while (cl->sub) {
    XOTclClass* subClass = cl->sub->cl;
    (void)RemoveSuper(subClass, cl);
    /* if there are no more super classes add the Object
     * class as superclasses
     * -> don't do that for Object itself!
     */
    if (subClass->super == 0 && cl != theobj)
      AddSuper(subClass, theobj);
  }
  while (cl->super) (void)RemoveSuper(cl, cl->super->cl);
}

/*
 * do class initialization & namespace creation
 */
static void
CleanupInitClass(Tcl_Interp* in, XOTclClass* cl,
		 Tcl_Namespace *namespacePtr) {
  XOTclObject *obj = (XOTclObject*)cl;

#ifdef OBJDELETION_TRACE
  fprintf(stderr,"+++ CleanupInitClass\n");
#endif

  /*
   * during init of Object and Class the theClass value is not set
   */
  /*
  if (RUNTIME_STATE(in)->theClass != 0)
    obj->type = RUNTIME_STATE(in)->theClass;
  */
  XOTclObjectSetClass(obj);

  cl->nsPtr = namespacePtr;

  cl->super = 0;
  cl->sub = 0;
  AddSuper(cl, RUNTIME_STATE(in)->theObject);
  cl->parent = RUNTIME_STATE(in)->theObject;
  cl->color = WHITE;
  cl->order = 0;
  cl->parameters = 0;

  Tcl_InitHashTable(&cl->instances, TCL_ONE_WORD_KEYS);
  MEM_COUNT_ALLOC("Tcl_InitHashTable",&cl->instances);

  cl->opt = 0;
}

/*
 * class physical destruction
 */
static void
PrimitiveCDestroy(ClientData cd) {
  XOTclClass* cl = (XOTclClass*)cd;
  XOTclObject* obj = (XOTclObject*)cd;
  Tcl_Interp* in;
  Tcl_Namespace* saved;

  /*
   * check and latch against recurrent calls with obj->teardown
   */
  if (!obj || !obj->teardown) return;
  in = obj->teardown;
  obj->teardown = 0;

  /*
   * Don't destroy, if the interpreted is destroyed already
   * e.g. TK calls Tcl_DeleteInterp directly, if Window is killed
   */
  if (Tcl_InterpDeleted(in)) return;

  /*
  if (cl->object.nsPtr) {
    fprintf(stderr, "->PrimitiveCDestroy %s\n", cl->object.nsPtr->fullName);
    NSDeleteChildren(in, cl->object.nsPtr);
  }
  */

  /*
   * call and latch user destroy with obj->id if we haven't
   */
  if (!(obj->flags & XOTCL_DESTROY_CALLED))
    callDestroyMethod(cd, in, obj, 0);
  /* we call cleanup here, before the instances
     hashtable is deleted

    result = callMethod(cd, in, global_objects[CLEANUP], 2,0, 0);
 */

  CleanupDestroyClass(in, cl);

  /*
   * handoff the primitive teardown
   */

  saved = cl->nsPtr;
  obj->teardown = in;

  /*
   * class object destroy + physical destroy
   */
  PrimitiveODestroy(cd);

  XOTcl_DeleteNamespace(in, saved);

  return;
}

/*
 * class init
 */
static void
PrimitiveCInit(void* mem, Tcl_Interp* in, char* name, XOTclClass* class) {
  XOTclClass* cl = (XOTclClass*)mem;

  Tcl_CallFrame frame;
  Tcl_Namespace* namespacePtr;
  char* n = name;

  /*
   * ensure that namespace is newly created during CleanupInitClass
   * ie. kill it, if it exists already
   */
  if (Tcl_PushCallFrame(in, &frame,
			(Tcl_Namespace*)RUNTIME_STATE(in)->XOTclClassesNS,
			0)!= TCL_OK) return;

  namespacePtr = NSGetFreshNamespace(in, (ClientData)cl, n);

  Tcl_PopCallFrame(in);

  CleanupInitClass(in, cl,  namespacePtr);
  return;
}

/*
 * class create: creation of namespace + class full name
 * calls class object creation
 */
static XOTclClass*
PrimitiveCCreate(Tcl_Interp* in, char* name, XOTclClass* class) {
  XOTclClass* cl = (XOTclClass*)ckalloc(sizeof(XOTclClass));
  Tcl_DString fullName, *fullNamePtr = &fullName;
  char *fn;
  int length;
  XOTclObject* obj = (XOTclObject*)cl;

  memset(cl, 0, sizeof(XOTclClass));

  MEM_COUNT_ALLOC("XOTclObject/XOTclClass",cl);
  /*
  fprintf(stderr, " +++ CLS alloc: %s\n", name);
  */
  if (*name == ':' && *(name+1) == ':') {
    fn = name;
    length = strlen(name);
  } else {
    Tcl_Namespace * currNsPtr = Tcl_GetCurrentNamespace(in);
    if (currNsPtr != Tcl_GetGlobalNamespace(in)
	&& !(currNsPtr->deleteProc == NSNamespaceDeleteProc)) {
      ALLOC_NAME_NS(fullNamePtr, currNsPtr->fullName, name);
    } else {
      ALLOC_TOP_NS(fullNamePtr, name);
    }
    fn = Tcl_DStringValue(fullNamePtr);
    length = Tcl_DStringLength(fullNamePtr);
  }
  /*
  fprintf(stderr,"Class alloc %p '%s'\n", cl, fn);
  */
  /* check whether Object parent NS already exists,
     otherwise: error */
  if (!NSCheckForParent(in, fn, length)) {
    if (fn != name) {
      DSTRING_FREE(fullNamePtr);
    }
    ckfree((char*) cl);
    return 0;
  }
  obj->id = Tcl_CreateObjCommand(in, fn, XOTclObjDispatch,
				 (ClientData)cl, PrimitiveCDestroy);

  PrimitiveOInit(obj, in, fn, class);
  obj->cmdName = NewXOTclObjectObjName(obj,fn,length);
  INCR_REF_COUNT(obj->cmdName);
  PrimitiveCInit(cl, in, fn+2, class);

  objTrace("PrimitiveCCreate", obj);

  if (fn != name) {
    DSTRING_FREE(fullNamePtr);
  }
  return cl;
}

/* change XOTcl class conditionally; obj must not be NULL */

static XOTCLINLINE void
changeClass(Tcl_Interp* in, XOTclObject* obj, XOTclClass* cl) {
  assert(obj);
  if (cl != obj->cl) {
    (void)RemoveInstance(obj, obj->cl);
    AddInstance(obj, cl);
    
    MixinComputeDefined(in, obj);
    FilterComputeDefined(in, obj);
  }
}


/*
 * Undestroy the object, reclass it, and call "cleanup" afterwards
 */
static int
doCleanup(Tcl_Interp* in, XOTclObject* newobj, XOTclObject* classobj,
	  int objc, Tcl_Obj *objv[]) {
  int destroyed = 0, result;
  XOTclCallStack *cs = &RUNTIME_STATE(in)->cs;
  XOTclCallStackContent *csc;
  /*
   * we check whether the object to be re-created is destroyed or not
   */
  for (csc = &cs->content[1]; csc <= cs->top; csc++) {
    if (newobj == csc->self && csc->destroyedCmd != 0) {
      destroyed = 1; break;
    }
  }

  if (destroyed)
    UndestroyObj(in, newobj);

  /*
   *  re-create, first ensure correct class for newobj
   */

  changeClass(in, newobj, (XOTclClass*) classobj);

  /*
  result = callMethodWithArg((ClientData)newobj, in,
			     global_objects[CLASS], classobj->cmdName,
			     3, 0, 0);
  if (result != TCL_OK)
    return result;
  */
  /*
   * dispatch "cleanup"
   */
  result = callMethod((ClientData) newobj, in,
		      global_objects[CLEANUP], 2, 0, 0);
  return result;
}

/*
 * Std initialization:
 *   call parameter default values
 *   apply "-" methods (call "configure" with given arguments)
 *   call constructor "init", if it was not called before
 */
static int
doObjInitialization(Tcl_Interp* in, XOTclObject* obj,
		    int objc, Tcl_Obj *objv[]) {
  int result, initCalled = 0, initArgsC = objc, i;

  /*
   * Search for default values of parameter on superclasses
   */
  result = callParameterMethodWithArg(obj, in,
				      global_objects[SEARCH_DEFAULTS],
				      obj->cmdName, 3, 0, 0);
  if (result != TCL_OK)
    return result;

  /*
   * call init methods (starting with '-')
   */
  result = callMethod((ClientData) obj, in,
		      global_objects[CONFIGURE], objc, objv+2, 0);
  if (result != TCL_OK)
    return result;

  /* check, whether init was called already, and determine where the
   * configure (with '-') start (we don't send them as args to
   * "init").  */

  for (i=1; i < objc; i++) {
      char* arg = ObjStr(objv[i]);
    if (isDashArg(arg)) {
      if (initArgsC == objc)
	initArgsC = i;
      arg++;
      if (isInitString(arg)) {
	initCalled = 1;
	break;
      }
    }
  }
  if (!initCalled) {
    /*
     * Call the user-defined constructor 'init'
     */
    result = callMethod((ClientData) obj, in, global_objects[INIT],
			initArgsC, objv+2, 0);
  }

  return result;
}


/*
 * experimental resolver implementation -> not used at the moment
 */
#ifdef NOT_USED
static int
XOTclResolveCmd(Tcl_Interp* in, char* name, Tcl_Namespace *contextNsPtr,
		int flags, Tcl_Command *rPtr) {

  Namespace *nsPtr[2], *cxtNsPtr;
  char *simpleName;
  register Tcl_HashEntry *entryPtr;
  register Command *cmdPtr;
  register int search;

  /*fprintf(stderr, "  ***%s->%s\n", contextNsPtr->fullName, name);*/

  /*
   * Find the namespace(s) that contain the command.
   */
  if ((flags & TCL_GLOBAL_ONLY) != 0) {
    cxtNsPtr = (Namespace *) Tcl_GetGlobalNamespace(in);
  }
  else if (contextNsPtr != NULL) {
    cxtNsPtr = (Namespace *) contextNsPtr;
  }
  else {
    cxtNsPtr = (Namespace *) Tcl_GetCurrentNamespace(in);
  }

  TclGetNamespaceForQualName(in, name, (Namespace *) contextNsPtr,
			     flags, &nsPtr[0], &nsPtr[1],
			     &cxtNsPtr, &simpleName);

  /*fprintf(stderr, "  ***Found %s, %s\n", nsPtr[0]->fullName, nsPtr[0]->fullName);*/

  /*
   * Look for the command in the command table of its namespace.
   * Be sure to check both possible search paths: from the specified
   * namespace context and from the global namespace.
   */

  cmdPtr = NULL;
  for (search = 0;  (search < 2) && (cmdPtr == NULL);  search++) {
    if ((nsPtr[search] != NULL) && (simpleName != NULL)) {
      entryPtr = Tcl_FindHashEntry(&nsPtr[search]->cmdTable,
				   simpleName);
      if (entryPtr != NULL) {
	cmdPtr = (Command *) Tcl_GetHashValue(entryPtr);
      }
    }
  }
  if (cmdPtr != NULL) {
    if (cxtNsPtr->deleteProc == NSNamespaceDeleteProc &&
	cmdPtr->objProc != XOTclObjDispatch &&
	cmdPtr->objProc != XOTclNextObjCmd &&
	cmdPtr->objProc != XOTclGetSelfObjCmd) {

      /*
       * the cmdPtr is defined in an XOTcl object or class namespace, but
       * not an object & not self/next -> redispatch in
       * global namespace
       */
      cmdPtr = 0;
      nsPtr[0] = (Namespace *) Tcl_GetGlobalNamespace(in);
      if ((nsPtr[0] != NULL) && (simpleName != NULL)) {
	entryPtr = Tcl_FindHashEntry(&nsPtr[0]->cmdTable,
				     simpleName);
	if (entryPtr != NULL) {
	  cmdPtr = (Command *) Tcl_GetHashValue(entryPtr);
	}
      }

      /*
      XOTclStackTrace(in);
      XOTclCallStackTrace(in);
      */
    }
    *rPtr = (Tcl_Command) cmdPtr;
    return TCL_OK;
  }

  return TCL_CONTINUE;
}
static int
XOTclResolveVar(Tcl_Interp* interp, char* name, Tcl_Namespace *context,
 	        Tcl_ResolvedVarInfo *rPtr) {
  /*fprintf(stderr, "Resolving %s in %s\n", name, context->fullName);*/

  return TCL_CONTINUE;
}
#endif

/*
 * object method implementations
 */

static int
XOTclODestroyMethod(ClientData cd, Tcl_Interp* in,
		    int objc, Tcl_Obj *CONST objv[]) {
  XOTclObject* obj = (XOTclObject*)cd;

  if (!obj) return XOTclObjErrType(in, objv[0], "Object");
  if (objc < 1) return XOTclObjErrArgCnt(in, obj->cmdName, "destroy");
  PRINTOBJ("XOTclODestroyMethod", obj);

  /*
   * call instdestroy for [self]
   */
  return callMethodWithArg((ClientData)obj->cl, in,
			   global_objects[INSTDESTROY], obj->cmdName,
			   objc+2, objv+1, 0);
}

static int
XOTclOCleanupMethod(ClientData cd, Tcl_Interp* in,
		    int objc, Tcl_Obj *CONST objv[]) {
  XOTclObject *obj = (XOTclObject*)cd;
  XOTclClass  *cl  = XOTclObjectToClass(obj);
  char* fn;
  Tcl_Obj* savedNameObj;

#if defined(OBJDELETION_TRACE)
  fprintf(stderr,"+++ XOTclOCleanupMethod\n");
#endif

  if (!obj) return XOTclObjErrType(in, objv[0], "Object");
  if (objc < 1) return XOTclObjErrArgCnt(in, obj->cmdName, "cleanup");
  PRINTOBJ("XOTclOCleanupMethod", obj);

  fn = ObjStr(obj->cmdName);
  savedNameObj = obj->cmdName;
  INCR_REF_COUNT(savedNameObj);

  CleanupDestroyObject(in, obj);
  CleanupInitObject(in, obj, obj->cl, obj->nsPtr);

  if (cl) {
    CleanupDestroyClass(in, cl);
    CleanupInitClass(in, cl, cl->nsPtr);
  }

  DECR_REF_COUNT(savedNameObj);

  return TCL_OK;
}

static int
XOTclOIsClassMethod(ClientData cd, Tcl_Interp* in,
		     int objc, Tcl_Obj *CONST objv[]) {
  XOTclObject* obj = (XOTclObject*)cd, *o;
  Tcl_Obj *className;
  if (!obj) return XOTclObjErrType(in, objv[0], "Object");
  if (objc < 1 || objc > 2) return XOTclObjErrArgCnt(in,
					  obj->cmdName,
					  "isclass ?className?");

  className = (objc == 2) ? objv[1] : obj->cmdName;

  if (GetXOTclObjectFromObj(in, className, &o) == TCL_OK
      && XOTclObjectIsClass(o)) {
    Tcl_SetIntObj(Tcl_GetObjResult(in), 1);
  } else {
    Tcl_SetIntObj(Tcl_GetObjResult(in), 0);
  }
  return TCL_OK;
}

static int
XOTclOIsObjectMethod(ClientData cd, Tcl_Interp* in,
		     int objc, Tcl_Obj *CONST objv[]) {
  XOTclObject* obj = (XOTclObject*)cd, *o;;

  if (!obj) return XOTclObjErrType(in, objv[0], "Object");
  if (objc != 2) return XOTclObjErrArgCnt(in,
					       obj->cmdName,
					       "isobject <objName>");

  if (GetXOTclObjectFromObj(in, objv[1], &o) == TCL_OK) {
    Tcl_SetIntObj(Tcl_GetObjResult(in), 1);
  } else {
    Tcl_SetIntObj(Tcl_GetObjResult(in), 0);
  }
  return TCL_OK;
}

static int
IsMetaClass(Tcl_Interp* in, XOTclClass* cl) {
  /* check if cl is a meta-class by checking is Class is a superclass of cl*/
  XOTclClasses* pl;
  if (cl == RUNTIME_STATE(in)->theClass)
    return 1;

  pl = ComputeOrder(cl, Super);
  while (pl != 0) {
    if (pl->cl == RUNTIME_STATE(in)->theClass)
      return 1;
    pl=pl->next;
  }
  return 0;
}

static int
XOTclOIsMetaClassMethod(ClientData cd, Tcl_Interp* in,
			int objc, Tcl_Obj *CONST objv[]) {
  XOTclObject *obj = (XOTclObject*)cd, *o;
  Tcl_Obj *className;
  if (!obj) return XOTclObjErrType(in, objv[0], "Object");
  if (objc < 1 || objc > 2) return XOTclObjErrArgCnt(in,
					  obj->cmdName,
					  "ismetaclass ?metaClassName?");

  className = (objc == 2) ? objv[1] : obj->cmdName;

  if (GetXOTclObjectFromObj(in, className, &o) == TCL_OK
      && XOTclObjectIsClass(o)
      && IsMetaClass(in, (XOTclClass*)o)) {
    Tcl_SetIntObj(Tcl_GetObjResult(in), 1);
  } else {
    Tcl_SetIntObj(Tcl_GetObjResult(in), 0);
  }
  return TCL_OK;
}

static int
XOTclOIsTypeMethod(ClientData cd, Tcl_Interp* in,
			int objc, Tcl_Obj *CONST objv[]) {
  XOTclObject *obj = (XOTclObject*)cd;
  XOTclClass *cl;
  XOTclClasses *t;

  if (!obj) return XOTclObjErrType(in, objv[0], "Object");
  if (objc != 2) return XOTclObjErrArgCnt(in,
					  obj->cmdName,
					  "istype <className>");
  if (obj->cl && GetXOTclClassFromObj(in,objv[1],&cl) == TCL_OK) {
    if (cl == obj->cl) {
      Tcl_SetIntObj(Tcl_GetObjResult(in), 1);
      return TCL_OK;
    }
    t = ComputeOrder(obj->cl, Super);
    while (t && t->cl && t->cl != cl) {
      t = t->next;
    }
    if (t && t->cl == cl) {
      Tcl_SetIntObj(Tcl_GetObjResult(in), 1);
      return TCL_OK;
    }
  }
  Tcl_SetIntObj(Tcl_GetObjResult(in), 0);
  return TCL_OK;
}


static int
XOTclOClassMethod(ClientData cd, Tcl_Interp* in,
		  int objc, Tcl_Obj *CONST objv[]) {
  XOTclObject* obj = (XOTclObject*)cd;
  XOTclClass* cl;

  if (!obj) return XOTclObjErrType(in, objv[0], "Object");
  if (objc != 2) return XOTclObjErrArgCnt(in, obj->cmdName, "class <class>");

  /*
   * allow a change to any class; type system enforces safety later
   */
  if (GetXOTclClassFromObj(in, objv[1], &cl) != TCL_OK)
    return XOTclErrBadVal(in, "a class", ObjStr(objv[1]));

  changeClass(in, obj, cl);

  return TCL_OK;
}

static int
XOTclOExistsMethod(ClientData cd, Tcl_Interp* in, int objc, Tcl_Obj *objv[]) {
  XOTclObject* obj = (XOTclObject*)cd;

  if (!obj) return XOTclObjErrType(in, objv[0], "Object");
  if (objc != 2) return XOTclObjErrArgCnt(in, obj->cmdName, "exists var");

  Tcl_SetIntObj(Tcl_GetObjResult(in),
		varExists(in, obj, ObjStr(objv[1])));
  return TCL_OK;
}

static int
countModifiers(int objc, Tcl_Obj * CONST objv[]) {
  int i, count = 0;
  char* to;
  for (i = 2; i < objc; i++) {
    to = ObjStr(objv[i]);
    if (to[0] == '-') {
      count++;
      /* '--' stops modifiers */
      if (to[1] == '-') break;
    }
  }
  return count;
}

static int
checkForModifier(Tcl_Obj * CONST objv[], int numberModifiers,
		 char* modifier) {
  int i;
  if (numberModifiers == 0) return 0;
  for (i = 2; i-2 < numberModifiers; i++) {
    char* ov = ObjStr(objv[i]);
    /* all start with a "-", so there must be a ov[1] */
    if (ov[1] == modifier[1] && !strcmp(ov, modifier))
      return 1;
  }
  return 0;
}

static int
XOTclOInfoMethod(ClientData cd, Tcl_Interp* in, int objc, Tcl_Obj* CONST objv[]) {
  XOTclObject* obj = (XOTclObject*)cd;
  Namespace *nsp = (Namespace *)obj->nsPtr;
  char *cmd, *pattern;
  int modifiers = 0;

  if (!obj) return XOTclObjErrType(in, objv[0], "Object");
  if (objc < 2)
    return XOTclObjErrArgCnt(in, obj->cmdName, "info <opt> ?args?");

  cmd = ObjStr(objv[1]);
  pattern = (objc > 2) ? ObjStr(objv[2]) : 0;

  /*
   * check for "-" modifiers
   */
  if (pattern && *pattern == '-') {
    modifiers = countModifiers(objc, objv);
    pattern = (objc > 2+modifiers) ? ObjStr(objv[2+modifiers]) : 0;
  }

  switch (*cmd) {
  case 'a':
    if (isArgsString(cmd)) {
      if (objc != 3 || modifiers > 0)
	return XOTclObjErrArgCnt(in, obj->cmdName, "info args <proc>");
      if (nsp)
	return ListProcArgs(in, &nsp->cmdTable, pattern);
      else
	return TCL_OK;
    }
    break;

  case 'b':
    if (!strcmp(cmd, "body")) {
      if (objc != 3 || modifiers > 0)
	return XOTclObjErrArgCnt(in, obj->cmdName, "info body <proc>");
      if (nsp)
	return ListProcBody(in, &nsp->cmdTable, pattern);
      else
	return TCL_OK;
    }
    break;

  case 'c':
    if (isClassString(cmd)) {
      if (objc > 3 || modifiers > 0)
	return XOTclObjErrArgCnt(in, obj->cmdName, "info class ?class?");
      return ListClass(in, obj, pattern, objc, objv);
    } else if (!strcmp(cmd, "commands")) {
      if (objc > 3 || modifiers > 0)
	return XOTclObjErrArgCnt(in, obj->cmdName, "info commands ?pat?");
      if (nsp)
	return ListKeys(in, &nsp->cmdTable, pattern);
      else
	return TCL_OK;
    } else if (!strcmp(cmd, "children")) {
      if (objc > 3 || modifiers > 0)
	return XOTclObjErrArgCnt(in, obj->cmdName, "info children ?pat?");
      return ListObjChildren(in, obj, pattern);
    } else if (!strcmp(cmd, "check")) {
      if (objc != 2 || modifiers > 0)
	return XOTclObjErrArgCnt(in, obj->cmdName, "info check");
      return AssertionListCheckOption(in, obj);
    }
    break;

  case 'd':
    if (!strcmp(cmd, "default")) {
      if (objc != 5 || modifiers > 0)
	return XOTclObjErrArgCnt(in, obj->cmdName,
				 "info default <proc> <arg> <var>");
      if (nsp)
	return ListProcDefault(in,  &nsp->cmdTable, pattern,
			       ObjStr(objv[3]), ObjStr(objv[4]));
      else
	return TCL_OK;
    }
    break;

  case 'f':
    if (!strcmp(cmd, "filter")) {
      int withGuards = 0, withOrder = 0;
      if (objc-modifiers > 3)
	return XOTclObjErrArgCnt(in, obj->cmdName,
				 "info filter ?-guards? ?-order? ?pat?");
      if (modifiers > 0) {
	withGuards = checkForModifier(objv, modifiers, "-guards");
	withOrder = checkForModifier(objv, modifiers, "-order");
	if (withGuards == 0 && withOrder == 0)
	  return XOTclVarErrMsg(in, "info filter: unknown modifier ",
				ObjStr(objv[2]));
	if (withGuards && withOrder)
	  return XOTclVarErrMsg(in, "info filter: cannot use -guards and -order together",
				ObjStr(objv[2]));
      }

      if (withOrder) {
	if (!(obj->flags & XOTCL_FILTER_ORDER_VALID))
	  FilterComputeDefined(in, obj);
	return FilterInfo(in, obj->filterOrder, pattern, 0, 1);
      }

      if (!obj->opt)
	return TCL_OK;
      return FilterInfo(in, obj->opt->filters, pattern, withGuards, 0);

    } else if (!strcmp(cmd, "filterguard")) {
      if (objc != 3 || modifiers > 0)
	return XOTclObjErrArgCnt(in, obj->cmdName, "info filterguard filter");
      if (!obj->opt)
	return TCL_OK;
      return FilterGuardList(in, obj->opt->filters, pattern);
    }
    break;

  case 'h':
    if (!strcmp(cmd, "hasNamespace")) {
      if (nsp)
	Tcl_SetObjResult(in, global_objects[ONE]);
      else
	Tcl_SetObjResult(in, global_objects[ZERO]);
      return TCL_OK;
    }
    break;

  case 'i':
    if (!strcmp(cmd, "invar")) {
      if (objc != 2 || modifiers > 0)
	return XOTclObjErrArgCnt(in, obj->cmdName, "info invar");
      if (obj->opt && obj->opt->assertions)
	Tcl_SetObjResult(in, AssertionList(in,
				   obj->opt->assertions->invariants));
      return TCL_OK;
    } else if (!strcmp(cmd, "info")) {
      if (objc > 2 || modifiers > 0)
	return XOTclObjErrArgCnt(in, obj->cmdName, "info info");
      return ListInfo(in, GetXOTclClassFromObj(in,obj->cmdName,NULL) == TCL_OK);
    }
    break;

  case 'm':
    if (!strcmp(cmd, "mixin")) {
      int withOrder = 0;
      if (objc-modifiers > 3)
	return XOTclObjErrArgCnt(in, obj->cmdName, "info mixin ?-order? ?class?");
      if (modifiers > 0) {
	withOrder = checkForModifier(objv, modifiers, "-order");
	if (withOrder == 0)
	  return XOTclVarErrMsg(in, "info mixin: unknown modifier ",
				ObjStr(objv[2]));
      }

     if (withOrder) {
	if (!(obj->flags & XOTCL_MIXIN_ORDER_VALID))
	  MixinComputeDefined(in, obj);
	return MixinInfo(in, obj->mixinOrder, pattern);
     }

     if (!obj->opt)
       return TCL_OK;
     return MixinInfo(in, obj->opt->mixins, pattern);

    } else if (!strcmp(cmd, "mixins")) {
      XOTclDeprecatedMsg("info mixins", "info mixin");
      if (objc > 3 || modifiers > 0)
	return XOTclObjErrArgCnt(in, obj->cmdName, "info mixin ?class?");
      if (!obj->opt)
	return TCL_OK;
      return MixinInfo(in, obj->opt->mixins, pattern);
    } else  if (!strcmp(cmd, "methods")) {
      int noprocs = 0, nocmds = 0, nomixins = 0;
      if (objc-modifiers > 3)
	return XOTclObjErrArgCnt(in, obj->cmdName,
				 "info methods ?-noprocs? ?-nocmds? ?-nomixins? ?pat?");
      if (modifiers > 0) {
	noprocs = checkForModifier(objv, modifiers, "-noprocs");
	nocmds = checkForModifier(objv, modifiers, "-nocmds");
	nomixins = checkForModifier(objv, modifiers, "-nomixins");
      }
      return ListMethods(in, obj, pattern, noprocs, nocmds, nomixins);
    }
#ifdef XOTCL_METADATA
     else  if (!strcmp(cmd, "metadata")) {
      if (objc > 3 || modifiers > 0)
	return XOTclObjErrArgCnt(in, obj->cmdName, "info metadata ?pat?");
      return ListKeys(in, &obj->metaData, pattern);
    }
#endif
    break;
  case 'p':
    if (!strcmp(cmd, "procs")) {
      if (objc > 3 || modifiers > 0)
	return XOTclObjErrArgCnt(in, obj->cmdName, "info procs ?pat?");
      if (nsp)
	return ListMethodKeys(in, &nsp->cmdTable, pattern,
			      /*noProcs*/ 0, /*noCmds*/ 1);
      else
	return TCL_OK;
    } else if (!strcmp(cmd, "parent")) {
      if (objc > 2 || modifiers > 0)
	return XOTclObjErrArgCnt(in, obj->cmdName, "info parent");
      return ListParent(in, obj);
    } else if (!strcmp(cmd, "pre")) {
      XOTclProcAssertion* procs;
      if (objc != 3 || modifiers > 0)
	return XOTclObjErrArgCnt(in, obj->cmdName, "info pre <proc>");
      if (obj->opt) {
	procs = AssertionFindProcs(obj->opt->assertions, ObjStr(objv[2]));
	if (procs) Tcl_SetObjResult(in, AssertionList(in, procs->pre));
      }
      return TCL_OK;
    } else if (!strcmp(cmd, "post")) {
      XOTclProcAssertion* procs;
      if (objc != 3 || modifiers > 0)
	return XOTclObjErrArgCnt(in, obj->cmdName, "info post <proc>");
      if (obj->opt) {
	procs = AssertionFindProcs(obj->opt->assertions, ObjStr(objv[2]));
	if (procs) Tcl_SetObjResult(in, AssertionList(in, procs->post));
      }
      return TCL_OK;
    }
    break;
  case 'v':
    if (!strcmp(cmd, "vars")) {
      if (objc > 3 || modifiers > 0)
	return XOTclObjErrArgCnt(in, obj->cmdName, "info vars ?pat?");
      return ListVars(in, obj, pattern);
    }
    break;
  }
  return XOTclErrBadVal
    (in, "an info option (use 'info info' to list all info options)", cmd);
  return TCL_OK;
}


static int
XOTclOProcMethod(ClientData cd, Tcl_Interp* in, int objc, Tcl_Obj * CONST objv[]) {
  XOTclObject* obj = (XOTclObject*)cd;
  char *argStr, *bdyStr, *name;

  if (!obj) return XOTclObjErrType(in, objv[0], "Object");
  if (objc != 4 && objc != 6)
      return XOTclObjErrArgCnt(in,
				    obj->cmdName,
				    "proc name args body ?preAssertion postAssertion?");
  argStr = ObjStr(objv[2]);
  bdyStr = ObjStr(objv[3]);
  name = ObjStr(objv[1]);

  if (*argStr == 0 && *bdyStr == 0) {
    if (obj->opt)
      AssertionRemoveProc(obj->opt->assertions, name);
    if (obj->nsPtr)
      NSDeleteCmd(in, obj->nsPtr, name);
  } else {
    requireObjectOpt(obj);
    if (!obj->opt->assertions)
      obj->opt->assertions = AssertionCreateStore();
    requireObjNamespace(in, obj);
    MakeProc(obj->nsPtr, obj->opt->assertions, in, objc, (Tcl_Obj **) objv);
  }

  /* could be a filter => recompute filter order */
  FilterComputeDefined(in, obj);

  return TCL_OK;
}

static int
XOTclOIncrMethod(ClientData cd, Tcl_Interp* in, int objc, Tcl_Obj * CONST objv[]) {
  XOTclObject* obj = (XOTclObject*)cd;
  VarFrameDecls;
  int result;

  if (!obj)
    return XOTclObjErrType(in, objv[0], "Object");
  if (objc < 2)
    return XOTclObjErrArgCnt(in, obj->cmdName, "incr <varName> ?increment?");

  VarFrameSwitchToObj(in, obj);
  result = XOTcl_IncrObjCmd(cd, in, objc, objv);
  VarFrameRestore(in, obj);

  return result;
}

Tcl_Obj*
XOTclOSetInstVar(XOTclObject* obj, Tcl_Interp* in,
		 Tcl_Obj* name, Tcl_Obj* value, int flgs) {
  return XOTclOSetInstVar2(obj, in, name, (Tcl_Obj *) NULL, value,
			   (flgs|TCL_PARSE_PART1));
}

Tcl_Obj*
XOTclOGetInstVar(XOTclObject* obj, Tcl_Interp* in, Tcl_Obj* name, int flgs) {
    return XOTclOGetInstVar2(obj, in, name, (Tcl_Obj *)NULL,
			     (flgs|TCL_PARSE_PART1));
}

int
XOTclUnsetInstVar(XOTclObject* obj, Tcl_Interp* in, char* name, int flgs) {
    return XOTclUnsetInstVar2 (obj, in, name,(char*)NULL, flgs);
}

extern int
XOTclCreateObject(Tcl_Interp* in, Tcl_Obj* name, XOTclClass* cl) {
  int result;
  INCR_REF_COUNT(name);
  result = callMethodWithArg((ClientData)cl, in,
			     global_objects[CREATE], name, 3, 0, 0);
  DECR_REF_COUNT(name);
  return result;
}

extern int
XOTclCreateClass(Tcl_Interp* in, Tcl_Obj* name, XOTclClass* cl) {
  int result;
  INCR_REF_COUNT(name);
  result = callMethodWithArg((ClientData)cl, in,
			     global_objects[CREATE], name, 3, 0, 0);
  DECR_REF_COUNT(name);
  return result;
}

int
XOTclDeleteObject(Tcl_Interp* in, XOTclObject* obj) {
  return callMethod((ClientData)obj, in, global_objects[DESTROY],2,0,0);
}

int
XOTclDeleteClass(Tcl_Interp* in, XOTclClass* cl) {
  return callMethod((ClientData)cl, in, global_objects[DESTROY],2,0,0);
}

extern Tcl_Obj*
XOTclOSetInstVar2(XOTclObject* obj, Tcl_Interp* in,
		  Tcl_Obj* name1, Tcl_Obj* name2,
		  Tcl_Obj* value, int flgs) {
  Tcl_Obj* result;
  VarFrameDecls;

  VarFrameSwitchToObj(in, obj);
  if (obj->nsPtr)
    flgs |= TCL_NAMESPACE_ONLY;

  result = Tcl_ObjSetVar2(in, name1, name2, value, flgs);

  VarFrameRestore(in, obj);

  return result;
}

extern int
XOTclUnsetInstVar2(XOTclObject* obj, Tcl_Interp* in, char* name1, char *name2,
        int flgs) {
  int result;
  VarFrameDecls;

  VarFrameSwitchToObj(in, obj);
  if (obj->nsPtr)
    flgs |= TCL_NAMESPACE_ONLY;

  result = Tcl_UnsetVar2(in, name1, name2, flgs);
  VarFrameRestore(in, obj);
  return result;
}

/*
 * We need NewVar from tclVar.c ... but its not exported
 */
static Var *NewVar() {
  register Var *varPtr;

  varPtr = (Var *) ckalloc(sizeof(Var));
  varPtr->value.objPtr = NULL;
  varPtr->name = NULL;
  varPtr->nsPtr = NULL;
  varPtr->hPtr = NULL;
  varPtr->refCount = 0;
  varPtr->tracePtr = NULL;
  varPtr->searchPtr = NULL;
  varPtr->flags = (VAR_SCALAR | VAR_UNDEFINED | VAR_IN_HASHTABLE);
  return varPtr;
}

static void
CleanupVar(Var * varPtr, Var *arrayPtr) {
    if (TclIsVarUndefined(varPtr) && (varPtr->refCount == 0)
	    && (varPtr->tracePtr == NULL)
	    && (varPtr->flags & VAR_IN_HASHTABLE)) {
	if (varPtr->hPtr != NULL) {
	    Tcl_DeleteHashEntry(varPtr->hPtr);
	}
	ckfree((char *) varPtr);
    }
    if (arrayPtr != NULL) {
	if (TclIsVarUndefined(arrayPtr) && (arrayPtr->refCount == 0)
		&& (arrayPtr->tracePtr == NULL)
	        && (arrayPtr->flags & VAR_IN_HASHTABLE)) {
	    if (arrayPtr->hPtr != NULL) {
		Tcl_DeleteHashEntry(arrayPtr->hPtr);
	    }
	    ckfree((char *) arrayPtr);
	}
    }
}

static int
GetInstVarIntoCurrentScope(Tcl_Interp* in, XOTclObject* obj,
			   char* varName, char* newName) {
  Interp *iPtr = (Interp *) in;
  Var *varPtr, *otherPtr = 0, *arrayPtr;
  int new;
  CallFrame *varFramePtr;
  Tcl_HashEntry *hPtr;
  Tcl_HashTable *tablePtr;
  VarFrameDecls;

  int flgs = TCL_LEAVE_ERR_MSG |
    /* PARSE_PART1 needed for 8.0.5 */ TCL_PARSE_PART1;

  VarFrameSwitchToObj(in, obj);
  if (obj->nsPtr) {
    flgs = flgs|TCL_NAMESPACE_ONLY;
  }

  otherPtr = TclLookupVar(in, varName, (char *) NULL, flgs, "define",
			  /*createPart1*/ 1, /*createPart2*/ 1, &arrayPtr);
  VarFrameRestore(in, obj);

  if (otherPtr == NULL) {
    return XOTclVarErrMsg(in, "can't make instvar ", varName,
			  ": can't find variable on ", ObjStr(obj->cmdName),
			  NULL);
  }

  /*
   * if newName == NULL -> there is no alias, use varName
   * as target link name
   */
  if (newName == NULL) {
    /*
     * Variable link into namespace cannot be an element in an array.
     * see Tcl_VariableObjCmd ...
     */
    if (arrayPtr) {
      return XOTclVarErrMsg(in, "can't make instvar ", varName,
			    " on ", ObjStr(obj->cmdName),
			    ": variable cannot be an element in an array;",
			    " use an alias or objeval.", NULL);
    }

    newName = varName;
  }


  /*
   * Mark the variable as a namespace variable
  if (obj->nsPtr) {
    if (!(otherPtr->flags & VAR_NAMESPACE_VAR)) {
      otherPtr->flags |= VAR_NAMESPACE_VAR;
      otherPtr->refCount++;
    }
  }
   */

  varFramePtr = iPtr->varFramePtr;

  /*
   * If we are executing inside a Tcl procedure, create a local
   * variable linked to the new namespace variable "varName".
   */
  if ((iPtr->varFramePtr != NULL)
      && iPtr->varFramePtr->isProcCallFrame) {
    Proc *procPtr = varFramePtr->procPtr;
    int localCt = procPtr->numCompiledLocals;
    CompiledLocal *localPtr = procPtr->firstLocalPtr;
    Var *localVarPtr = varFramePtr->compiledLocals;
    int nameLen = strlen(newName);
    int i;

    varPtr = NULL;
    for (i = 0;  i < localCt;  i++) {    /* look in compiled locals */
      if (!TclIsVarTemporary(localPtr)) {
	char *localName = localVarPtr->name;
	if ((newName[0] == localName[0])
	    && (nameLen == localPtr->nameLength)
	    && (strcmp(newName, localName) == 0)) {
	  varPtr = localVarPtr;
	  new = 0;
	  break;
	}
      }
      localVarPtr++;
      localPtr = localPtr->nextPtr;
    }

    if (varPtr == NULL) {	/* look in frame's local var hashtable */
      tablePtr = varFramePtr->varTablePtr;
      if (tablePtr == NULL) {
	tablePtr = (Tcl_HashTable *) ckalloc(sizeof(Tcl_HashTable));
	Tcl_InitHashTable(tablePtr, TCL_STRING_KEYS);
	varFramePtr->varTablePtr = tablePtr;
      }
      hPtr = Tcl_CreateHashEntry(tablePtr, newName, &new);
      if (new) {
	varPtr = NewVar();
	Tcl_SetHashValue(hPtr, varPtr);
	varPtr->hPtr = hPtr;
	varPtr->nsPtr = varFramePtr->nsPtr;
      } else {
	varPtr = (Var *) Tcl_GetHashValue(hPtr);
      }
    }
    /*
     * if we define an alias (newName != varName), be sure that
     * the target does not exist already
     */
    if (!new) {
      if (varPtr == otherPtr)
	return XOTclVarErrMsg(in, "can't instvar to variable itself", 0);
      if (TclIsVarLink(varPtr)) {
	/* we try to make the same instvar again ... this is ok */
	Var *linkPtr = varPtr->value.linkPtr;
	if (linkPtr == otherPtr) {
	  return TCL_OK;
	}

	linkPtr->refCount--;
	if (TclIsVarUndefined(linkPtr)) {
	  CleanupVar(linkPtr, (Var *) NULL);
	}

	    /*
	return XOTclVarErrMsg(in, "can't instvar to link", 0);
	    */
      } else if (!TclIsVarUndefined(varPtr)) {
	Tcl_AppendResult((Tcl_Interp *) iPtr, "variable \"", newName,
			 "\" already exists", (char *) NULL);
	return TCL_ERROR;
      } else if (varPtr->tracePtr != NULL) {
	Tcl_AppendResult((Tcl_Interp *) iPtr, "variable \"", newName,
			 "\" has traces: can't use for instvar", (char *) NULL);
	return TCL_ERROR;
      }
    }
    TclSetVarLink(varPtr);
    TclClearVarUndefined(varPtr);
    varPtr->value.linkPtr = otherPtr;
    otherPtr->refCount++;
  }
  return TCL_OK;
}

static int
XOTclOInstVarMethod(ClientData cd, Tcl_Interp* in, int objc, Tcl_Obj *objv[]);

extern int
XOTclInstVar(XOTclObject* obj, Tcl_Interp* in, char* name, char *destName) {
  int result;
  Tcl_Obj* alias = 0;
  DEFINE_NEW_TCL_OBJECTS_ON_STACK(2, objv);

  objv[0] = global_objects[INSTVAR];
  objv[1] = Tcl_NewStringObj(name, -1);
  INCR_REF_COUNT(objv[1]);

  if (destName != 0) {
    alias = Tcl_NewStringObj(destName, -1);
    INCR_REF_COUNT(alias);
    Tcl_ListObjAppendElement(in, objv[1], alias);
  }

  result = XOTclOInstVarMethod((ClientData) obj, in, 2, objv);

  if (destName != 0) {
    DECR_REF_COUNT(alias);
  }
  DECR_REF_COUNT(objv[1]);
  FREE_TCL_OBJECTS_ON_STACK(objv);
  return result;
}

extern void
XOTclRemovePMethod(Tcl_Interp* in, XOTclObject* obj, char* nm) {
  if (obj->nsPtr)
    NSDeleteCmd(in, obj->nsPtr, nm);
}

extern void
XOTclRemoveIMethod(Tcl_Interp* in, XOTclClass* cl, char* nm) {
  NSDeleteCmd(in, cl->nsPtr, nm);
}

/*
 * obj/cl ClientData setter/getter
 */
extern void
XOTclSetObjClientData(XOTclObject* obj, ClientData data) {
  requireObjectOpt(obj);
  obj->opt->clientData = data;
}
extern ClientData
XOTclGetObjClientData(XOTclObject* obj) {
  return (obj && obj->opt) ? obj->opt->clientData : 0;
}
extern void
XOTclSetClassClientData(XOTclClass* cl, ClientData data) {
  requireClassOpt(cl);
  cl->opt->clientData = data;
}
extern ClientData
XOTclGetClassClientData(XOTclClass* cl) {
  return (cl && cl->opt) ? cl->opt->clientData : 0;
}

static int
XOTclOSetMethod(ClientData cd, Tcl_Interp* in, int objc, Tcl_Obj *objv[]) {
  XOTclObject* obj = (XOTclObject*)cd;
  Tcl_Obj *result;

  if (!obj) return XOTclObjErrType(in, objv[0], "Object");

  if (objc == 2) {
    /*fprintf(stderr,"+++ read var '%s'\n", ObjStr(objv[1]));*/
    result = XOTclOGetInstVar2(obj, in, objv[1], NULL,
			       (TCL_LEAVE_ERR_MSG|TCL_PARSE_PART1));
  } else if (objc == 3) {
    /*fprintf(stderr,"+++ write var '%s' = '%s'\n", 
      ObjStr(objv[1]),ObjStr(objv[2]));*/
    result = XOTclOSetInstVar2(obj, in, objv[1], NULL, objv[2],
			       (TCL_LEAVE_ERR_MSG|TCL_PARSE_PART1));
  } else
    return XOTclObjErrArgCnt(in, obj->cmdName, "set var ?value?");

  if (result) {
    Tcl_SetObjResult(in, result);
    return TCL_OK;
  } else {
    return XOTclVarErrMsg(in, "Can't find result of set ",
			  ObjStr(objv[1]), 0);
  }
}

static int
XOTclSetterMethod(ClientData cd, Tcl_Interp* in, int objc, Tcl_Obj *objv[]) {
  XOTclObject* obj = (XOTclObject*)cd;
  Tcl_Obj *result;

  if (!obj) return XOTclObjErrType(in, objv[0], "Object");

  if (objc == 1)
    result = XOTclOGetInstVar2(obj, in, objv[0], NULL,
			       (TCL_LEAVE_ERR_MSG|TCL_PARSE_PART1));
  else if (objc == 2)
    result = XOTclOSetInstVar2(obj, in, objv[0], NULL, objv[1],
			       (TCL_LEAVE_ERR_MSG|TCL_PARSE_PART1));
  else
    return XOTclObjErrArgCnt(in, obj->cmdName, "<parameter> ?value?");

  if (result) {
    Tcl_SetObjResult(in, result);
    return TCL_OK;
  } else {
    return XOTclVarErrMsg(in, "Can't find result of parameter ",
			  ObjStr(objv[0]), 0);
  }
}

static int
XOTclOUnsetMethod(ClientData cd, Tcl_Interp* in,
		  int objc, Tcl_Obj * CONST objv[]) {
  XOTclObject* obj = (XOTclObject*)cd;
  int i, result = TCL_ERROR;
  VarFrameDecls;
  int flgs = TCL_LEAVE_ERR_MSG | TCL_PARSE_PART1;

  if (!obj) return XOTclObjErrType(in, objv[0], "Object");
  if (objc < 2) return XOTclObjErrArgCnt(in, obj->cmdName, "unset ?vars?");

  VarFrameSwitchToObj(in, obj);

  if (obj->nsPtr)
    flgs = flgs|TCL_NAMESPACE_ONLY;

  for (i=1; i<objc; i++) {
    result =  Tcl_UnsetVar2(in, ObjStr(objv[i]), (char *) NULL, flgs);
    if (result != TCL_OK) break;
  }

  VarFrameRestore(in, obj);

  return result;
}

static int
XOTclOEvalMethod(ClientData cd, Tcl_Interp* in,
			int objc, Tcl_Obj * CONST objv[]) {
  XOTclObject* obj = (XOTclObject*)cd;
  VarFrameDecls;
  int result;
  if (!obj) return XOTclObjErrType(in, objv[0], "Object");
  VarFrameSwitchToObj(in, obj);
  result = Tcl_EvalObjv(in, objc, objv, 0);
  VarFrameRestore(in, obj);
  return result;
}

static int
XOTclOInstVarMethod(ClientData cd, Tcl_Interp* in, int objc, Tcl_Obj *objv[]) {
  XOTclObject  *obj   = (XOTclObject*)cd;
  Tcl_Obj      **ov;
  int          i, oc, result = TCL_OK;
  Interp       *iPtr = (Interp *) in;
  int filterFrameCutOff = 0;
  CallFrame *savedFramePtr = 0, *savedVarFramePtr = 0;

  if (!obj) return XOTclObjErrType(in, objv[0], "Object");
  if (objc < 2) return XOTclObjErrArgCnt(in, obj->cmdName, "instvar ?vars?");

  if (obj && obj->filterStack) {
    filterFrameCutOff =
      FilterCutFrames(in, &savedFramePtr, &savedVarFramePtr);
  }
  if (!iPtr->varFramePtr) {
    if (filterFrameCutOff)
      FilterRestoreFrames(in, savedFramePtr, savedVarFramePtr);
    return XOTclVarErrMsg(in, "instvar used on ", ObjStr(obj->cmdName),
			  ", but callstack is not in procedure scope",
			  (char*) 0);
  }

  for (i=1; i<objc; i++) {
    if ((result = Tcl_ListObjGetElements(in, objv[i], &oc, &ov)) == TCL_OK) {
      char* alias = 0;
      if (oc == 2)
	alias = ObjStr(ov[1]);
      result = GetInstVarIntoCurrentScope(in, obj, ObjStr(ov[0]),
					  alias);
      if (result != TCL_OK)
	break;
    } else
      break;
  }

  if (filterFrameCutOff)
    FilterRestoreFrames(in, savedFramePtr, savedVarFramePtr);

  return result;
}

/*
 * copied from Tcl, since not exported
 */
static char *
VwaitVarProc(clientData, interp, name1, name2, flags)
    ClientData clientData;	/* Pointer to integer to set to 1. */
    Tcl_Interp *interp;		/* Interpreter containing variable. */
    char *name1;		/* Name of variable. */
    char *name2;		/* Second part of variable name. */
    int flags;			/* Information about what happened. */
{
    int *donePtr = (int *) clientData;

    *donePtr = 1;
    return (char *) NULL;
}
static int
XOTclOVwaitMethod(ClientData cd, Tcl_Interp* in, int objc, Tcl_Obj *objv[]) {
  XOTclObject* obj = (XOTclObject*)cd;
  int done, foundEvent;
  char* nameString;
  int flgs = TCL_TRACE_WRITES|TCL_TRACE_UNSETS;
  VarFrameDecls;

  if (!obj) return XOTclObjErrType(in, objv[0], "Object");
  if (objc != 2)
      return XOTclObjErrArgCnt(in, obj->cmdName, "vwait varname");

  nameString = ObjStr(objv[1]);

  /*
   * Make sure the var table exists and the varname is in there
   */
  if (NSRequireVariableOnObj(in, obj, nameString, flgs) == 0)
    return XOTclVarErrMsg(in, "Can't lookup (and create) variable ",
			  nameString, " on ", ObjStr(obj->cmdName));

  VarFrameSwitchToObj(in, obj);
  /*
   * much of this is copied from Tcl, since we must avoid
   * access with flag TCL_GLOBAL_ONLY ... doesn't work on
   * obj->varTable vars
   */
  if (Tcl_TraceVar(in, nameString, flgs,
		   (Tcl_VarTraceProc *)VwaitVarProc,
		   (ClientData) &done) != TCL_OK) {
    return TCL_ERROR;
  }
  done = 0;
  foundEvent = 1;
  while (!done && foundEvent) {
    foundEvent = Tcl_DoOneEvent(TCL_ALL_EVENTS);
  }
  Tcl_UntraceVar(in, nameString, flgs,
		 (Tcl_VarTraceProc *)VwaitVarProc,
		 (ClientData) &done);
  VarFrameRestore(in, obj);
  /*
   * Clear out the interpreter's result, since it may have been set
   * by event handlers.
   */
  Tcl_ResetResult(in);

  if (!foundEvent) {
    Tcl_AppendResult(in, "can't wait for variable \"", nameString,
		     "\":  would wait forever", (char *) NULL);
    return TCL_ERROR;
  }
  return TCL_OK;
}

static int
XOTclOInvariantsMethod(ClientData cd, Tcl_Interp* in, int objc, Tcl_Obj *objv[]) {
  XOTclObject* obj = (XOTclObject*)cd;

  if (!obj) return XOTclObjErrType(in, objv[0], "Object");
  if (objc != 2)
      return XOTclObjErrArgCnt(in, obj->cmdName, "invar <invariantList>");

  requireObjectOpt(obj);

  if (obj->opt->assertions)
    TclObjListDeleteList(obj->opt->assertions->invariants);
  else
    obj->opt->assertions = AssertionCreateStore();

  obj->opt->assertions->invariants = AssertionNewList(in, objv[1]);
  return TCL_OK;
}

static int
XOTclOAutonameMethod(ClientData cd, Tcl_Interp* in,
		     int objc, Tcl_Obj *CONST objv[]) {
  XOTclObject* obj = (XOTclObject*)cd;
  int instanceOpt = 0, resetOpt = 0;
  Tcl_Obj* autoname;

  if (!obj) return XOTclObjErrType(in, objv[0], "Object");
  if (objc == 3) {
    instanceOpt = (strcmp(ObjStr(objv[1]), "-instance") == 0);
    resetOpt = (strcmp(ObjStr(objv[1]), "-reset") == 0);
  }
  if ((objc < 2 || objc > 3) || (objc == 3 && !instanceOpt && !resetOpt))
    return XOTclObjErrArgCnt(in, obj->cmdName,
			     "autoname [-instance | -reset] name");
  autoname = AutonameIncr(in, objv[objc-1], obj, instanceOpt, resetOpt);

  if (autoname) {
    Tcl_SetObjResult(in, autoname);
    DECR_REF_COUNT(autoname);
  }
  else
    return XOTclVarErrMsg(in,
      "Autoname failed. Probably format string (with %) was not well-formed", 0);

  return TCL_OK;
}

static int
XOTclOCheckMethod(ClientData cd, Tcl_Interp* in, int objc, Tcl_Obj *objv[]) {
  XOTclObject* obj = (XOTclObject*)cd;
  int ocArgs; Tcl_Obj** ovArgs;
  int i;

  if (!obj) return XOTclObjErrType(in, objv[0], "Object");
  if (objc != 2)
    return XOTclObjErrArgCnt(in, obj->cmdName,
			     "check (?all? ?pre? ?post? ?invar? ?instinvar?)");

  requireObjectOpt(obj);

  obj->opt->checkoptions = CHECK_NONE;
  if (Tcl_ListObjGetElements(in, objv[1], &ocArgs, &ovArgs) == TCL_OK
      && ocArgs > 0) {
    for (i = 0; i < ocArgs; i++) {
      char* option = ObjStr(ovArgs[i]);
      if (option != 0) {
	switch (*option) {
	case 'i':
	  if (strcmp(option, "instinvar") == 0) {
	    optionSet(obj->opt->checkoptions, CHECK_CLINVAR);
	  } else if (strcmp(option, "invar") == 0) {
	    optionSet(obj->opt->checkoptions, CHECK_OBJINVAR);
	  }
	  break;
	case 'p':
	  if (strcmp(option, "pre") == 0) {
	    optionSet(obj->opt->checkoptions, CHECK_PRE);
	  } else if (strcmp(option, "post") == 0) {
	    optionSet(obj->opt->checkoptions, CHECK_POST);
	  }
	  break;
	case 'a':
	  if (strcmp(option, "all") == 0) {
	    optionSet(obj->opt->checkoptions, CHECK_ALL);
	  }
	  break;	
	}
      }
    }
  }
  if (obj->opt->checkoptions == CHECK_NONE && ocArgs>0) {
    return XOTclVarErrMsg(in, "Unknown check option in command '",
			  ObjStr(obj->cmdName), " ", ObjStr(objv[0]),
			  " ", ObjStr(objv[1]),
			  "', valid: all pre post invar instinvar",
			  (char*) NULL);
  }

  Tcl_ResetResult(in);
  return TCL_OK;
}

static int
XOTclOMixinMethod(ClientData cd, Tcl_Interp* in, int objc, Tcl_Obj *objv[]) {
  int oc; Tcl_Obj** ov;
  XOTclObject* obj = (XOTclObject*)cd;
  int i, result = TCL_OK;

  if (!obj) return XOTclObjErrType(in, objv[0], "Object");
  if (objc < 2)
    return XOTclObjErrArgCnt(in, obj->cmdName, "mixin <classes> ?args?");

  if (Tcl_ListObjGetElements(in, objv[1], &oc, &ov)!= TCL_OK)
    return TCL_ERROR;

  if (obj->opt)
    CmdListRemoveList(&obj->opt->mixins, 0);
  obj->flags &= ~XOTCL_MIXIN_ORDER_VALID;
  requireObjectOpt(obj);

  /*
   * since mixin procs may be used as filters -> we have to invalidate
   */
  obj->flags &= ~XOTCL_FILTER_ORDER_VALID;

  for (i = 0; i < oc; i++) {
    result = MixinAdd(in, &obj->opt->mixins, ov[i]);
    if (result != TCL_OK)
      return result;
  }

  MixinComputeDefined(in, obj);
  FilterComputeDefined(in, obj);

  return result;
}

static int
XOTclOFilterMethod(ClientData cd, Tcl_Interp* in, int objc, Tcl_Obj *objv[]) {
  int oc; Tcl_Obj** ov;
  XOTclObject* obj = (XOTclObject*)cd;
  int i, result = TCL_OK;

  if (!obj) return XOTclObjErrType(in, objv[0], "Object");
  if (objc < 2)
    return XOTclObjErrArgCnt(in, obj->cmdName, "filter filterNameList");

  if (Tcl_ListObjGetElements(in, objv[1], &oc, &ov) != TCL_OK)
    return TCL_ERROR;

  if (obj->opt) {
    CmdListRemoveList(&obj->opt->filters, FilterGuardDelList);
  }

  obj->flags &= ~XOTCL_FILTER_ORDER_VALID;
  requireObjectOpt(obj);

  for (i = 0; i < oc; i ++) {
    result = FilterAdd(in, &obj->opt->filters, ov[i], obj, 0);
    if (result != TCL_OK)
      return result;
  }
  /*FilterComputeDefined(in, obj);*/
  return result;
}

static int
XOTclOFilterGuardMethod(ClientData cd, Tcl_Interp* in, int objc, Tcl_Obj *objv[]) {
  XOTclObject* obj = (XOTclObject*)cd;
  XOTclCmdList* h;

  if (!obj) return XOTclObjErrType(in, objv[0], "Object");
  if (objc != 3)
    return XOTclObjErrArgCnt(in, obj->cmdName, "filterguard filtername filterGuards");

  if (obj->opt && obj->opt->filters) {
    h = CmdListFindNameInList(in, ObjStr(objv[1]), obj->opt->filters);
    if (h) {
      if (h->clientData)
	FilterGuardDelList((XOTclCmdList*) h->clientData);
      FilterGuardAdd(in, h, objv[2]);
      obj->flags &= ~XOTCL_FILTER_ORDER_VALID;
      return TCL_OK;
    }
  }

  return XOTclVarErrMsg(in, "Filterguard: can't find filter ",
			  ObjStr(objv[1]), " on ", ObjStr(obj->cmdName),
			  0);
}

static int
XOTclOGetGuardedScopeMethod(ClientData cd, Tcl_Interp* in, int objc, Tcl_Obj *objv[]) {
  XOTclObject* obj = (XOTclObject*)cd;
  XOTclCallStack *cs = &RUNTIME_STATE(in)->cs;
  XOTclCallStackContent *csc = cs->top,
    *bot =  cs->content;
  int level = 0;

  if (!obj) return XOTclObjErrType(in, objv[0], "Object");
  if (objc != 1)
    return XOTclObjErrArgCnt(in, obj->cmdName, "getGuardedScope");

  if (!cs->filtersForbidden)
    XOTclVarErrMsg(in, "getGuardedScope called outside of a filter guard", 0);

  while (csc > bot) {
    if (isFilterEntry(csc))
      break;
    level ++;
    csc--;
  }
  if (csc <= bot)
    XOTclVarErrMsg(in, "getGuardedScope can't find filter callstack entry\n");

  Tcl_SetObjResult(in, Tcl_NewIntObj(level));
  return TCL_OK;
}


/*
 *  Searches for filter on [self] and returns fully qualified name
 *  if it is not found it returns an empty string
 */
static int
XOTclOFilterSearchMethod(ClientData cd, Tcl_Interp* in, int objc, Tcl_Obj *objv[]) {
  XOTclObject *obj = (XOTclObject*)cd;
  char *methodName;
  XOTclCmdList *cmdList;
  XOTclClass* fcl;
  XOTclObject* fobj;

  if (!obj) return XOTclObjErrType(in, objv[0], "Object");
  if (objc < 2) return XOTclObjErrArgCnt(in, obj->cmdName,
					 "filtersearch name");
  Tcl_ResetResult(in);

  if (!(obj->flags & XOTCL_FILTER_ORDER_VALID))
    FilterComputeDefined(in, obj);
  if (!(obj->flags & XOTCL_FILTER_ORDER_DEFINED))
    return TCL_OK;

  methodName = ObjStr(objv[1]);
  cmdList = obj->filterOrder;

  while (cmdList) {
    CONST84 char* filterName =
      Tcl_GetCommandName(in, (Tcl_Command)cmdList->cmdPtr);
    if (filterName[0] == methodName[0] &&
	!strcmp(filterName, methodName))
      break;
    cmdList = cmdList->next;
  }

  if (!cmdList)
    return TCL_OK;

  fcl = GetClass(in, NSCutXOTclClasses(cmdList->cmdPtr->nsPtr->fullName));
  fobj = fcl ? 0 : GetObject(in,
			     cmdList->cmdPtr->nsPtr->fullName);
  Tcl_SetObjResult(in, getFullProcQualifier(in, methodName, fobj, fcl));
  return TCL_OK;
}

static int
XOTclOProcSearchMethod(ClientData cd, Tcl_Interp* in, int objc, Tcl_Obj *objv[]) {
  XOTclObject* obj = (XOTclObject*)cd;
  XOTclClass* cl=0;
  Command *cmdPtr = 0;
  char *simpleName, *methodName;

  if (!obj) return XOTclObjErrType(in, objv[0], "Object");
  if (objc < 2) return XOTclObjErrArgCnt(in, obj->cmdName,
    "procsearch name");

  Tcl_ResetResult(in);

  methodName = ObjStr(objv[1]);

  if (obj->nsPtr)
    cmdPtr = FindMethod(methodName, obj->nsPtr);

  if (!cmdPtr) {
    if (!(obj->flags & XOTCL_MIXIN_ORDER_VALID))
      MixinComputeDefined(in, obj);

    if ((obj->flags & XOTCL_MIXIN_ORDER_VALID) &&
	(obj->flags & XOTCL_MIXIN_ORDER_DEFINED)) {
      XOTclCmdList* mixinList = obj->mixinOrder;
      while (mixinList) {
	XOTclClass* mcl = GetClass(in,
          (char*) Tcl_GetCommandName(in, (Tcl_Command)mixinList->cmdPtr));
	cl = SearchCMethod(mcl, NULL, methodName, &cmdPtr);
	if (cmdPtr)
	  break;
	mixinList = mixinList->next;
      }
    }
  }

  if (!cmdPtr) {
    cl = SearchCMethod(obj->cl, NULL, methodName, &cmdPtr);
  }

  if (cmdPtr) {
    XOTclClass* pcl = GetClass(in,(char*) NSCutXOTclClasses(cmdPtr->nsPtr->fullName));
    XOTclObject* pobj = pcl ? 0 : GetObject(in,
      cmdPtr->nsPtr->fullName);
    simpleName = (char*)
      Tcl_GetCommandName(in, (Tcl_Command)cmdPtr);
    Tcl_SetObjResult(in, getFullProcQualifier(in, simpleName, pobj, pcl));
  }
  return TCL_OK;
}

static int
XOTclORequireNamespaceMethod(ClientData cd, Tcl_Interp* in,
			     int objc, Tcl_Obj *objv[]) {
  XOTclObject* obj = (XOTclObject*)cd;

  if (!obj) return XOTclObjErrType(in, objv[0], "Object");
  if (objc != 1) return XOTclObjErrArgCnt(in, obj->cmdName,
    "requireNamespace");

  requireObjNamespace(in, obj);
  return TCL_OK;
}

static int
XOTclOConfigureMethod(ClientData cd, Tcl_Interp* in,
			int objc, Tcl_Obj *objv[]) {
  XOTclObject* obj = (XOTclObject*)cd;
  int i, result = TCL_OK;

  if (!obj) return XOTclObjErrType(in, objv[0], "Object");
  if (objc < 1) return XOTclObjErrArgCnt(in, obj->cmdName,
					 "configure ?args?");
  /*
   * Check Args for '-' followed by an alpha character
   * these arguments are called as methods
   */
  for (i=1; i < objc; i++) {
    char *flag = ObjStr(objv[i]);

    if (isDashArg(flag)) {
      int j, nrArgs;
      Tcl_Obj *method = Tcl_NewStringObj(flag+1,-1);
      INCR_REF_COUNT(method);

      for (j = i+1, nrArgs = 0; j < objc; j++, nrArgs++) {
	if (isDashArg(ObjStr(objv[j])))
	  break;
      }
      result = callMethod((ClientData)obj, in,
			  method, nrArgs+2, objv+i+1,
			  XOTCL_CM_NO_UNKNOWN);
      DECR_REF_COUNT(method);
      if (result != TCL_OK) {
	Tcl_AppendResult(in, " during '", ObjStr(obj->cmdName), " ",
			 flag, "'", 0);
	return result;
      }

      i += nrArgs;
    }
  }
  return result;
}


/*
 * class method implementations
 */

static int
XOTclCInstDestroyMethod(ClientData cd, Tcl_Interp* in,
		    int objc, Tcl_Obj *CONST objv[]) {
  XOTclClass* cl = XOTclObjectToClass(cd);
  XOTclObject* delobj;

  if (!cl) return XOTclObjErrType(in, objv[0], "Class");
  if (objc < 2)
      return XOTclObjErrArgCnt(in, cl->object.cmdName,
			       "instdestroy <obj/cl>");

  if (GetXOTclObjectFromObj(in, objv[1], &delobj) != TCL_OK)
    return XOTclVarErrMsg(in, "Can't destroy object ",
			  ObjStr(objv[1]), " that does not exist.",
			  0);
  /*
   * latch, and call delete command if not already in progress
   */
  delobj->flags |= XOTCL_DESTROY_CALLED;
  RUNTIME_STATE(in)->callIsDestroy = 1;

  if (RUNTIME_STATE(in)->exitHandlerDestroyRound !=
      XOTCL_EXITHANDLER_ON_SOFT_DESTROY) {
    CallStackDestroyObject(in, delobj);
  }

  return TCL_OK;
}

static int
XOTclCAllocMethod(ClientData cd, Tcl_Interp* in, int objc, Tcl_Obj *objv[]) {
  XOTclClass* cl = XOTclObjectToClass(cd);
  XOTclClass* newcl;
  XOTclObject* newobj;

  if (!cl) return XOTclObjErrType(in, objv[0], "Class");
  if (objc < 2)
      return XOTclObjErrArgCnt(in, cl->object.cmdName,
			       "alloc <obj/cl> ?args?");
  /*
   * if the lookup via GetObject for the object succeeds,
   * the object exists already,
   * and we do not overwrite it, but re-create it
   */
  if (GetXOTclObjectFromObj(in, objv[1], &newobj) == TCL_OK) {
    int result = doCleanup(in, newobj, &cl->object, objc, objv);
    if (result != TCL_OK)
      return result;

    Tcl_ResetResult(in);
    return TCL_OK;
  } else {
    /*
     * create a new object from scratch
     */
    if (IsMetaClass(in, cl)) {
      /*
       * if the base class is a meta-class, we create a class
       */
      newcl = PrimitiveCCreate(in, ObjStr(objv[1]), cl);
      if (newcl == 0)
	return XOTclVarErrMsg(in, "Class alloc failed for '",ObjStr(objv[1]),
			      "' (possibly parent namespace does not exist)",NULL);
    } else {
      /*
       * if the base class is an ordinary class, we create an object
       */
      newobj = PrimitiveOCreate(in, ObjStr(objv[1]), cl);
      if (newobj == 0)
	return XOTclVarErrMsg(in, "Object alloc failed for '",ObjStr(objv[1]),
			      "' (possibly parent namespace does not exist)",NULL);
    }
  }
  Tcl_ResetResult(in);
  return TCL_OK;
}


static int
XOTclCCreateMethod(ClientData cd, Tcl_Interp* in, int objc, Tcl_Obj *objv[]) {
  XOTclClass* cl = XOTclObjectToClass(cd);
  XOTclObject* obj = &cl->object;
  XOTclObject* newobj;

  int result, nameLength;
  char *objName;
  Tcl_Obj* nameObj;

  if (!cl) return XOTclObjErrType(in, objv[0], "Class");
  if (!obj) return XOTclObjErrType(in, objv[0], "Object");
  if (objc < 2)
    return XOTclObjErrArgCnt(in, cl->object.cmdName, "create <obj> ?args?");

  /*
   * Check whether we have to call recreate (i.e. when the
   * object exists already)
   */
  /* fprintf(stderr,"+++ create objv[1] = %p (%d) %s\n",
     objv[1],objv[1]->refCount,ObjStr(objv[1])); */
  
  if (GetXOTclObjectFromObj(in, objv[1], &newobj) == TCL_OK) {
    /* cleanup */
    /*fprintf(stderr, "+++ recreate, call cleanup ... %s\n", ObjStr(objv[1]));*/
    result = doCleanup(in, newobj, obj, objc, objv);
    if (result != TCL_OK)
      return result;

    /*XOTclObjectRefCountDecr(newobj);*/

    /* call recreate --> initialization */
    result = callMethod((ClientData) obj, in,
                         global_objects[RECREATE], objc+1, objv+1, 0);
    if (result != TCL_OK)
      return result;

    nameObj = newobj->cmdName;
    INCR_REF_COUNT(nameObj);

    objTrace("RECREATE", newobj);

  } else {
    objName = Tcl_GetStringFromObj(objv[1],&nameLength);
    if (!NSCheckColons(objName, nameLength))
      return XOTclVarErrMsg(in, "Cannot create object -- illegal name '",
			    objName, "'", 0);
    /*
     * call "alloc"
     */
    /* fprintf(stderr, "alloc ... %s\n", ObjStr(objv[1]));*/
    result = callMethod((ClientData) obj, in,
                         global_objects[ALLOC], objc+1, objv+1, 0);
    if (result != TCL_OK)
      return result;

    newobj = GetObject(in, objName);
    if (newobj == 0)
      return XOTclErrMsg(in, "couldn't find result of alloc", TCL_STATIC);
    nameObj = newobj->cmdName;

    (void)RemoveInstance(newobj, newobj->cl);
    AddInstance(newobj, cl);
    objTrace("CREATE", newobj);

    /* in case, the object is destroyed during initialization, we incr refcount */
    INCR_REF_COUNT(nameObj);
    result = doObjInitialization(in, newobj, objc, objv);
  }
  /* fprintf(stderr, "alloc -- end ... %s\n", ObjStr(objv[1]));*/

  if (result == TCL_OK) {
    Tcl_SetObjResult(in, nameObj);
  }
  DECR_REF_COUNT(nameObj);

  return result;
}

static char *
XOTclUnsetTrace(ClientData cd, Tcl_Interp *in, CONST84 char *name, CONST84 char *name2, int flags)
{
  Tcl_Obj *obj = (Tcl_Obj *)cd;
  XOTclObject *o;
  char *result = NULL;

  if (GetXOTclObjectFromObj(in, obj, &o) == TCL_OK) {
    Tcl_Obj *res = Tcl_GetObjResult(in); /* save the result */
    INCR_REF_COUNT(res);

    if (callMethod((ClientData)o, in, global_objects[DESTROY],2,0,0) != TCL_OK) {
      result = "Destroy for volatile object failed";
    } else 
      result = "No XOTcl Object passed";

    Tcl_SetObjResult(in, res);  /* restore the result */
    DECR_REF_COUNT(res);
  }
  DECR_REF_COUNT(obj);
  return result;
}

static int
XOTclCNewMethod(ClientData cd, Tcl_Interp* in, int objc, Tcl_Obj *objv[]) {
  XOTclClass  *cl = XOTclObjectToClass(cd);
  XOTclObject *obj = &cl->object, *child = NULL;
  Tcl_Obj *fullname, *autoname;
  int result, offset=1, isvolatile=0, isrefcount=0, i;

  if (!cl) return XOTclObjErrType(in, objv[0], "Class");
  if (!obj) return XOTclObjErrType(in, objv[0], "Object");
  if (objc < 1)
    return XOTclObjErrArgCnt(in, cl->object.cmdName,
			     "new [-childof obj] ?args?");

  for (i=1; i<objc; i++) {
    char *option = ObjStr(objv[i]);
    if (strcmp(option,"-volatile")==0) {
      isvolatile = 1;
      offset += 1;
    } else if (strcmp(option,"-childof")==0 && i<objc-1) {
      offset += 2;
      if (GetXOTclObjectFromObj(in, objv[i+1], &child) != TCL_OK) {
	return XOTclErrMsg(in, "not a valid object specified as child\n", TCL_STATIC);
      }
    } else if (strcmp(option,"-refcount")==0) {
      isrefcount = 1;
      offset += 1;
    } 
    else
      break;
  }
  
  if (child) {
    fullname = Tcl_NewStringObj(ObjStr(child->cmdName),-1);
    Tcl_AppendToObj(fullname, "::", 2);
  } else {
    fullname = Tcl_NewStringObj("::xotcl::",9);
  }
  INCR_REF_COUNT(fullname);

  autoname = AutonameIncr(in, global_objects[NEWOBJ], 
			  (XOTclObject *)RUNTIME_STATE(in)->theClass, 0, 0);
  /*fprintf(stderr, "AUTO DONE %p %s\n",fullname,ObjStr(fullname));*/

#ifdef PRE81
  Tcl_AppendToObj(fullname, ObjStr(autoname), -1);
#else
  Tcl_AppendObjToObj(fullname, autoname);
#endif

  objc -= offset;
  {
    DEFINE_NEW_TCL_OBJECTS_ON_STACK(objc+2, ov);

    ov[0] = global_objects[CREATE];
    ov[1] = fullname;
    if (objc >= 1) 
      memcpy(ov+2, objv+offset, sizeof(Tcl_Obj *)*objc);
    
    result = XOTclCCreateMethod(cd, in, objc+2, ov);
    FREE_TCL_OBJECTS_ON_STACK(ov);
  }

  if (result == TCL_OK) {
    if (isvolatile) {
      /* register trace for for the magic variable to destroy the object */
      Tcl_Obj *obj = Tcl_GetObjResult(in);
      char *vn = ObjStr(autoname);
      INCR_REF_COUNT(obj);
      Tcl_SetVar2(in, vn, 0, ObjStr(fullname), 0);
      Tcl_TraceVar(in, vn, TCL_TRACE_UNSETS, 
		   (Tcl_VarTraceProc*)XOTclUnsetTrace,
		   (ClientData)obj); 
    }
#if REFCOUNTED
    if (isrefcount) {
      Tcl_Obj *obj = Tcl_GetObjResult(in);
      XOTclObject *o = (XOTclObject*) obj->internalRep.otherValuePtr;
      o->flags |= XOTCL_REFCOUNTED;
      o->teardown = in;
      DECR_REF_COUNT(obj);
    }
#endif
  }

  DECR_REF_COUNT(fullname);
  DECR_REF_COUNT(autoname);
  return result;
}


static int
XOTclCRecreateMethod(ClientData cd, Tcl_Interp* in,
		     int objc, Tcl_Obj *objv[]) {
  XOTclClass* cl = XOTclObjectToClass(cd);
  XOTclObject* obj = &cl->object, *newobj;
  int result;

  if (!cl) return XOTclObjErrType(in, objv[0], "Class");
  if (!obj) return XOTclObjErrType(in, objv[0], "Object");
  if (objc < 2)
    return XOTclObjErrArgCnt(in, cl->object.cmdName, "recreate <obj> ?args?");

  if (GetXOTclObjectFromObj(in, objv[1], &newobj) != TCL_OK)
    return XOTclVarErrMsg(in, "can't recreate not existing obj ",
			  ObjStr(objv[1]), 0);

  INCR_REF_COUNT(objv[1]);

  result = doObjInitialization(in, newobj, objc, objv);

  if (result == TCL_OK)
    Tcl_SetObjResult(in, objv[1]);

  DECR_REF_COUNT(objv[1]);

  return result;
}

static int
XOTclCSuperClassMethod(ClientData cd, Tcl_Interp* in,
                      int objc, Tcl_Obj *objv[]) {
  XOTclClass* cl = XOTclObjectToClass(cd);
  XOTclClasses* osl = 0;
  int oc; Tcl_Obj** ov;
  XOTclClass** scl = 0;
  int reversed = 0;
  int i, j;
  XOTclClasses* filterCheck = ComputeOrder(cl, Super);

  if (!cl) return XOTclObjErrType(in, objv[0], "Class");
  if (objc != 2)
    return XOTclObjErrArgCnt(in, cl->object.cmdName, "superclass <classes>");

  if (Tcl_ListObjGetElements(in, objv[1], &oc, &ov)!= TCL_OK || oc==0)
    return XOTclObjErrArgCnt(in, cl->object.cmdName, "superclass <classes>");


  /*
   * we have to remove all dependent superclass filter referenced
   * by class or one of its subclasses
   *
   * do not check the class "cl" itself (first entry in
   * filterCheck class list)
   */
  if (filterCheck)
    filterCheck = filterCheck->next;
  while (filterCheck) {
    FilterRemoveDependentFilterCmds(cl, filterCheck->cl);
    filterCheck = filterCheck->next;
  }

  /* invalidate all interceptors orders of instances of this
     and of all depended classes */
  MixinInvalidateObjOrders(cl);
  FilterInvalidateObjOrders(in, cl);

  scl = NEW_ARRAY(XOTclClass*,oc);
  for (i = 0; i < oc; i++) {
    if (GetXOTclClassFromObj(in, ov[i], &scl[i]) != TCL_OK) {
      FREE(XOTclClass**, scl);
      return XOTclErrBadVal(in, "a list of classes", ObjStr(objv[1]));
    }
  }

  /*
   * check that superclasses don't precede their classes
   */

  for (i = 0; i < oc; i++) {
    if (reversed != 0) break;
    for (j = i+1; j < oc; j++) {
      XOTclClasses* dl = ComputeOrder(scl[j], Super);
      if (reversed != 0) break;
      while (dl != 0) {
        if (dl->cl == scl[i]) break;
        dl = dl->next;
      }
      if (dl != 0) reversed = 1;
    }
  }

  if (reversed != 0) {
    return XOTclErrBadVal(in, "classes in dependence order", ObjStr(objv[1]));
  }

  while (cl->super != 0) {

    /*
     * build up an old superclass list in case we need to revert
     */

    XOTclClass* sc = cl->super->cl;
    XOTclClasses* l = osl;
    osl = NEW(XOTclClasses);
    osl->cl = sc;
    osl->next = l;
    (void)RemoveSuper(cl, cl->super->cl);
  }
  for (i = 0; i < oc; i++)
    AddSuper(cl, scl[i]);
  FREE(XOTclClass**,scl);
  FlushPrecedences(cl);

  if (!ComputeOrder(cl, Super)) {

    /*
     * cycle in the superclass graph, backtrack
     */

    XOTclClasses* l;
    while (cl->super != 0) (void)RemoveSuper(cl, cl->super->cl);
    for (l = osl; l != 0; l = l->next) AddSuper(cl, l->cl);
    XOTclRemoveClasses(osl);
    return XOTclErrBadVal(in, "a cycle-free graph", ObjStr(objv[1]));
  }
  XOTclRemoveClasses(osl);

  /* if there are no more super classes add the Object
     class as superclasses */
  if (cl->super == 0)
    AddSuper(cl, RUNTIME_STATE(in)->theObject);

  Tcl_ResetResult(in);
  return TCL_OK;
}

static int
XOTclCInfoMethod(ClientData cd, Tcl_Interp* in, int objc, Tcl_Obj * CONST objv[]) {
  XOTclClass* cl = XOTclObjectToClass(cd);
  Namespace *nsp = (Namespace *)cl->nsPtr;
  char *pattern, *cmd;
  int modifiers = 0;

  if (!cl)
    return XOTclObjErrType(in, objv[0], "Class");
  if (objc < 2)
    return XOTclObjErrArgCnt(in, cl->object.cmdName, "info <opt> ?args?");

  cmd = ObjStr(objv[1]);
  pattern = (objc > 2) ? ObjStr(objv[2]) : 0;

  /*
   * check for "-" modifiers
   */
  if (pattern && *pattern == '-') {
    modifiers = countModifiers(objc, objv);
    pattern = (objc > 2+modifiers) ? ObjStr(objv[2+modifiers]) : 0;
  }

  switch (*cmd) {
  case 'c':
    if (!strcmp(cmd, "classchildren")) {
      if (objc > 3 || modifiers > 0)
	return XOTclObjErrArgCnt(in, cl->object.cmdName,
				 "info classchildren ?pat?");
      return ListClassChildren(in, cl, pattern);
    } else if (!strcmp(cmd, "classparent")) {
      if (objc > 2 || modifiers > 0)
	return XOTclObjErrArgCnt(in, cl->object.cmdName,
				 "info classparent");
      return ListParent(in, &cl->object);
    }
    break;

  case 'h':
    if (!strcmp(cmd, "heritage")) {
      if (objc > 3 || modifiers > 0)
	return XOTclObjErrArgCnt(in,
				 cl->object.cmdName,
				 "info heritage ?pat?");
      return ListHeritage(in, cl, pattern);
    }
    break;

  case 'i':
    if (cmd[1] == 'n' && cmd[2] == 's' && cmd[3] == 't') {
      char* cmdTail = cmd + 4;
      switch (*cmdTail) {
      case 'a':
	if (!strcmp(cmdTail, "ances")) {
	  if (objc > 3 || modifiers > 0)
	    return XOTclObjErrArgCnt(in, cl->object.cmdName,
				     "info instances ?pat?");
	  return ListObjPtrHashTable(in, &cl->instances, pattern);
	} else if (!strcmp(cmdTail, "args")) {
	  if (objc != 3 || modifiers > 0)
	    return XOTclObjErrArgCnt(in, cl->object.cmdName,
				     "info instargs <instproc>");
	  return ListProcArgs(in, &nsp->cmdTable, pattern);
	}
	break;

      case 'b':
	if (!strcmp(cmdTail, "body")) {
	  if (objc != 3 || modifiers > 0)
	    return XOTclObjErrArgCnt(in, cl->object.cmdName,
				     "info instbody <instproc>");
	  return ListProcBody(in, &nsp->cmdTable, pattern);
	}
	break;

      case 'c':
	if (!strcmp(cmdTail, "commands")) {
	  if (objc > 3 || modifiers > 0)
	    return XOTclObjErrArgCnt(in, cl->object.cmdName,
				     "info instcommands ?pat?");
	  return ListKeys(in, &nsp->cmdTable, pattern);
	}
	break;

      case 'd':
	if (!strcmp(cmdTail, "default")) {
	  if (objc != 5 || modifiers > 0)
	    return XOTclObjErrArgCnt(in, cl->object.cmdName,
				     "info instdefault <instproc> <arg> <var>");
	  return ListProcDefault(in, &nsp->cmdTable, pattern,
				 ObjStr(objv[3]), ObjStr(objv[4]));
	}
	break;

      case 'f':
	if (!strcmp(cmdTail, "filter")) {
	  int withGuards = 0;
	  if (objc-modifiers > 3)
	    return XOTclObjErrArgCnt(in, cl->object.cmdName,
				     "info instfilter ?-guards? ?pat?");
	  if (modifiers > 0) {
	    withGuards = checkForModifier(objv, modifiers, "-guards");
	    if (withGuards == 0)
	      return XOTclVarErrMsg(in, "info instfilter: unknown modifier ",
				    ObjStr(objv[2]));
	  }
	  if (!cl->opt)
	    return TCL_OK;
	  return FilterInfo(in, cl->opt->instfilters, pattern, withGuards, 0);
	} else if (!strcmp(cmdTail, "filterguard")) {
	  if (objc != 3 || modifiers > 0)
	    return XOTclObjErrArgCnt(in, cl->object.cmdName,
				     "info instfilterguard filter");
	  if (!cl->opt)
	    return TCL_OK;
	  return FilterGuardList(in, cl->opt->instfilters, pattern);
	}
	break;

      case 'i':
	if (!strcmp(cmdTail, "invar")) {
	  if (objc != 2 || modifiers > 0)
	    return XOTclObjErrArgCnt(in, cl->object.cmdName,
				     "info instinvar");
	  if (cl->opt && cl->opt->assertions &&
	      cl->opt->assertions->invariants)
	    Tcl_SetObjResult(in, AssertionList(in,
					    cl->opt->assertions->invariants));
	  return TCL_OK;
	}
	break;

      case 'm':
	if (!strcmp(cmdTail, "mixin")) {
	  if (objc > 3 || modifiers > 0)
	    return XOTclObjErrArgCnt(in, cl->object.cmdName,
				     "info instmixin ?class?");
	  if (!cl->opt)
	    return TCL_OK;
	  return MixinInfo(in, cl->opt->instmixins, pattern);
	}
	break;

      case 'p':
	if (!strcmp(cmdTail, "procs")) {
	  if (objc > 3 || modifiers > 0)
	    return XOTclObjErrArgCnt(in, cl->object.cmdName,
				     "info instprocs ?pat?");
	  return ListMethodKeys(in, &nsp->cmdTable, pattern,
				/*noProcs*/ 0, /*noCmds*/ 1);
	} else if (!strcmp(cmdTail, "pre")) {
	  XOTclProcAssertion* procs;
	  if (objc != 3 || modifiers > 0)
	    return XOTclObjErrArgCnt(in, cl->object.cmdName,
				     "info instpre <proc>");
	  if (cl->opt && cl->opt->assertions) {
	    procs = AssertionFindProcs(cl->opt->assertions, ObjStr(objv[2]));
	    if (procs) Tcl_SetObjResult(in, AssertionList(in, procs->pre));
	  }
	  return TCL_OK;
	} else if (!strcmp(cmdTail, "post")) {
	  XOTclProcAssertion* procs;
	  if (objc != 3 || modifiers > 0)
	    return XOTclObjErrArgCnt(in, cl->object.cmdName,
				     "info instpost <proc>");
	  if (cl->opt && cl->opt->assertions) {
	    procs = AssertionFindProcs(cl->opt->assertions, ObjStr(objv[2]));
	    if (procs) Tcl_SetObjResult(in, AssertionList(in, procs->post));
	  }
	  return TCL_OK;
	}
	break;
      }
    }
    break;

  case 'p':
    if (!strcmp(cmd, "parameterclass")) {
      if (cl->opt && cl->opt->parameterClass) {
	Tcl_SetObjResult(in, cl->opt->parameterClass);
      } else {
	Tcl_SetObjResult(in, global_objects[PARAM_CL]);
      }
      return TCL_OK;
    } else if (!strcmp(cmd, "parameter")) {
      if (cl->parameters) {
	Tcl_SetObjResult(in, cl->parameters);
      } else {
	Tcl_SetObjResult(in, global_objects[EMPTY]);
      }
      return TCL_OK;
    }
    break;

  case 's':
    if (!strcmp(cmd, "superclass")) {
      if (objc > 3 || modifiers > 0)
        return XOTclObjErrArgCnt(in, cl->object.cmdName,
				 "info superclass ?class?");
      return ListSuperclasses(in, cl, pattern);
    } else if (!strcmp(cmd, "subclass")) {
      if (objc > 3 || modifiers > 0)
	return XOTclObjErrArgCnt(in,
				 cl->object.cmdName,
				 "info subclass ?class?");
      return ListSubclasses(in, cl, pattern);
    }
    break;
  }
  return XOTclOInfoMethod(cd, in, objc, (Tcl_Obj **)objv);
}

static int
XOTclCParameterMethod(ClientData cd, Tcl_Interp* in, int objc, Tcl_Obj *objv[]) {
  XOTclClass* cl = XOTclObjectToClass(cd);
  XOTclObject* obj = &cl->object;
  Tcl_Obj     **pv=0;
  Tcl_Obj     **ov=0;
  int         elts, pc, oc, result;
  char*       params;
  if (!cl) return XOTclObjErrType(in, objv[0], "Class");
  if (objc != 2)
    return XOTclObjErrArgCnt(in, cl->object.cmdName,
			     "parameter ?params?");
  if (cl->parameters) {
    DECR_REF_COUNT(cl->parameters);
  }
  /* did we delete the parameters ? */
  params = ObjStr(objv[1]);
  if ((params == NULL) || (*params == '\0')) {
    cl->parameters = 0;
    return TCL_OK;
  }

  /* ok, remember the params */
  cl->parameters = objv[1];
  INCR_REF_COUNT(cl->parameters);

  /* call getter/setter methods in params */
  result = Tcl_ListObjGetElements(in, objv[1], &pc, &pv);
  if (result == TCL_OK) {
    for (elts = 0; elts < pc; elts++) {
      result = Tcl_ListObjGetElements(in, pv[elts], &oc, &ov);
      if (result == TCL_OK && oc > 0 ) {
	result = callParameterMethodWithArg(obj, in,
			    global_objects[MKGETTERSETTER],
			    cl->object.cmdName, 3+oc, ov,0);
      }
      if (result != TCL_OK)
	return result;
    }
  }
  return result;
}
static int
XOTclCParameterClassMethod(ClientData cd, Tcl_Interp* in,
			   int objc, Tcl_Obj *objv[]) {
  XOTclClass* cl = XOTclObjectToClass(cd);
  char* paramClStr;

  if (!cl) return XOTclObjErrType(in, objv[0], "Class");
  if (objc != 2)
    return XOTclObjErrArgCnt(in, cl->object.cmdName,
			     "parameterclass cl");

  paramClStr = ObjStr(objv[1]);
  if (cl->opt && cl->opt->parameterClass) {
    DECR_REF_COUNT(cl->opt->parameterClass);
  }
  if ((paramClStr == NULL) || (*paramClStr == '\0')) {
    if (cl->opt)
      cl->opt->parameterClass = 0;
  } else {
    requireClassOpt(cl);
    cl->opt->parameterClass = objv[1];
    INCR_REF_COUNT(cl->opt->parameterClass);
  }
  return TCL_OK;
}

static int
XOTclCInstParameterCmdMethod(ClientData cd, Tcl_Interp* in,
		      int objc, Tcl_Obj * CONST objv[]) {
  XOTclClass* cl = XOTclObjectToClass(cd);

  if (!cl) return XOTclObjErrType(in, objv[0], "Class");
  if (objc < 2) return XOTclObjErrArgCnt(in, cl->object.cmdName,
					 "instparametercmd name");
  XOTclAddIMethod(in, cl, ObjStr(objv[1]),
		  (Tcl_ObjCmdProc*)XOTclSetterMethod, 0, 0);
  return TCL_OK;
}
static int
XOTclCParameterCmdMethod(ClientData cd, Tcl_Interp* in,
		      int objc, Tcl_Obj * CONST objv[]) {
  XOTclObject* obj = (XOTclObject*) cd;

  if (objc < 2) return XOTclObjErrArgCnt(in, obj->cmdName,
					 "parametercmd name");
  XOTclAddPMethod(in, obj, ObjStr(objv[1]),
		  (Tcl_ObjCmdProc*)XOTclSetterMethod, 0, 0);
  return TCL_OK;
}

static int
XOTclCInstTclCmdMethod(ClientData cd, Tcl_Interp* in,
		      int objc, Tcl_Obj * CONST objv[]) {
  XOTclClass* cl = XOTclObjectToClass(cd);

  if (!cl) return XOTclObjErrType(in, objv[0], "Class");
  if (objc < 2) return XOTclObjErrArgCnt(in, cl->object.cmdName,
					 "insttclcmd name");
  XOTclAddIMethod(in, cl, ObjStr(objv[1]),
		  (Tcl_ObjCmdProc*)XOTclOEvalMethod, 0, 0);
  return TCL_OK;
}
static int
XOTclCTclCmdMethod(ClientData cd, Tcl_Interp* in,
		      int objc, Tcl_Obj * CONST objv[]) {
  XOTclObject* obj = (XOTclObject*) cd;

  if (objc < 2) return XOTclObjErrArgCnt(in, obj->cmdName,
					 "tclcmd name");
  XOTclAddPMethod(in, obj, ObjStr(objv[1]),
		  (Tcl_ObjCmdProc*)XOTclOEvalMethod, (ClientData) obj, 0);
  return TCL_OK;
}


static int
XOTclCInstProcMethod(ClientData cd, Tcl_Interp* in, int objc, Tcl_Obj *objv[]) {
  XOTclClass* cl = XOTclObjectToClass(cd);
  char *argStr, *bdyStr, *name;

  if (!cl) return XOTclObjErrType(in, objv[0], "Class");
  if (objc != 4 && objc != 6)
    return XOTclObjErrArgCnt(in, cl->object.cmdName,
			     "instproc name args body ?preAssertion postAssertion?");

  argStr = ObjStr(objv[2]);
  bdyStr = ObjStr(objv[3]);
  name = ObjStr(objv[1]);

  if ((cl == RUNTIME_STATE(in)->theObject && isDestroyString(name)) ||
      (cl == RUNTIME_STATE(in)->theClass && isInstDestroyString(name)) ||
      (cl == RUNTIME_STATE(in)->theClass && isAllocString(name)) ||
      (cl == RUNTIME_STATE(in)->theClass && isCreateString(name)))
    return XOTclVarErrMsg(in, className(cl), " instproc: '", name, "' of ",
			  className(cl), " can not be overwritten. Derive a ",
			  "sub-class", (char*) 0);

  if (*argStr == 0 && *bdyStr == 0) {
    if (cl->opt)
      AssertionRemoveProc(cl->opt->assertions, name);
    NSDeleteCmd(in, cl->nsPtr, name);
  } else {
    requireClassOpt(cl);
    if (!cl->opt->assertions)
      cl->opt->assertions = AssertionCreateStore();
    MakeProc(cl->nsPtr, cl->opt->assertions, in, objc, (Tcl_Obj **) objv);
  }

  /* could be a filter or filter inheritance ... update filter orders */
  FilterInvalidateObjOrders(in, cl);

  return TCL_OK;
}

static int
XOTclCInstFilterMethod(ClientData cd, Tcl_Interp* in, int objc, Tcl_Obj *objv[]) {
  XOTclClass* cl = XOTclObjectToClass(cd);
  int i, result = TCL_OK;
  Tcl_Obj **ov; int oc;

  if (!cl) return XOTclObjErrType(in, objv[0], "Class");
  if (objc < 2) return XOTclObjErrArgCnt(in, cl->object.cmdName,
    "instfilter filterNameList");

  if (Tcl_ListObjGetElements(in, objv[1], &oc, &ov) != TCL_OK)
    return TCL_ERROR;

  if (cl->opt)
    CmdListRemoveList(&cl->opt->instfilters, FilterGuardDelList);

  FilterInvalidateObjOrders(in, cl);
  requireClassOpt(cl);

  for (i = 0; i < oc; i ++) {
    result = FilterAdd(in, &cl->opt->instfilters, ov[i], 0, cl);
    if (result != TCL_OK)
      return result;
  }
  return result;
}

static int
XOTclCInstFilterGuardMethod(ClientData cd, Tcl_Interp* in, int objc, Tcl_Obj *objv[]) {
  XOTclClass* cl = XOTclObjectToClass(cd);
  XOTclCmdList* h;

  if (!cl) return XOTclObjErrType(in, objv[0], "Class");
  if (objc != 3) return XOTclObjErrArgCnt(in, cl->object.cmdName,
    "instfilterguard filtername filterGuard");

  if (cl->opt && cl->opt->instfilters) {
    h = CmdListFindNameInList(in, ObjStr(objv[1]), cl->opt->instfilters);
    if (h) {
      if (h->clientData)
	FilterGuardDelList((XOTclCmdList*) h->clientData);
      FilterGuardAdd(in, h, objv[2]);
      FilterInvalidateObjOrders(in, cl);
      return TCL_OK;
    }
  }

  return XOTclVarErrMsg(in, "Filterguard: can't find filter ",
			ObjStr(objv[1]), " on ", ObjStr(cl->object.cmdName),
			0);
}


static int
XOTclCInstMixinMethod(ClientData cd, Tcl_Interp* in, int objc, Tcl_Obj *objv[]) {
  int oc; Tcl_Obj** ov;
  XOTclClass* cl = XOTclObjectToClass(cd);
  int i, result = TCL_OK;

  if (!cl) return XOTclObjErrType(in, objv[0], "Class");
  if (objc < 2) return XOTclObjErrArgCnt(in, cl->object.cmdName,
    "instmixin classList");

  if (Tcl_ListObjGetElements(in, objv[1], &oc, &ov)!= TCL_OK)
    return TCL_ERROR;

  if (cl->opt)
    CmdListRemoveList(&cl->opt->instmixins, 0);

  MixinInvalidateObjOrders(cl);
  /*
   * since mixin procs may be used as filters -> we have to invalidate
   */
  FilterInvalidateObjOrders(in, cl);

  requireClassOpt(cl);

  for (i = 0; i < oc; i++) {
    result = MixinAdd(in, &cl->opt->instmixins, ov[i]);
    if (result != TCL_OK)
      return result;
  }
  return result;
}

static int
XOTclCInvariantsMethod(ClientData cd, Tcl_Interp* in, int objc, Tcl_Obj *objv[]) {
  XOTclClass* cl = XOTclObjectToClass(cd);

  if (!cl) return XOTclObjErrType(in, objv[0], "Class");
  if (objc != 2)
      return XOTclObjErrArgCnt(in, cl->object.cmdName,
			       "instinvar <invariantList>");
  requireClassOpt(cl);

  if (cl->opt->assertions)
    TclObjListDeleteList(cl->opt->assertions->invariants);
  else
    cl->opt->assertions = AssertionCreateStore();

  cl->opt->assertions->invariants = AssertionNewList(in, objv[1]);
  return TCL_OK;
}

static int
XOTclCUnknownMethod(ClientData cd, Tcl_Interp *in,
		    int objc, Tcl_Obj *CONST objv[]) {
  XOTclObject* obj = (XOTclObject*) cd;
  char *self = ObjStr(obj->cmdName);
  int rc;

  if (objc < 2) return XOTclObjErrArgCnt(in, objv[0], "message ?args .. args?");
  if (isCreateString(self))
    return XOTclVarErrMsg(in, "error ", self, ": unable to dispatch '",
			  ObjStr(objv[1]), "'", (char*)NULL);

  rc = callMethod(cd, in, global_objects[CREATE], objc+1, objv+1, 0);
  return rc;
}

/*
 * New Tcl Commands
 */
static int
XOTcl_NSCopyCmds(ClientData cd, Tcl_Interp *in, int objc, Tcl_Obj *CONST objv[]) {
  Interp *iPtr = (Interp *) in;
  Tcl_Command cmd;
  Command *cmdPtr;
  Tcl_Obj* newFullCmdName, *oldFullCmdName;
  char* newName, *oldName, *name;
  Namespace *nsPtr, *newNsPtr;
  Tcl_HashSearch hSrch;
  Tcl_HashEntry* hPtr;

  if (objc != 3)
    return XOTclObjErrArgCnt(in, NULL, "namespace_copycmds fromNs toNs");

  nsPtr = (Namespace*) Tcl_FindNamespace(in, ObjStr(objv[1]),
					 (Tcl_Namespace *) NULL, 0);
  if (!nsPtr)
    return TCL_OK;

  newNsPtr = (Namespace*) Tcl_FindNamespace(in, ObjStr(objv[2]),
					    (Tcl_Namespace *) NULL, 0);
  if (!newNsPtr)
    return XOTclVarErrMsg(in, "CopyCmds: Destination namespace ",
			  ObjStr(objv[2]), " does not exist",0);
  /*
   * copy all procs & commands in the ns
   */
  hPtr = &nsPtr->cmdTable ?
    Tcl_FirstHashEntry(&nsPtr->cmdTable, &hSrch) : 0;
  while (hPtr != NULL) {
    name = Tcl_GetHashKey(&nsPtr->cmdTable, hPtr);

    /*
     * construct full cmd names
     */
    newFullCmdName = Tcl_NewStringObj(newNsPtr->fullName,-1);
    oldFullCmdName = Tcl_NewStringObj(nsPtr->fullName,-1);

    INCR_REF_COUNT(newFullCmdName); INCR_REF_COUNT(oldFullCmdName);
    Tcl_AppendStringsToObj(newFullCmdName, "::", name, (char*)NULL);
    Tcl_AppendStringsToObj(oldFullCmdName, "::", name, (char*)NULL);
    newName = ObjStr(newFullCmdName);
    oldName = ObjStr(oldFullCmdName);

    /*
     * Make sure that the destination command does not already exist.
     * Otherwise: do not copy
     */
    cmd = Tcl_FindCommand(in, newName, 0, 0);
    cmdPtr = (Command *) cmd;
    if (cmdPtr != NULL) {
      /*fprintf(stderr, "%s already exists\n", newName);*/
      if (!GetObject(in, newName)) {
	/* command or instproc will be deleted & than copied */
	Tcl_DeleteCommandFromToken(in, cmd);
      } else {
	/* don't overwrite objects -> will be recreated */
	hPtr = Tcl_NextHashEntry(&hSrch);
	DECR_REF_COUNT(newFullCmdName); DECR_REF_COUNT(oldFullCmdName);
	continue;
      }
    }

    /*
     * Find the existing command. An error is returned if simpleName can't
     * be found
     */
    cmd = Tcl_FindCommand(in, oldName, 0, 0);
    cmdPtr = (Command *) cmd;
    if (cmdPtr == NULL) {
      Tcl_AppendStringsToObj(Tcl_GetObjResult(in), "can't copy ", " \"",
			     oldName, "\": command doesn't exist", (char *) NULL);
      DECR_REF_COUNT(newFullCmdName); DECR_REF_COUNT(oldFullCmdName);
      return TCL_ERROR;
    }
    /*
     * Do not copy Objects or Classes
     */
    if (!GetObject(in, oldName)) {
      if (TclIsProc(cmdPtr)) {
	Proc *procPtr = TclFindProc(iPtr, oldName);
	Tcl_Obj *arglistObj;
	CompiledLocal *localPtr;

	/*
	 * Build a list containing the arguments of the proc
	 */

	arglistObj = Tcl_NewListObj(0, (Tcl_Obj **) NULL);
	INCR_REF_COUNT(arglistObj);
	for (localPtr = procPtr->firstLocalPtr;  localPtr != NULL;
	     localPtr = localPtr->nextPtr) {
	  if (TclIsVarArgument(localPtr)) {
	    Tcl_Obj *defVal, *defStringObj = Tcl_NewStringObj(localPtr->name, -1);
	    INCR_REF_COUNT(defStringObj);
	    /* check for default values */
	    if ((GetProcDefault(in, &nsPtr->cmdTable, name,
				localPtr->name, &defVal) == TCL_OK) &&
		(defVal != 0)) {
	      Tcl_AppendStringsToObj(defStringObj, " ",
				     ObjStr(defVal), 0);
	    }
	    Tcl_ListObjAppendElement(in, arglistObj, defStringObj);
	    DECR_REF_COUNT(defStringObj);
	  }
	}
	
	if (cmdPtr->objProc == RUNTIME_STATE(in)->objInterpProc) {
	  Tcl_DString ds, *dsPtr = &ds;
	  char *clObjName = NSCutXOTclClasses(nsPtr->fullName);

	  if (clObjName != nsPtr->fullName) {
	    /* it started with ::xotcl::classes */
	    XOTclClass* cl = GetClass(in, clObjName);
	    XOTclProcAssertion* procs;

	    if (cl) {
	      procs = cl->opt ?
		AssertionFindProcs(cl->opt->assertions, name) : 0;
	    } else {
	      DECR_REF_COUNT(newFullCmdName);
	      DECR_REF_COUNT(oldFullCmdName);
	      DECR_REF_COUNT(arglistObj);
	      return XOTclVarErrMsg(in, "No class for inst - assertions", 0);
	    }

	    /* XOTcl InstProc */
	    DSTRING_INIT(dsPtr);
	    Tcl_DStringAppendElement(dsPtr, NSCutXOTclClasses(newNsPtr->fullName));
	    Tcl_DStringAppendElement(dsPtr, "instproc");
	    Tcl_DStringAppendElement(dsPtr, name);
	    Tcl_DStringAppendElement(dsPtr, ObjStr(arglistObj));
	    Tcl_DStringAppendElement(dsPtr, ObjStr(procPtr->bodyPtr));
	    if (procs) {
	      requireClassOpt(cl);
	      AssertionAppendPrePost(in, dsPtr, procs);
	    }
	    Tcl_Eval(in, Tcl_DStringValue(dsPtr));
	    DSTRING_FREE(dsPtr);
	  } else {
	    XOTclObject *obj = GetObject(in, nsPtr->fullName);
	    XOTclProcAssertion* procs;
	    if (obj) {
	      procs = obj->opt ?
		AssertionFindProcs(obj->opt->assertions, name) : 0;
	    } else {
	      DECR_REF_COUNT(newFullCmdName);
	      DECR_REF_COUNT(oldFullCmdName);
	      DECR_REF_COUNT(arglistObj);
	      return XOTclVarErrMsg(in, "No object for assertions", 0);
	    }

	    /* XOTcl Proc */
	    DSTRING_INIT(dsPtr);
	    Tcl_DStringAppendElement(dsPtr, newNsPtr->fullName);
	    Tcl_DStringAppendElement(dsPtr, "proc");
	    Tcl_DStringAppendElement(dsPtr, name);
	    Tcl_DStringAppendElement(dsPtr, ObjStr(arglistObj));
	    Tcl_DStringAppendElement(dsPtr, ObjStr(procPtr->bodyPtr));
	    if (procs) {
	      requireObjectOpt(obj);
	      AssertionAppendPrePost(in, dsPtr, procs);
	    }
	    Tcl_Eval(in, Tcl_DStringValue(dsPtr));
	    DSTRING_FREE(dsPtr);
	  }
	  DECR_REF_COUNT(arglistObj);
	} else {
	  /* Tcl Proc */
	  Tcl_VarEval(in, "proc ", newName, " {", ObjStr(arglistObj),"} {\n",
		      ObjStr(procPtr->bodyPtr), "}", 0);
	}
      } else {
	/*
	 * Otherwise copy command
	 */
	if (cmdPtr->objProc) {
	  Tcl_CreateObjCommand(in, newName, cmdPtr->objProc,
			       cmdPtr->objClientData, cmdPtr->deleteProc);
	} else {
	  Tcl_CreateCommand(in, newName, cmdPtr->proc,
			    cmdPtr->clientData, cmdPtr->deleteProc);
	}
      }
    }
    hPtr = Tcl_NextHashEntry(&hSrch);
    DECR_REF_COUNT(newFullCmdName); DECR_REF_COUNT(oldFullCmdName);
  }
  return TCL_OK;
}

static int
XOTcl_NSCopyVars(ClientData cd, Tcl_Interp *in, int objc, Tcl_Obj *CONST objv[]) {
  Namespace *nsPtr, *newNsPtr;
  Var *varPtr=0;
  Tcl_DString ds, *dsPtr = &ds;
  Tcl_HashSearch hSrch;
  Tcl_HashEntry* hPtr;
  Tcl_HashTable* tablePtr;
  int rc = TCL_OK;
  char* varName;
  XOTclObject* obj;
  char* destFullName;

  if (objc != 3)
    return XOTclObjErrArgCnt(in, NULL, "namespace_copyvars fromNs toNs");

  nsPtr = (Namespace*) Tcl_FindNamespace(in, ObjStr(objv[1]),
					 (Tcl_Namespace *) NULL, 0);
  if (nsPtr) {
    newNsPtr = (Namespace*) Tcl_FindNamespace(in, ObjStr(objv[2]),
					      (Tcl_Namespace *) NULL, 0);
    if (!newNsPtr)
      return XOTclVarErrMsg(in, "CopyVars: Destination namespace ",
			    ObjStr(objv[2]), " does not exist",0);
    obj = GetObject(in, nsPtr->fullName);
    tablePtr = &nsPtr->varTable;
    destFullName = newNsPtr->fullName;
  } else {
    XOTclObject* newObj;
    obj = GetObject(in, ObjStr(objv[1]));
    if (!obj)
      return XOTclVarErrMsg(in, "CopyVars: Origin object/namespace ",
			    ObjStr(objv[1]), " does not exist",0);
    newObj = GetObject(in, ObjStr(objv[2]));
    if (!newObj)
      return XOTclVarErrMsg(in, "CopyVars: Destination object/namespace ",
			    ObjStr(objv[2]), " does not exist",0);
    tablePtr = obj->varTable;
    destFullName = ObjStr(newObj->cmdName);
  }

  /* copy all vars in the ns */
  hPtr = tablePtr ? Tcl_FirstHashEntry(tablePtr, &hSrch) : 0;
  while (hPtr != NULL) {
    varPtr = (Var *) Tcl_GetHashValue(hPtr);
    if (!TclIsVarUndefined(varPtr) && !TclIsVarLink(varPtr)) {
      varName = Tcl_GetHashKey(tablePtr, hPtr);

      if (TclIsVarScalar(varPtr)) {
	/* it may seem odd that we do not copy obj vars with the
	 * same SetVar2 as normal vars, but we want to dispatch it in order to
	 * be able to intercept the copying */
	if (obj) {
	  DSTRING_INIT(dsPtr);
	  Tcl_DStringAppendElement(dsPtr, destFullName);
	  Tcl_DStringAppendElement(dsPtr, "set");
	  Tcl_DStringAppendElement(dsPtr, varName);
	  Tcl_DStringAppendElement(dsPtr, ObjStr(varPtr->value.objPtr));
	  rc = Tcl_Eval(in, Tcl_DStringValue(dsPtr));
	  DSTRING_FREE(dsPtr);
	} else {
	  ALLOC_NAME_NS(&ds, destFullName, varName);
	  Tcl_SetVar2(in, Tcl_DStringValue(&ds), 0,
		      ObjStr(varPtr->value.objPtr), TCL_GLOBAL_ONLY);
	  DSTRING_FREE(&ds);
	}
      } else {
	if (TclIsVarArray(varPtr)) {
	  Tcl_HashTable* aTable = varPtr->value.tablePtr;
	  Tcl_HashSearch ahSrch;
	  Tcl_HashEntry* ahPtr = aTable ? Tcl_FirstHashEntry(aTable, &ahSrch) : 0;
	
	  for (; ahPtr != 0; ahPtr = Tcl_NextHashEntry(&ahSrch)) {
	    char *eltName = Tcl_GetHashKey(aTable, ahPtr);
	    Var  *eltVar  = (Var*) Tcl_GetHashValue(ahPtr);
	
	    if (TclIsVarScalar(eltVar)) {
	      if (obj) {
		Tcl_DString ds2, *ds2Ptr = &ds2;
		DSTRING_INIT(dsPtr);
		Tcl_DStringAppendElement(dsPtr, destFullName);
		Tcl_DStringAppendElement(dsPtr, "set");
		DSTRING_INIT(ds2Ptr);
		Tcl_DStringAppend(ds2Ptr, varName, -1);
		Tcl_DStringAppend(ds2Ptr, "(", 1);
		Tcl_DStringAppend(ds2Ptr, eltName, -1);
		Tcl_DStringAppend(ds2Ptr, ")", 1);
		Tcl_DStringAppendElement(dsPtr,Tcl_DStringValue(ds2Ptr));
		Tcl_DStringAppendElement(dsPtr,
					 ObjStr(eltVar->value.objPtr));
		rc = Tcl_Eval(in, Tcl_DStringValue(dsPtr));
		DSTRING_FREE(dsPtr);
		DSTRING_FREE(ds2Ptr);
	      } else {
		ALLOC_NAME_NS(&ds, destFullName, varName);
		Tcl_SetVar2(in, Tcl_DStringValue(&ds), eltName,
			    ObjStr(eltVar->value.objPtr), TCL_GLOBAL_ONLY);
		DSTRING_FREE(&ds);
	      }
	    }
	  }
	}
      }
    }
    hPtr = Tcl_NextHashEntry(&hSrch);
  }
  return rc;
}

static int
XOTcl_InfoObjCmd(ClientData cd, Tcl_Interp *in, int objc, Tcl_Obj *CONST objv[]) {
  int result, filterFrameCutOff = 0, isBody = 0;
  CallFrame *savedFramePtr = 0, *savedVarFramePtr = 0;
  if (objc > 1) {
    char* opt = ObjStr(objv[1]);
    if (isLevelString(opt) &&
	FilterEntryOnCallStack(in)) {
      filterFrameCutOff =
	FilterCutFrames(in, &savedFramePtr, &savedVarFramePtr);
    } else if (isBodyString(opt) && objc > 2)
      isBody = 1;
  }
  result = callCommand(in, INFO, objc, objv);
  if (filterFrameCutOff)
    FilterRestoreFrames(in, savedFramePtr, savedVarFramePtr);

  if (isBody && result == TCL_OK) {
    char* body = ObjStr(Tcl_GetObjResult(in));
    if (strncmp(body, "::xotcl::initProcNS\n",20) == 0)
      body+=20;
    Tcl_SetObjResult(in, Tcl_NewStringObj(body, -1));
  }
  return result;

}

static int
XOTcl_UplevelObjCmd(ClientData cd, Tcl_Interp *in, int objc, Tcl_Obj *CONST objv[]) {
  int result, filterFrameCutOff = 0;
  CallFrame *savedFramePtr = 0, *savedVarFramePtr = 0;
  XOTclCallStack *cs = &RUNTIME_STATE(in)->cs;
  /* invalidate the top frame cache */
  cs->topFrameCache = 0;

  if (FilterEntryOnCallStack(in)) {
    filterFrameCutOff =
      FilterCutFrames(in, &savedFramePtr, &savedVarFramePtr);
  }
  result = callCommand(in, UPLEVEL, objc, objv);
  if (filterFrameCutOff)
    FilterRestoreFrames(in, savedFramePtr, savedVarFramePtr);
  /* invalidate the top frame cache */
  cs->topFrameCache = 0;
  return result;
}
static int
XOTcl_UpvarObjCmd(ClientData cd, Tcl_Interp *in, int objc, Tcl_Obj *CONST objv[]) {
  int result, filterFrameCutOff = 0;
  CallFrame *savedFramePtr = 0, *savedVarFramePtr = 0;
  XOTclCallStack *cs = &RUNTIME_STATE(in)->cs;
  /* invalidate the top frame cache */
  cs->topFrameCache = 0;

  if (FilterEntryOnCallStack(in)) {
    filterFrameCutOff =
      FilterCutFrames(in, &savedFramePtr, &savedVarFramePtr);
  }
  result = callCommand(in, UPVAR, objc, objv);
  if (filterFrameCutOff)
    FilterRestoreFrames(in, savedFramePtr, savedVarFramePtr);
  /* invalidate the top frame cache */
  cs->topFrameCache = 0;
  return result;
}


int
XOTclSelfDispatchCmd(ClientData cd, Tcl_Interp *in, int objc, Tcl_Obj *CONST objv[]) {
  XOTclObject *self;
  int result;
  if ((self = GetSelfObj(in))) {
    result = callMethod((ClientData)self, in, objv[1], objc, objv+2, 0);
  } else {
    result = TCL_ERROR;
    panic("Self Dispatch: could not resolve self");
  }
  return result;
}

int
XOTclInitProcNSCmd(ClientData cd, Tcl_Interp *in, int objc, Tcl_Obj *CONST objv[]) {
  Interp *iPtr = ((Interp *) in);

  if (iPtr->varFramePtr) {
    if (iPtr->varFramePtr->callerVarPtr) {
      iPtr->varFramePtr->nsPtr = iPtr->varFramePtr->callerVarPtr->nsPtr;
    } else {
      iPtr->varFramePtr->nsPtr = iPtr->globalNsPtr;
    }
  }
  return TCL_OK;
}



static int
XOTcl_RenameObjCmd(ClientData cd, Tcl_Interp *in, int objc, Tcl_Obj *CONST objv[]) {
  /* this call the Tcl_RenameObjCmd, but it ensures before that
     the renamed obj, functions, etc. are not part of XOTcl */
  XOTclObject* obj;
  XOTclClass* cl;
  Tcl_Command cmd;
  Command *cmdPtr;
  int result;
  char *oldName, *newName;
  char* oldSimpleCmdName=0, *newSimpleCmdName;
  Tcl_DString oldDS;

  /* wrong # args => normal Tcl ErrMsg*/
  if (objc != 3)
    return callCommand(in, RENAME, objc, objv);

  oldName = ObjStr(objv[1]);
  newName = ObjStr(objv[2]);

  /* if an obj/cl should be rename => call the XOTcl-rename command */
  cmd = Tcl_FindCommand(in, oldName, (Tcl_Namespace *) NULL,0);
  cmdPtr = (Command *) cmd;

  obj = GetObjectFromCmdPtr(cmdPtr);
  cl  = obj ? XOTclObjectToClass(obj) : NULL;

  DSTRING_INIT(&oldDS);
  if (cmdPtr != NULL) {
    if (obj) {
      DSTRING_FREE(&oldDS);
      return callMethodWithArg((ClientData)obj, in,
			       global_objects[MOVE], objv[2], 3, 0, 0);
    }

    /* Save the SimpleName */
    Tcl_DStringAppend(&oldDS, Tcl_GetCommandName(in, cmd), -1);
    oldSimpleCmdName = Tcl_DStringValue(&oldDS);

    /* saving obj/cl for proc's; they could be destroyed */

    if (cmdPtr->objProc == XOTclObjDispatch) {
      obj = (XOTclObject*)cmdPtr->objClientData;
      cl = XOTclObjectToClass(obj);
    } else {
      obj = 0;
      cl = 0;
    }
  }

  /* Actually rename the cmd */
  result = callCommand(in, RENAME, objc, objv);

  /* has a XOTcl Proc/Instproc been renamed */
  if (result == TCL_OK && cmdPtr != NULL && obj) {
    if ((newName != NULL) && (*newName != '\0')) {
      /* it was renamed -> correct the assertion lists*/
      Tcl_Obj* pre; Tcl_Obj* post;
      XOTclAssertionStore *as;
      XOTclProcAssertion *pa;
      if (obj)
	as = obj->opt ? obj->opt->assertions : 0;
      else
	as = cl->opt ? cl->opt->assertions : 0;
      pa = AssertionFindProcs(as, oldSimpleCmdName);
      if (pa) {
	newSimpleCmdName = (char*) Tcl_GetCommandName(in, cmd);	
	pre = AssertionList(in, pa->pre);
	post = AssertionList(in, pa->post);
	AssertionRemoveProc(as, oldSimpleCmdName);
	AssertionAddProc(in, newSimpleCmdName, as, pre, post);
      }
    }
  }
  DSTRING_FREE(&oldDS);
  return result;
}

/* create a slave interp that calls XOTcl Init */
static int
XOTcl_InterpObjCmd(ClientData cd, Tcl_Interp *in, int objc, Tcl_Obj *objv[]) {
  Tcl_Interp* slave;
  Tcl_Obj* saved;

  if (objc < 1)
      return XOTclObjErrArgCnt(in, NULL, "xotcl_interp name ?args?");

  saved = objv[0];
  objv[0] = global_objects[INTERP];
  if (Tcl_EvalObjv(in, objc, objv, 
		   TCL_GLOBAL_ONLY|TCL_LEAVE_ERR_MSG) != TCL_OK) {
    objv[0] = saved;
    return TCL_ERROR;
  }
  objv[0] = saved;

  if (isCreateString(ObjStr(objv[1]))) {
    slave = Tcl_GetSlave(in, ObjStr(objv[2]));
    if (!slave)
      return XOTclVarErrMsg(in, "Creation of slave interpreter failed",
			    (char*)NULL);
    if (Xotcl_Init(slave) == TCL_ERROR) {
      return TCL_ERROR;
    }
  }

  return TCL_OK;
}

/*
 * C interface routines for manipulating objects and classes
 */

extern XOTclObject*
XOTclGetObject(Tcl_Interp* in, char* name) {
  return GetObject(in, name);
}

/*
 * Find an object using a char* name
 */
static XOTclObject*
GetObject(Tcl_Interp* in, char* name) {
  register Command *cmdPtr;
  assert(name);
  cmdPtr = (Command*) NSFindCommand(in, name);
  if (cmdPtr && cmdPtr->objProc == XOTclObjDispatch) {
    return (XOTclObject*)cmdPtr->objClientData;
  }
  return 0;
}

extern XOTclClass*
XOTclGetClass(Tcl_Interp* in, char* name) {
  return GetClass(in, name);
}

/*
 * Find a class using a char* name
 */
static XOTclClass*
GetClass(Tcl_Interp* in, char* name) {
  XOTclObject *obj = GetObject(in, name);
  return (obj && XOTclObjectIsClass(obj)) ? (XOTclClass *)obj : NULL;
}

void
XOTclAddPMethod(Tcl_Interp* in, XOTclObject* obj, char* nm, Tcl_ObjCmdProc* proc,
               ClientData cd, Tcl_CmdDeleteProc* dp) {
  Tcl_DString newCmd, *cptr=&newCmd;
  requireObjNamespace(in, obj);
  ALLOC_NAME_NS(cptr, obj->nsPtr->fullName, nm);
  Tcl_CreateObjCommand(in, Tcl_DStringValue(cptr), proc, cd, dp);
  DSTRING_FREE(cptr);
}

void
XOTclAddIMethod(Tcl_Interp* in, XOTclClass *cl, char* nm,
		Tcl_ObjCmdProc* proc, ClientData cd, Tcl_CmdDeleteProc* dp) {
  Tcl_DString newCmd, *cptr=&newCmd;
  ALLOC_NAME_NS(cptr, cl->nsPtr->fullName, nm);
  Tcl_CreateObjCommand(in, Tcl_DStringValue(cptr), proc, cd, dp);
  DSTRING_FREE(cptr);
}


extern Tcl_Obj*
XOTclOGetInstVar2(XOTclObject* obj, Tcl_Interp* in,
		  Tcl_Obj* name1, Tcl_Obj *name2,
		  int flgs) {

  Tcl_Obj* result;
  VarFrameDecls;

  VarFrameSwitchToObj(in, obj);
  if (obj->nsPtr)
    flgs |= TCL_NAMESPACE_ONLY;

#if defined(V81)
  result = Tcl_GetObjVar2(in, ObjStr(name1),
			  ((name2==NULL) ? (char*)NULL :
			   ObjStr(name2)), flgs);
#else
  result = Tcl_ObjGetVar2(in, name1, name2, flgs);
#endif

  VarFrameRestore(in, obj);

  return result;
}

/*
 * get all instances of a class recursively to an initialized
 * String key hashtable
 */
static void
getAllInstances(Tcl_HashTable* destTable, XOTclClass* startCl) {
  Tcl_HashTable* table = &startCl->instances;
  Tcl_HashSearch search;
  Tcl_HashEntry *hPtr;

  for (hPtr = Tcl_FirstHashEntry(table, &search);  hPtr != NULL;
       hPtr = Tcl_NextHashEntry(&search)) {
    XOTclObject* inst = (XOTclObject*) Tcl_GetHashKey(table, hPtr);
    Tcl_HashEntry *hPtrDest;
    int new;
    hPtrDest = Tcl_CreateHashEntry(destTable, ObjStr(inst->cmdName),
				   &new);
    /*
    fprintf (stderr, " -- %s (%s)\n", ObjStr(inst->cmdName), ObjStr(startCl->object.cmdName));
    */
    if (new && XOTclObjectIsClass(inst)) {
      getAllInstances(destTable, (XOTclClass*) inst);
    }
  }
}
#if !defined(NDEBUG)
static void
checkAllInstances(Tcl_Interp *in, XOTclClass* cl, int lvl) {
  Tcl_HashSearch search;
  Tcl_HashEntry *hPtr;
  if (cl && cl->object.refCount>0) {
    /*fprintf(stderr,"checkallinstances %d cl=%p '%s'\n", lvl, cl, ObjStr(cl->object.cmdName));*/
    for (hPtr = Tcl_FirstHashEntry(&cl->instances, &search);  hPtr != NULL;
	 hPtr = Tcl_NextHashEntry(&search)) {
      XOTclObject* inst = (XOTclObject*) Tcl_GetHashKey(&cl->instances, hPtr);
      assert(inst);
      assert(inst->refCount>0);
      assert(inst->cmdName->refCount>0);
      if (XOTclObjectIsClass(inst) && (XOTclClass*)inst != RUNTIME_STATE(in)->theClass) {
	checkAllInstances(in, (XOTclClass*) inst, lvl+1);
      }
    }
  }
}
#endif

#ifdef DO_FULL_CLEANUP
/* delete global variables and procs */
static void deleteProcsAndVars(Tcl_Interp *in) {
  Namespace *nsPtr = (Namespace *) Tcl_GetGlobalNamespace(in);
  Tcl_HashSearch search;
  Var *varPtr;
  Command *cmd;
  register Tcl_HashEntry *entryPtr;
  char *varName;

  entryPtr = Tcl_FirstHashEntry(&nsPtr->varTable, &search);
  while (entryPtr != NULL) {
    varPtr = (Var *) Tcl_GetHashValue(entryPtr);
    if (!TclIsVarUndefined(varPtr) || (varPtr->flags & VAR_NAMESPACE_VAR)) {
      varName = Tcl_GetHashKey(&nsPtr->varTable, entryPtr);
      /* fprintf(stderr, "unsetting var %s\n", varName);*/
      Tcl_UnsetVar2(in, varName, (char *) NULL, TCL_GLOBAL_ONLY);
    }
    entryPtr = Tcl_NextHashEntry(&search);
  }
  
  entryPtr = Tcl_FirstHashEntry(&nsPtr->cmdTable, &search);
  while (entryPtr) {
    cmd = (Command *)Tcl_GetHashValue(entryPtr);
    if (cmd->proc == RUNTIME_STATE(in)->interpProc) {
      /*
	char* key = Tcl_GetHashKey(&nsPtr->cmdTable, entryPtr);
	fprintf(stderr, "cmdname = %s cmd %p proc %p objProc %p %d\n",
		key,cmd,cmd->proc,cmd->objProc,
		cmd->proc==RUNTIME_STATE(in)->interpProc);
      
	*/
      Tcl_DeleteCommandFromToken(in, (Tcl_Command)cmd);
    }
    entryPtr = Tcl_NextHashEntry(&search);
  }
}
#endif


#ifdef DO_CLEANUP
static int
ClassHasSubclasses(XOTclClass* cl) {
  return (cl->sub != NULL);
}

static int
ClassHasInstances(XOTclClass* cl) {
  Tcl_HashSearch hSrch;
  return (Tcl_FirstHashEntry(&cl->instances, &hSrch) != NULL);
}

static int
ObjectHasChildren(Tcl_Interp* in, XOTclObject* obj) {
  Namespace *nsp = (Namespace *)obj->nsPtr;
  int result = 0;

  if (nsp != NULL) {
    Tcl_HashEntry *hPtr;
    Tcl_HashSearch hSrch;
    VarFrameDecls;
    
    VarFrameSwitchToObj(in, obj);
    for (hPtr = Tcl_FirstHashEntry(&nsp->cmdTable, &hSrch); hPtr; hPtr = Tcl_NextHashEntry(&hSrch)) {
      char* key = Tcl_GetHashKey(&nsp->cmdTable, hPtr);
      if (GetObject(in, key)) {
	/*fprintf(stderr,"child = %s\n",key);*/
	result = 1;
	break;
      }
    }
    VarFrameRestore(in,obj);
  }
  return result;
}

static void freeAllXOTclObjectsAndClasses(Tcl_Interp *in, Tcl_HashTable *commandTable) {
  Tcl_HashEntry *hPtr, *hDel;  
  Tcl_HashSearch hSrch;
  XOTclObject *obj;
  XOTclClass  *thecls, *theobj, *cl;
  int i;
  
  /*fprintf(stderr,"freeAllXOTclObjectsAndClasses\n");*/

  thecls = RUNTIME_STATE(in)->theClass;
  theobj = RUNTIME_STATE(in)->theObject;
  /***** PHYSICAL DESTROY *****/
  RUNTIME_STATE(in)->exitHandlerDestroyRound = XOTCL_EXITHANDLER_ON_PHYSICAL_DESTROY;
  while (1) {
    int deleted = 0;
    hPtr = Tcl_FirstHashEntry(commandTable, &hSrch);
    while (hPtr) {
      char* key = Tcl_GetHashKey(commandTable, hPtr);
      obj = GetObject(in, key);
      if (obj && !XOTclObjectIsClass(obj) && !ObjectHasChildren(in,obj)) {
	/*fprintf(stderr,"  ... delete object %s %p\n",key,obj);*/
	Tcl_DeleteCommandFromToken(in, obj->id);
	hDel = hPtr;
	deleted++;
      } else {
	hDel = NULL;
      }
      hPtr = Tcl_NextHashEntry(&hSrch);
      if (hDel) 
	Tcl_DeleteHashEntry(hDel);
    }
    /*fprintf(stderr, "deleted %d Objects\n",deleted);*/
    if (deleted>0) 
      continue;
    
    
    hPtr = Tcl_FirstHashEntry(commandTable, &hSrch);
    while (hPtr) {
      char* key = Tcl_GetHashKey(commandTable, hPtr);
      cl = GetClass(in, key);
      /*fprintf(stderr,"cl key = %s %p\n", key, cl);*/
      if (cl 
	  && !ObjectHasChildren(in, (XOTclObject*)cl)
	  && !ClassHasInstances(cl)
	  && !ClassHasSubclasses(cl)
	  && cl != RUNTIME_STATE(in)->theClass 
	  && cl != RUNTIME_STATE(in)->theObject
	  ) {
	/*fprintf(stderr,"  ... delete class %s %p\n",key,cl);*/
	Tcl_DeleteCommandFromToken(in, cl->object.id);
	hDel = hPtr;
	deleted++;
      } else {
	hDel = NULL;
      }
      hPtr = Tcl_NextHashEntry(&hSrch);
      if (hDel) 
	Tcl_DeleteHashEntry(hDel);
    }
    /*fprintf(stderr, "deleted %d Classes\n",deleted);*/
    if (deleted>0) 
      continue;
    
    break;
    
  }
  
#ifdef DO_FULL_CLEANUP
  deleteProcsAndVars(in);
#endif
  
  RUNTIME_STATE(in)->callDestroy = 0;
  RemoveSuper(thecls, theobj);
  RemoveInstance((XOTclObject*)thecls, thecls);
  RemoveInstance((XOTclObject*)theobj, thecls);
  
  Tcl_DeleteCommandFromToken(in, theobj->object.id);
  RUNTIME_STATE(in)->theObject = NULL;

  Tcl_DeleteCommandFromToken(in, thecls->object.id);
  RUNTIME_STATE(in)->theClass = NULL;

  XOTcl_DeleteNamespace(in, (Tcl_Namespace*) RUNTIME_STATE(in)->fakeNS);
  XOTcl_DeleteNamespace(in, (Tcl_Namespace*) RUNTIME_STATE(in)->XOTclClassesNS);
  XOTcl_DeleteNamespace(in, (Tcl_Namespace*) RUNTIME_STATE(in)->XOTclNS);
  
  for (i = 0; i < nr_elements(global_strings); i++) {
    DECR_REF_COUNT(global_objects[i]);
  }
  MEM_COUNT_FREE("Tcl_Obj**",global_objects);
  ckfree((char*) global_objects);
  FREE(XOTclRuntimeState, RUNTIME_STATE(in));
} 
#endif /* DO_CLEANUP */


/*
 *  Exit Handler
 */
static void
ExitHandler(ClientData cd) {
  Tcl_Interp* in = (Tcl_Interp*) cd;
  Interp* iPtr = (Interp*) in;
  XOTclObject *obj;
  XOTclClass  *thecls, *theobj, *cl;
  int flags;
  Tcl_HashSearch hSrch;
  Tcl_HashEntry *hPtr;
  Tcl_HashTable objTable, *commandTable = &objTable;
  XOTclCallStack *cs= &RUNTIME_STATE(in)->cs;
  int result;

#if defined(PROFILE)
  XOTclProfilePrintData(in);
#endif
  /*
   * Don't use exit handler, if the interpreted is destroyed already
   * Call to exit handler comes after freeing namespaces, commands, etc.
   * e.g. TK calls Tcl_DeleteInterp directly, if Window is killed
   */

  /*
   * Ahem ...
   *
   * Since we *must* be sure that our destroy methods will run
   * we must *cheat* (I mean CHEAT) here: we flip the interp
   * flag, saying, "hey boy, you're not deleted any more".
   * After our handlers are done, we restore the old state...
   * All this is needed so we can do an eval in the interp which
   * is potentially marked for delete when we start working here.
   *
   * I know, I know, this is not realy elegant. But...  I'd need a
   * standard way of invoking some code at interpreter delete time
   * but JUST BEFORE the actual deletion process starts. Sadly,
   * there is no such hook in Tcl as of Tcl8.3.2, that I know of.
   *
   * So, for the rest of procedure, assume the interp is alive !
   */

  /*fprintf(stderr,"+++ EXIT handler\n");*/

  flags = iPtr->flags;
  iPtr->flags &= ~DELETED;

  TclCommands(in, 0);

  thecls = RUNTIME_STATE(in)->theClass;
  theobj = RUNTIME_STATE(in)->theObject;

  /* 
   * evaluate user-defined exit handler
   */
  result = callMethod((ClientData) theobj, in,
		      global_objects[EXIT_HANDLER], 2, 0, 0);
  if (result != TCL_OK) {
    fprintf(stderr,"User defined exit handler contains errors!\n"
            "Error in line %d: %s\nExecution interrupted.\n",
            in->errorLine, ObjStr(Tcl_GetObjResult(in)));
  }

  /*
   * Pop any callstack entry that is still alive (e.g.
   * if "exit" is called and we were jumping out of the
   * callframe
   */
   while (cs->top > cs->content)
     CallStackPop(in);
   
   while (iPtr->framePtr)
     Tcl_PopCallFrame(in);

  /*
   * deleting in two rounds:
   *  (a) SOFT DESTROY: call all user-defined destroys
   *  (b) PHYSICAL DESTROY: delete the commands, user-defined
   *      destroys are not executed anymore
   *
   * this is to prevent user-defined destroys from overriding physical
   * destroy during exit handler, but still ensure that all
   * user-defined destroys are called.
   */

   Tcl_InitHashTable(commandTable, TCL_STRING_KEYS);
   MEM_COUNT_ALLOC("Tcl_InitHashTable",commandTable);
   getAllInstances(commandTable, thecls);

   /***** SOFT DESTROY *****/
   RUNTIME_STATE(in)->exitHandlerDestroyRound = XOTCL_EXITHANDLER_ON_SOFT_DESTROY;

   hPtr = Tcl_FirstHashEntry(commandTable, &hSrch);
   while (hPtr) {
     char* key = Tcl_GetHashKey(commandTable, hPtr);
     obj = GetObject(in, key);
     /*fprintf(stderr,"key = %s %p %d\n", 
       key, obj, obj && !XOTclObjectIsClass(obj));*/
     if (obj && !XOTclObjectIsClass(obj)
	 && !(obj->flags & XOTCL_DESTROY_CALLED))
       callDestroyMethod((ClientData)obj, in, obj, 0);
     hPtr = Tcl_NextHashEntry(&hSrch);
   }
   hPtr = Tcl_FirstHashEntry(commandTable, &hSrch);
   while (hPtr) {
     char* key = Tcl_GetHashKey(commandTable, hPtr);
     cl = GetClass(in, key);
     if (cl
	 && !(cl->object.flags & XOTCL_DESTROY_CALLED))
       callDestroyMethod((ClientData)cl, in, (XOTclObject*)cl, 0);
     hPtr = Tcl_NextHashEntry(&hSrch);
   }
#ifdef DO_CLEANUP
   freeAllXOTclObjectsAndClasses(in, commandTable);
#endif
  MEM_COUNT_FREE("Tcl_InitHashTable",commandTable);
  Tcl_DeleteHashTable(commandTable);

#if defined(TCL_MEM_DEBUG)
  TclDumpMemoryInfo (stderr);
  Tcl_DumpActiveMemory ("./xotclActiveMem");
  /* Tcl_GlobalEval(in, "puts {checkmem to checkmemFile}; 
    checkmem checkmemFile"); */
#endif
  MEM_COUNT_DUMP();

  iPtr->flags = flags;
  Tcl_Release((ClientData) in);
}



#ifdef TCL_THREADS
/*
 * Gets activated at the thread-exit
 */
static void
XOTcl_ThreadExitProc(ClientData cd) {
  /*fprintf(stderr,"+++ ThreadExit\n");*/
  ExitHandler(cd);
}
#endif

/*
 * Gets activated at the application-exit
 */
static void
XOTcl_ExitProc(ClientData cd) {
  /* if main exit handler advances to this point -> don't call thread
     exit handler again since we'll panic in Tcl_Release ! */
  /*fprintf(stderr,"+++ ExitProc\n");*/

#ifdef TCL_THREADS
# if !defined(PRE83)
  Tcl_DeleteThreadExitHandler(XOTcl_ThreadExitProc, cd);
# endif
  ExitHandler(cd);
#else
  /* In threaded applications, the final exit handler is a no-no.
     The main thread already terminated and invoked the
     XOTcl_ThreadExitProc so there is no need to do it again...*/
  ExitHandler(cd);
#endif
}

/*
 * Registers thread/appl exit handlers
 * to keep the table clean when we exit.
 */
static void
RegisterExitHandlers(ClientData cd) {
#if !defined(PRE83)
  static int initialized = 0;
  static XOTclMutex initMutex = 0;
  Tcl_Preserve(cd);
  if (initialized == 0) {
    XOTclMutexLock(&initMutex);
    if (initialized == 0) {
      Tcl_CreateExitHandler(XOTcl_ExitProc, cd);
      initialized = 1;
    }
    XOTclMutexUnlock(&initMutex);
  }
# if defined(TCL_THREADS)
  Tcl_CreateThreadExitHandler(XOTcl_ThreadExitProc, cd);
# endif
#else
  Tcl_Preserve(cd);
  Tcl_CreateExitHandler(XOTcl_ExitProc, cd);
#endif /*!defined(PRE83)*/
}

/*
 * Tcl extension initialization routine
 */
static int
XOTclReplaceCommandCleanup(Tcl_Interp* in, global_names name) {
  Command *cmdPtr;
  int result = TCL_OK;

  cmdPtr = (Command *)Tcl_GetCommandFromObj(in, global_objects[name]);
  if (cmdPtr != NULL) {
    cmdPtr->objProc = tcl_commands[name-EXPR];
  } else {
    result = TCL_ERROR;
  }
  return result;
}

static int
XOTclReplaceCommand(Tcl_Interp* in, global_names name,
		    Tcl_ObjCmdProc* proc,
		    Tcl_ObjCmdProc* tclProc, int pass) {
  Command *cmdPtr;
  int result = TCL_OK;

  cmdPtr = (Command *)Tcl_GetCommandFromObj(in, global_objects[name]);
  if (cmdPtr != NULL) {
    if (proc != cmdPtr->objProc) {
      if (pass == 0) { /* Only once (must be locked here) */
	if (tclProc)
	  tcl_commands[name-EXPR] = tclProc;
	else
	  tcl_commands[name-EXPR] = cmdPtr->objProc;
      }
      if (proc)
	cmdPtr->objProc = proc;
      /*Tcl_CreateObjCommand(in, global_strings[name], proc, 0, 0);*/
    } else {
      /* fprintf(stderr,"!!! fishy replace\n");*/
    }
  } else {
    result = TCL_ERROR;
  }

  return result;
}
/*
 * Ohtain the names of the tcl commands
 * not available through our the stub interface
 */
static int
TclCommands(Tcl_Interp* in, int load) {
  static int initialized = 0;
  int rc = TCL_OK;
  static XOTclMutex initMutex = 0;

  if (load) {
    if (initialized == 0) {
      XOTclMutexLock(&initMutex);
    }
#ifdef USE_TCL_STUBS
    rc|= XOTclReplaceCommand(in, EXPR,     0, 0, initialized);
    rc|= XOTclReplaceCommand(in, INCR,     0, 0, initialized);
#endif
    rc|= XOTclReplaceCommand(in, SUBST,    0, SUBST_CMD, initialized);
    rc|= XOTclReplaceCommand(in, INFO,     XOTcl_InfoObjCmd, 0, initialized);
    rc|= XOTclReplaceCommand(in, RENAME,   XOTcl_RenameObjCmd, 0, initialized);
    rc|= XOTclReplaceCommand(in, UPLEVEL,  XOTcl_UplevelObjCmd, 0, initialized);
    rc|= XOTclReplaceCommand(in, UPVAR,    XOTcl_UpvarObjCmd, 0, initialized);
    
    if (initialized == 0) {
      initialized = 1;   
      XOTclMutexUnlock(&initMutex);
    }
  } else {
    XOTclReplaceCommandCleanup(in, INFO);
    XOTclReplaceCommandCleanup(in, RENAME);
    XOTclReplaceCommandCleanup(in, UPLEVEL);
    XOTclReplaceCommandCleanup(in, UPVAR);
  }
  return rc;
}

extern int
Xotcl_Init(Tcl_Interp* in) {
  XOTclClass* theobj = 0;
  XOTclClass* thecls = 0;
  XOTclClass* paramCl = 0;
  ClientData runtimeState;

  int i;
#ifdef XOTCL_BYTECODE
  XOTclCompEnv *instructions = XOTclGetCompEnv();
#endif

#ifndef PRE81
# ifdef USE_TCL_STUBS
  if (Tcl_InitStubs(in, "8.1", 0) == NULL) {
    return TCL_ERROR;
  }
# endif
#endif

#if defined(TCL_MEM_DEBUG)
  TclDumpMemoryInfo(stderr);
#endif
  MEM_COUNT_INIT();

  /*
  fprintf(stderr, "SIZES: obj=%d, tcl_obj=%d, DString=%d, class=%d, namespace=%d,  command=%d,  HashTable=%d\n", sizeof(XOTclObject), sizeof(Tcl_Obj), sizeof(Tcl_DString), sizeof(XOTclClass), sizeof(Namespace),  sizeof(Command), sizeof(Tcl_HashTable));
  */

  /*
   * Runtime State stored in the client data of the Interp's global
   * Namespace in order to avoid global state information
   */
  runtimeState = (ClientData) NEW(XOTclRuntimeState);
  Tcl_SetAssocData(in, "XOTclRuntimeState", NULL, runtimeState);

  /* CallStack initialization */
  memset(RUNTIME_STATE(in),
         0, sizeof(XOTclRuntimeState));

  memset(RUNTIME_STATE(in)->cs.content,
	 0, sizeof(XOTclCallStackContent));

  RUNTIME_STATE(in)->cs.top = RUNTIME_STATE(in)->cs.content;
  /* RUNTIME_STATE(in)->cs.frameCutRound = 0; */
  RUNTIME_STATE(in)->callDestroy = 1;

  /* create xotcl namespace */
  RUNTIME_STATE(in)->XOTclNS = (Namespace*)
    Tcl_CreateNamespace(in, "::xotcl",
			(ClientData)NULL,(Tcl_NamespaceDeleteProc*)NULL);
  MEM_COUNT_ALLOC("TclNamespace",RUNTIME_STATE(in)->XOTclNS);

  /*
   * init an empty, faked proc structure in the RUNTIME state
   */
  RUNTIME_STATE(in)->fakeProc.iPtr = (Interp *)in;
  RUNTIME_STATE(in)->fakeProc.refCount = 1;
  RUNTIME_STATE(in)->fakeProc.cmdPtr = NULL;
  RUNTIME_STATE(in)->fakeProc.bodyPtr = NULL;
  RUNTIME_STATE(in)->fakeProc.numArgs  = 0;
  RUNTIME_STATE(in)->fakeProc.numCompiledLocals = 0;
  RUNTIME_STATE(in)->fakeProc.firstLocalPtr = NULL;
  RUNTIME_STATE(in)->fakeProc.lastLocalPtr = NULL;
  RUNTIME_STATE(in)->fakeNS = (Namespace*)
    Tcl_CreateNamespace(in, "::xotcl::fakeNS",
			(ClientData)NULL,(Tcl_NamespaceDeleteProc*)NULL);
  MEM_COUNT_ALLOC("TclNamespace",RUNTIME_STATE(in)->fakeNS);

  /* XOTclClasses in separate Namespace / Objects */
  RUNTIME_STATE(in)->XOTclClassesNS = (Namespace*)
    Tcl_CreateNamespace(in, "::xotcl::classes",
			(ClientData)NULL,(Tcl_NamespaceDeleteProc*)NULL);
  MEM_COUNT_ALLOC("TclNamespace",RUNTIME_STATE(in)->XOTclClassesNS);


  /* cache interpreters proc interpretation functions */
  RUNTIME_STATE(in)->objInterpProc = TclGetObjInterpProc();
  RUNTIME_STATE(in)->interpProc = TclGetInterpProc();
  RUNTIME_STATE(in)->exitHandlerDestroyRound = XOTCL_EXITHANDLER_OFF;

  RegisterObjTypes();
  RegisterExitHandlers((ClientData)in);

  /* initialize global Tcl_Obj*/
  global_objects = NEW_ARRAY(Tcl_Obj*,nr_elements(global_strings));

  for (i = 0; i < nr_elements(global_strings); i++) {
    global_objects[i] = Tcl_NewStringObj(global_strings[i],-1);
    INCR_REF_COUNT(global_objects[i]);
  }

  /* create Object and Class, and store them in the RUNTIME STATE */
  theobj = PrimitiveCCreate(in, "::xotcl::Object", 0);
  RUNTIME_STATE(in)->theObject = theobj;
  if (!theobj) panic("Cannot create ::xotcl::Object",0);

  thecls = PrimitiveCCreate(in, "::xotcl::Class", 0);
  RUNTIME_STATE(in)->theClass = thecls;
  if (!thecls) panic("Cannot create ::xotcl::Class",0);

  theobj->parent = 0;
  thecls->parent = theobj;

  Tcl_Export(in, (Tcl_Namespace*) RUNTIME_STATE(in)->XOTclNS,
	     "Object", 0);
  Tcl_Export(in, (Tcl_Namespace*) RUNTIME_STATE(in)->XOTclNS,
	     "Class", 0);
  /*Tcl_AddInterpResolvers(in, "XOTcl", XOTclResolveCmd, 0, 0);*/

#ifdef XOTCL_BYTECODE
  instructions[INST_SELF].cmdPtr = (Command *)
#endif
  Tcl_CreateObjCommand(in, "::xotcl::self", XOTclGetSelfObjCmd, 0, 0);
#ifdef XOTCL_BYTECODE
  instructions[INST_NEXT].cmdPtr = (Command *)
#endif
  Tcl_CreateObjCommand(in, "::xotcl::next", (Tcl_ObjCmdProc*) XOTclNextObjCmd, 0, 0);
  /*XOTclAddIMethod(in, theobj, "next", (Tcl_ObjCmdProc*) XOTclNextObjCmd, 0, 0);*/

  Tcl_Export(in, (Tcl_Namespace*) RUNTIME_STATE(in)->XOTclNS,
	     "self", 0);
  
  Tcl_Export(in, (Tcl_Namespace*) RUNTIME_STATE(in)->XOTclNS,
	     "next", 0);
  Tcl_Export(in, (Tcl_Namespace*) RUNTIME_STATE(in)->XOTclNS,
	     "my", 0);

#if defined(PROFILE)
  XOTclProfileInit(in);
#endif

  /* test Object and Class creation */
  if (!theobj || !thecls) {
    RUNTIME_STATE(in)->callDestroy = 0;

    if (thecls) PrimitiveCDestroy((ClientData) thecls);
    if (theobj) PrimitiveCDestroy((ClientData) theobj);

    for (i = 0; i < nr_elements(global_strings); i++) {
      DECR_REF_COUNT(global_objects[i]);
    }
    FREE(Tcl_Obj**, global_objects);
    FREE(XOTclRuntimeState, RUNTIME_STATE(in));

    return XOTclErrMsg(in, "Object/Class failed", TCL_STATIC);
  }

  AddInstance((XOTclObject*)theobj, thecls);
  AddInstance((XOTclObject*)thecls, thecls);
  AddSuper(thecls, theobj);

  /*
   * and fill them with functionality
   */
  XOTclAddIMethod(in, theobj, "autoname", XOTclOAutonameMethod, 0, 0);
  XOTclAddIMethod(in, theobj, "check", (Tcl_ObjCmdProc*)XOTclOCheckMethod, 0, 0);
  XOTclAddIMethod(in, theobj, "class", XOTclOClassMethod, 0, 0);
  XOTclAddIMethod(in, theobj, "cleanup", XOTclOCleanupMethod, 0, 0);
  XOTclAddIMethod(in, theobj, "configure", (Tcl_ObjCmdProc*) XOTclOConfigureMethod, 0, 0);
  XOTclAddIMethod(in, theobj, "destroy", XOTclODestroyMethod, 0, 0);
  XOTclAddIMethod(in, theobj, "exists",  (Tcl_ObjCmdProc*)XOTclOExistsMethod, 0, 0);
  XOTclAddIMethod(in, theobj, "filter", (Tcl_ObjCmdProc*)XOTclOFilterMethod, 0, 0);
  XOTclAddIMethod(in, theobj, "filterguard", (Tcl_ObjCmdProc*)XOTclOFilterGuardMethod, 0, 0);
  XOTclAddIMethod(in, theobj, "filtersearch", (Tcl_ObjCmdProc*)XOTclOFilterSearchMethod, 0, 0);
  XOTclAddIMethod(in, theobj, "getGuardedScope", (Tcl_ObjCmdProc*)XOTclOGetGuardedScopeMethod, 0, 0);
  XOTclAddIMethod(in, theobj, "incr", (Tcl_ObjCmdProc*)XOTclOIncrMethod, 0, 0);
  XOTclAddIMethod(in, theobj, "info", XOTclOInfoMethod, 0, 0);
  XOTclAddIMethod(in, theobj, "instvar", (Tcl_ObjCmdProc*)XOTclOInstVarMethod, 0, 0);
  XOTclAddIMethod(in, theobj, "invar", (Tcl_ObjCmdProc*)XOTclOInvariantsMethod, 0, 0);
  XOTclAddIMethod(in, theobj, "isclass", XOTclOIsClassMethod, 0, 0);
  XOTclAddIMethod(in, theobj, "ismetaclass", XOTclOIsMetaClassMethod, 0, 0);
  XOTclAddIMethod(in, theobj, "isobject", XOTclOIsObjectMethod, 0, 0);
  XOTclAddIMethod(in, theobj, "istype", XOTclOIsTypeMethod, 0, 0);
#ifdef XOTCL_METADATA
  XOTclAddIMethod(in, theobj, "metadata", (Tcl_ObjCmdProc*)XOTclOMetaDataMethod, 0, 0);
#endif
  XOTclAddIMethod(in, theobj, "mixin", (Tcl_ObjCmdProc*)XOTclOMixinMethod, 0, 0);
  XOTclAddIMethod(in, theobj, "parametercmd", (Tcl_ObjCmdProc*)XOTclCParameterCmdMethod, 0, 0);
  XOTclAddIMethod(in, theobj, "proc", XOTclOProcMethod, 0, 0);
  XOTclAddIMethod(in, theobj, "procsearch", (Tcl_ObjCmdProc*)XOTclOProcSearchMethod, 0, 0);
  XOTclAddIMethod(in, theobj, "requireNamespace", (Tcl_ObjCmdProc*)XOTclORequireNamespaceMethod, 0, 0);
  XOTclAddIMethod(in, theobj, "set",  (Tcl_ObjCmdProc*)XOTclOSetMethod, 0, 0);
  XOTclAddIMethod(in, theobj, "tclcmd", (Tcl_ObjCmdProc*)XOTclCTclCmdMethod, 0, 0);
  XOTclAddIMethod(in, theobj, "unset", XOTclOUnsetMethod, 0, 0);
  XOTclAddIMethod(in, theobj, "vwait", (Tcl_ObjCmdProc*)XOTclOVwaitMethod, 0, 0);



  XOTclAddIMethod(in, thecls, "alloc", (Tcl_ObjCmdProc*)XOTclCAllocMethod, 0, 0);
  XOTclAddIMethod(in, thecls, "create", (Tcl_ObjCmdProc*)XOTclCCreateMethod, 0, 0);
  XOTclAddIMethod(in, thecls, "new", (Tcl_ObjCmdProc*)XOTclCNewMethod, 0, 0);
  XOTclAddIMethod(in, thecls, "info", XOTclCInfoMethod, 0, 0);
  XOTclAddIMethod(in, thecls, "instdestroy", XOTclCInstDestroyMethod, 0, 0);
  XOTclAddIMethod(in, thecls, "instfilter", (Tcl_ObjCmdProc*)XOTclCInstFilterMethod, 0, 0);
  XOTclAddIMethod(in, thecls, "instfilterguard", (Tcl_ObjCmdProc*)XOTclCInstFilterGuardMethod, 0, 0);
  XOTclAddIMethod(in, thecls, "instinvar", (Tcl_ObjCmdProc*)XOTclCInvariantsMethod, 0, 0);
  XOTclAddIMethod(in, thecls, "instmixin", (Tcl_ObjCmdProc*)XOTclCInstMixinMethod, 0, 0);
  XOTclAddIMethod(in, thecls, "instparametercmd", (Tcl_ObjCmdProc*)XOTclCInstParameterCmdMethod, 0, 0);
  XOTclAddIMethod(in, thecls, "instproc", (Tcl_ObjCmdProc*)XOTclCInstProcMethod, 0, 0);
  XOTclAddIMethod(in, thecls, "insttclcmd", (Tcl_ObjCmdProc*)XOTclCInstTclCmdMethod, 0, 0);
  XOTclAddIMethod(in, thecls, "parameter", (Tcl_ObjCmdProc*)XOTclCParameterMethod, 0, 0);
  XOTclAddIMethod(in, thecls, "parameterclass", (Tcl_ObjCmdProc*)XOTclCParameterClassMethod, 0, 0);
  XOTclAddIMethod(in, thecls, "recreate", (Tcl_ObjCmdProc*) XOTclCRecreateMethod, 0, 0);
  XOTclAddIMethod(in, thecls, "superclass", (Tcl_ObjCmdProc*)XOTclCSuperClassMethod, 0, 0);
  XOTclAddIMethod(in, thecls, "unknown", (Tcl_ObjCmdProc*) XOTclCUnknownMethod, 0, 0);
  /*
   * overwritten tcl objs
   */
  /*Tcl_CreateObjCommand(in, "exit", XOTcl_ExitObjCmd, 0, 0);*/

  i = TclCommands(in, 1);
  if (i != TCL_OK)
    return i;

  /*
   * new tcl cmds
   */
  Tcl_CreateObjCommand(in, "::xotcl::interp", (Tcl_ObjCmdProc*)XOTcl_InterpObjCmd, 0, 0);
  Tcl_CreateObjCommand(in, "::xotcl::trace", XOTcl_TraceObjCmd, 0, 0);
  Tcl_CreateObjCommand(in, "::xotcl::namespace_copyvars", XOTcl_NSCopyVars, 0, 0);
  Tcl_CreateObjCommand(in, "::xotcl::namespace_copycmds", XOTcl_NSCopyCmds, 0, 0);
  Tcl_CreateObjCommand(in, "::xotcl::deprecated", XOTcl_DeprecatedCmd, 0, 0);
#ifdef XOTCL_BYTECODE
  instructions[INST_INITPROC].cmdPtr = (Command *)
#endif
  Tcl_CreateObjCommand(in, "::xotcl::initProcNS", XOTclInitProcNSCmd, 0, 0);
#ifdef XOTCL_BYTECODE
  instructions[INST_SELF_DISPATCH].cmdPtr = (Command *)
#endif
  Tcl_CreateObjCommand(in, "::xotcl::my", XOTclSelfDispatchCmd, 0, 0);

#ifdef XOTCL_BYTECODE
  XOTclBytecodeInit();
#endif

  /*
   *  Parameter Class
   */
  paramCl = PrimitiveCCreate(in, global_strings[PARAM_CL], thecls);
  XOTclAddPMethod(in, &paramCl->object, global_strings[SEARCH_DEFAULTS],
		  (Tcl_ObjCmdProc*) ParameterSearchDefaultsMethod, 0, 0);
  /*
   * set runtime version information in Tcl variable
   */
  Tcl_SetVar(in, "::xotcl::version", XOTCLVERSION, TCL_GLOBAL_ONLY);
  Tcl_SetVar(in, "::xotcl::patchlevel", XOTCLPATCHLEVEL, TCL_GLOBAL_ONLY);

  /*
   * with some methods and library procs in tcl - they could go in a
   * xotcl.tcl file, but they're embedded here with Tcl_GlobalEval
   * to avoid the need to carry around a separate file at runtime.
   */
  {

#include "predefined.h"

    /*fprintf(stderr, "predefined=<<%s>>\n",cmd); */
    if (Tcl_GlobalEval(in, cmd) != TCL_OK) return TCL_ERROR;

  }

#ifndef AOL_SERVER
  /* the AOL server uses a different package loading mechanism */
  {
    /* set xotcl_lib to the compiled in default value */
    Tcl_SetVar(in, "::xotcl::lib", XOLIBPKG, TCL_GLOBAL_ONLY);
    /* search for the location of the xotcl library */
    if (Tcl_GlobalEval(in, "::xotcl::config::check_library_path") != TCL_OK)
      return TCL_ERROR;
  }
  Tcl_PkgProvide(in, "XOTcl", XOTCLVERSION);
  Tcl_ResetResult(in);
  Tcl_SetIntObj(Tcl_GetObjResult(in), 1);
#endif

  return TCL_OK;
}

#ifdef NEVER
extern int
Xotcl_SafeInit(interp)
    Tcl_Interp *interp;
{
  /*** dummy for now **/
  return Xotcl_Init(interp);
}
#endif
